/* Generated by Cython 0.29.23 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "_pyfxr",
        "sources": [
            "_pyfxr.pyx"
        ]
    },
    "module_name": "_pyfxr"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_23"
#define CYTHON_HEX_VERSION 0x001D17F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE___pyfxr
#define __PYX_HAVE_API___pyfxr
/* Early includes */
#include <stdint.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "_pyfxr.pyx",
  "stringsource",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif


/*--- Type declarations ---*/
struct __pyx_obj_6_pyfxr_Wavetable;
struct __pyx_obj_6_pyfxr_SoundBuffer;
struct __pyx_obj_6_pyfxr_CachedSound;
struct __pyx_obj_6_pyfxr_PygletSource;

/* "_pyfxr.pyx":21
 * 
 * 
 * cdef class Wavetable:             # <<<<<<<<<<<<<<
 *     cdef int16_t[1024] wavetable
 * 
 */
struct __pyx_obj_6_pyfxr_Wavetable {
  PyObject_HEAD
  int16_t wavetable[0x400];
};


/* "_pyfxr.pyx":138
 * 
 * 
 * cdef class SoundBuffer:             # <<<<<<<<<<<<<<
 *     cdef size_t n_samples
 *     cdef int16_t *samples
 */
struct __pyx_obj_6_pyfxr_SoundBuffer {
  PyObject_HEAD
  size_t n_samples;
  int16_t *samples;
};


/* "_pyfxr.pyx":208
 * 
 * 
 * cdef class CachedSound:             # <<<<<<<<<<<<<<
 *     cdef SoundBuffer buf
 * 
 */
struct __pyx_obj_6_pyfxr_CachedSound {
  PyObject_HEAD
  struct __pyx_obj_6_pyfxr_SoundBuffer *buf;
};


/* "_pyfxr.pyx":243
 * 
 * 
 * cdef class PygletSource:             # <<<<<<<<<<<<<<
 *     cdef SoundBuffer buf
 *     cdef size_t pos
 */
struct __pyx_obj_6_pyfxr_PygletSource {
  PyObject_HEAD
  struct __pyx_obj_6_pyfxr_SoundBuffer *buf;
  size_t pos;
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_TrueDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_size_t(size_t value, Py_ssize_t width, char padding_char, char format_char);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_str(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_str(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int16_t(int16_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int16_t __Pyx_PyInt_As_int16_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from '_pyfxr' */
static PyTypeObject *__pyx_ptype_6_pyfxr_Wavetable = 0;
static PyTypeObject *__pyx_ptype_6_pyfxr_SoundBuffer = 0;
static PyTypeObject *__pyx_ptype_6_pyfxr_CachedSound = 0;
static PyTypeObject *__pyx_ptype_6_pyfxr_PygletSource = 0;
static float __pyx_v_6_pyfxr_AMPLITUDE;
static uint32_t __pyx_v_6_pyfxr_SAMPLE_RATE;
static int16_t __pyx_f_6_pyfxr_samp(float); /*proto*/
static void __pyx_f_6_pyfxr_reset_sample(float, float, float, float, float, float, float, float, double *, int *, double *, double *, double *, float *, float *, double *, int *, int *); /*proto*/
static void __pyx_f_6_pyfxr_fill_noise(float *); /*proto*/
static float __pyx_f_6_pyfxr_frnd(float); /*proto*/
static PyObject *__pyx_f_6_pyfxr___pyx_unpickle_Wavetable__set_state(struct __pyx_obj_6_pyfxr_Wavetable *, PyObject *); /*proto*/
static PyObject *__pyx_f_6_pyfxr___pyx_unpickle_PygletSource__set_state(struct __pyx_obj_6_pyfxr_PygletSource *, PyObject *); /*proto*/
static void __pyx_fuse_0__pyx_f_6_pyfxr_clamp(float *, float, float); /*proto*/
static CYTHON_INLINE PyObject *__Pyx_carray_to_py_int16_t(int16_t *, Py_ssize_t); /*proto*/
static CYTHON_INLINE PyObject *__Pyx_carray_to_tuple_int16_t(int16_t *, Py_ssize_t); /*proto*/
static int __Pyx_carray_from_py_int16_t(PyObject *, int16_t *, Py_ssize_t); /*proto*/
#define __Pyx_MODULE_NAME "_pyfxr"
extern int __pyx_module_is_main__pyfxr;
int __pyx_module_is_main__pyfxr = 0;

/* Implementation of '_pyfxr' */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_round;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_OverflowError;
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_fp[] = "fp";
static const char __pyx_k_ni[] = "ni";
static const char __pyx_k_pp[] = "pp";
static const char __pyx_k_si[] = "si";
static const char __pyx_k_wb[] = "wb";
static const char __pyx_k_buf[] = "buf";
static const char __pyx_k_gen[] = "gen";
static const char __pyx_k_get[] = "_get";
static const char __pyx_k_ipp[] = "ipp";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_saw[] = "saw";
static const char __pyx_k_sfx[] = "sfx";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_fltp[] = "fltp";
static const char __pyx_k_fltw[] = "fltw";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_prev[] = "prev";
static const char __pyx_k_sine[] = "sine";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_tone[] = "tone";
static const char __pyx_k_wave[] = "wave";
static const char __pyx_k_build[] = "_build";
static const char __pyx_k_decay[] = "decay";
static const char __pyx_k_delay[] = "delay";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_fltdp[] = "fltdp";
static const char __pyx_k_flthp[] = "flthp";
static const char __pyx_k_omega[] = "omega";
static const char __pyx_k_phase[] = "phase";
static const char __pyx_k_pitch[] = "pitch";
static const char __pyx_k_pluck[] = "pluck";
static const char __pyx_k_pyfxr[] = "_pyfxr";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_round[] = "round";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_attack[] = "attack";
static const char __pyx_k_fltdmp[] = "fltdmp";
static const char __pyx_k_fltphp[] = "fltphp";
static const char __pyx_k_fltw_d[] = "fltw_d";
static const char __pyx_k_fphase[] = "fphase";
static const char __pyx_k_fslide[] = "fslide";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_iphase[] = "iphase";
static const char __pyx_k_p_duty[] = "p_duty";
static const char __pyx_k_period[] = "period";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_sample[] = "sample";
static const char __pyx_k_square[] = "square";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_arp_mod[] = "arp_mod";
static const char __pyx_k_env_vol[] = "env_vol";
static const char __pyx_k_fdphase[] = "fdphase";
static const char __pyx_k_fdslide[] = "fdslide";
static const char __pyx_k_flthp_d[] = "flthp_d";
static const char __pyx_k_fperiod[] = "fperiod";
static const char __pyx_k_fsample[] = "fsample";
static const char __pyx_k_randval[] = "randval";
static const char __pyx_k_release[] = "release";
static const char __pyx_k_samples[] = "samples";
static const char __pyx_k_ssample[] = "ssample";
static const char __pyx_k_sustain[] = "sustain";
static const char __pyx_k_vib_amp[] = "vib_amp";
static const char __pyx_k_arp_time[] = "arp_time";
static const char __pyx_k_channels[] = "channels";
static const char __pyx_k_duration[] = "duration";
static const char __pyx_k_env_time[] = "env_time";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_rep_time[] = "rep_time";
static const char __pyx_k_rfperiod[] = "rfperiod";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_triangle[] = "triangle";
static const char __pyx_k_AudioData[] = "AudioData";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_Wavetable[] = "Wavetable";
static const char __pyx_k_amplitude[] = "amplitude";
static const char __pyx_k_arp_limit[] = "arp_limit";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_env_stage[] = "env_stage";
static const char __pyx_k_for_pitch[] = " for pitch ";
static const char __pyx_k_n_samples[] = "n_samples";
static const char __pyx_k_p_arp_mod[] = "p_arp_mod";
static const char __pyx_k_pyfxr_pyx[] = "_pyfxr.pyx";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_rep_limit[] = "rep_limit";
static const char __pyx_k_vib_phase[] = "vib_phase";
static const char __pyx_k_vib_speed[] = "vib_speed";
static const char __pyx_k_wave_type[] = "wave_type";
static const char __pyx_k_wavetable[] = "wavetable";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_duty_cycle[] = "duty_cycle";
static const char __pyx_k_env_length[] = "env_length";
static const char __pyx_k_fmaxperiod[] = "fmaxperiod";
static const char __pyx_k_p_hpf_freq[] = "p_hpf_freq";
static const char __pyx_k_p_hpf_ramp[] = "p_hpf_ramp";
static const char __pyx_k_p_lpf_freq[] = "p_lpf_freq";
static const char __pyx_k_p_lpf_ramp[] = "p_lpf_ramp";
static const char __pyx_k_p_pha_ramp[] = "p_pha_ramp";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_setnframes[] = "setnframes";
static const char __pyx_k_AudioFormat[] = "AudioFormat";
static const char __pyx_k_CachedSound[] = "CachedSound";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_SoundBuffer[] = "SoundBuffer";
static const char __pyx_k_p_arp_speed[] = "p_arp_speed";
static const char __pyx_k_p_base_freq[] = "p_base_freq";
static const char __pyx_k_p_duty_ramp[] = "p_duty_ramp";
static const char __pyx_k_p_env_decay[] = "p_env_decay";
static const char __pyx_k_p_env_punch[] = "p_env_punch";
static const char __pyx_k_p_freq_ramp[] = "p_freq_ramp";
static const char __pyx_k_p_vib_delay[] = "p_vib_delay";
static const char __pyx_k_p_vib_speed[] = "p_vib_speed";
static const char __pyx_k_sample_rate[] = "sample_rate";
static const char __pyx_k_sample_size[] = "sample_size";
static const char __pyx_k_square_duty[] = "square_duty";
static const char __pyx_k_PygletSource[] = "PygletSource";
static const char __pyx_k_noise_buffer[] = "noise_buffer";
static const char __pyx_k_p_env_attack[] = "p_env_attack";
static const char __pyx_k_p_freq_dramp[] = "p_freq_dramp";
static const char __pyx_k_p_freq_limit[] = "p_freq_limit";
static const char __pyx_k_p_pha_offset[] = "p_pha_offset";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_setframerate[] = "setframerate";
static const char __pyx_k_setnchannels[] = "setnchannels";
static const char __pyx_k_setsampwidth[] = "setsampwidth";
static const char __pyx_k_square_slide[] = "square_slide";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_video_format[] = "video_format";
static const char __pyx_k_OverflowError[] = "OverflowError";
static const char __pyx_k_from_function[] = "from_function";
static const char __pyx_k_p_env_sustain[] = "p_env_sustain";
static const char __pyx_k_phaser_buffer[] = "phaser_buffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_p_repeat_speed[] = "p_repeat_speed";
static const char __pyx_k_p_vib_strength[] = "p_vib_strength";
static const char __pyx_k_writeframesraw[] = "writeframesraw";
static const char __pyx_k_p_lpf_resonance[] = "p_lpf_resonance";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_release_samples[] = "release_samples";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_index_out_of_range[] = "index out of range";
static const char __pyx_k_pyglet_media_codecs[] = "pyglet.media.codecs";
static const char __pyx_k_pyx_unpickle_Wavetable[] = "__pyx_unpickle_Wavetable";
static const char __pyx_k_pyx_unpickle_PygletSource[] = "__pyx_unpickle_PygletSource";
static const char __pyx_k_n_samples_must_be_at_least[] = "n_samples must be at least ";
static const char __pyx_k_Incompatible_checksums_s_vs_0x58[] = "Incompatible checksums (%s vs 0x58d03cd = (buf, pos))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x6f[] = "Incompatible checksums (%s vs 0x6f0316b = (wavetable))";
static const char __pyx_k_Wavetable_generator_generated_to[] = "Wavetable generator generated too few values.";
static const char __pyx_k_duty_cycle_must_be_between_0_and[] = "duty_cycle must be between 0 and 1";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_n_s_AudioData;
static PyObject *__pyx_n_s_AudioFormat;
static PyObject *__pyx_n_s_CachedSound;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x58;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x6f;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_OverflowError;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_PygletSource;
static PyObject *__pyx_n_s_SoundBuffer;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_Wavetable;
static PyObject *__pyx_kp_u_Wavetable_generator_generated_to;
static PyObject *__pyx_n_s_amplitude;
static PyObject *__pyx_n_s_arp_limit;
static PyObject *__pyx_n_s_arp_mod;
static PyObject *__pyx_n_s_arp_time;
static PyObject *__pyx_n_s_attack;
static PyObject *__pyx_n_s_buf;
static PyObject *__pyx_n_s_build;
static PyObject *__pyx_n_s_channels;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_decay;
static PyObject *__pyx_n_s_delay;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_duration;
static PyObject *__pyx_n_s_duty_cycle;
static PyObject *__pyx_kp_u_duty_cycle_must_be_between_0_and;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_env_length;
static PyObject *__pyx_n_s_env_stage;
static PyObject *__pyx_n_s_env_time;
static PyObject *__pyx_n_s_env_vol;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_fdphase;
static PyObject *__pyx_n_s_fdslide;
static PyObject *__pyx_n_s_fltdmp;
static PyObject *__pyx_n_s_fltdp;
static PyObject *__pyx_n_s_flthp;
static PyObject *__pyx_n_s_flthp_d;
static PyObject *__pyx_n_s_fltp;
static PyObject *__pyx_n_s_fltphp;
static PyObject *__pyx_n_s_fltw;
static PyObject *__pyx_n_s_fltw_d;
static PyObject *__pyx_n_s_fmaxperiod;
static PyObject *__pyx_kp_u_for_pitch;
static PyObject *__pyx_n_s_fp;
static PyObject *__pyx_n_s_fperiod;
static PyObject *__pyx_n_s_fphase;
static PyObject *__pyx_n_s_from_function;
static PyObject *__pyx_n_s_fsample;
static PyObject *__pyx_n_s_fslide;
static PyObject *__pyx_n_s_gen;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_kp_u_index_out_of_range;
static PyObject *__pyx_n_s_iphase;
static PyObject *__pyx_n_s_ipp;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_n_samples;
static PyObject *__pyx_kp_u_n_samples_must_be_at_least;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_ni;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_noise_buffer;
static PyObject *__pyx_n_s_omega;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_p_arp_mod;
static PyObject *__pyx_n_s_p_arp_speed;
static PyObject *__pyx_n_s_p_base_freq;
static PyObject *__pyx_n_s_p_duty;
static PyObject *__pyx_n_s_p_duty_ramp;
static PyObject *__pyx_n_s_p_env_attack;
static PyObject *__pyx_n_s_p_env_decay;
static PyObject *__pyx_n_s_p_env_punch;
static PyObject *__pyx_n_s_p_env_sustain;
static PyObject *__pyx_n_s_p_freq_dramp;
static PyObject *__pyx_n_s_p_freq_limit;
static PyObject *__pyx_n_s_p_freq_ramp;
static PyObject *__pyx_n_s_p_hpf_freq;
static PyObject *__pyx_n_s_p_hpf_ramp;
static PyObject *__pyx_n_s_p_lpf_freq;
static PyObject *__pyx_n_s_p_lpf_ramp;
static PyObject *__pyx_n_s_p_lpf_resonance;
static PyObject *__pyx_n_s_p_pha_offset;
static PyObject *__pyx_n_s_p_pha_ramp;
static PyObject *__pyx_n_s_p_repeat_speed;
static PyObject *__pyx_n_s_p_vib_delay;
static PyObject *__pyx_n_s_p_vib_speed;
static PyObject *__pyx_n_s_p_vib_strength;
static PyObject *__pyx_n_s_period;
static PyObject *__pyx_n_s_phase;
static PyObject *__pyx_n_s_phaser_buffer;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pitch;
static PyObject *__pyx_n_s_pluck;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_s_pp;
static PyObject *__pyx_n_s_prev;
static PyObject *__pyx_n_s_pyfxr;
static PyObject *__pyx_kp_s_pyfxr_pyx;
static PyObject *__pyx_n_s_pyglet_media_codecs;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_PygletSource;
static PyObject *__pyx_n_s_pyx_unpickle_Wavetable;
static PyObject *__pyx_n_s_randval;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_release;
static PyObject *__pyx_n_s_release_samples;
static PyObject *__pyx_n_s_rep_limit;
static PyObject *__pyx_n_s_rep_time;
static PyObject *__pyx_n_s_rfperiod;
static PyObject *__pyx_n_s_round;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_sample;
static PyObject *__pyx_n_s_sample_rate;
static PyObject *__pyx_n_s_sample_size;
static PyObject *__pyx_n_s_samples;
static PyObject *__pyx_n_s_saw;
static PyObject *__pyx_n_s_setframerate;
static PyObject *__pyx_n_s_setnchannels;
static PyObject *__pyx_n_s_setnframes;
static PyObject *__pyx_n_s_setsampwidth;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_sfx;
static PyObject *__pyx_n_s_si;
static PyObject *__pyx_n_s_sine;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_square;
static PyObject *__pyx_n_s_square_duty;
static PyObject *__pyx_n_s_square_slide;
static PyObject *__pyx_n_s_ssample;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_sustain;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_tone;
static PyObject *__pyx_n_s_triangle;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_vib_amp;
static PyObject *__pyx_n_s_vib_phase;
static PyObject *__pyx_n_s_vib_speed;
static PyObject *__pyx_n_s_video_format;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_n_s_wave;
static PyObject *__pyx_n_s_wave_type;
static PyObject *__pyx_n_s_wavetable;
static PyObject *__pyx_n_u_wb;
static PyObject *__pyx_n_s_writeframesraw;
static int __pyx_pf_6_pyfxr_9Wavetable___init__(struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self, PyObject *__pyx_v_gen); /* proto */
static PyObject *__pyx_pf_6_pyfxr_9Wavetable_2from_function(PyObject *__pyx_v_f); /* proto */
static PyObject *__pyx_pf_6_pyfxr_9Wavetable_4sine(void); /* proto */
static PyObject *__pyx_pf_6_pyfxr_9Wavetable_6triangle(void); /* proto */
static PyObject *__pyx_pf_6_pyfxr_9Wavetable_8saw(void); /* proto */
static PyObject *__pyx_pf_6_pyfxr_9Wavetable_10square(float __pyx_v_duty_cycle); /* proto */
static int __pyx_pf_6_pyfxr_9Wavetable_12__getbuffer__(struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_6_pyfxr_9Wavetable_14__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_6_pyfxr_9Wavetable_16__reduce_cython__(struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_9Wavetable_18__setstate_cython__(struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6_pyfxr_11SoundBuffer___cinit__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, size_t __pyx_v_n_samples); /* proto */
static void __pyx_pf_6_pyfxr_11SoundBuffer_2__dealloc__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_6_pyfxr_11SoundBuffer_4__len__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_6__getitem__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, Py_ssize_t __pyx_v_i); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_8duration___get__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_8save(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, PyObject *__pyx_v_filename); /* proto */
static int __pyx_pf_6_pyfxr_11SoundBuffer_10__getbuffer__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_6_pyfxr_11SoundBuffer_12__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_14get_queue_source(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6_pyfxr_11CachedSound___cinit__(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11CachedSound_2_clear(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11CachedSound_4_set(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self, struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_sound); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11CachedSound_6_get(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self); /* proto */
static struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_pf_6_pyfxr_11CachedSound_8_build(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self); /* proto */
static int __pyx_pf_6_pyfxr_11CachedSound_10__getbuffer__(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /* proto */
static void __pyx_pf_6_pyfxr_11CachedSound_12__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11CachedSound_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_11CachedSound_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6_pyfxr_12PygletSource___init__(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self, struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_buf); /* proto */
static PyObject *__pyx_pf_6_pyfxr_12PygletSource_12audio_format___get__(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_12PygletSource_2get_audio_data(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_length); /* proto */
static PyObject *__pyx_pf_6_pyfxr_12PygletSource_4__reduce_cython__(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6_pyfxr_12PygletSource_6__setstate_cython__(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6_pyfxr_tone(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_wavetable, double __pyx_v_pitch, uint32_t __pyx_v_attack, uint32_t __pyx_v_decay, uint32_t __pyx_v_sustain, uint32_t __pyx_v_release); /* proto */
static PyObject *__pyx_pf_6_pyfxr_2sfx(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_wave_type, float __pyx_v_p_base_freq, float __pyx_v_p_freq_limit, float __pyx_v_p_freq_ramp, float __pyx_v_p_freq_dramp, float __pyx_v_p_duty, float __pyx_v_p_duty_ramp, float __pyx_v_p_vib_strength, float __pyx_v_p_vib_speed, CYTHON_UNUSED float __pyx_v_p_vib_delay, float __pyx_v_p_env_attack, float __pyx_v_p_env_sustain, float __pyx_v_p_env_decay, float __pyx_v_p_env_punch, float __pyx_v_p_lpf_resonance, float __pyx_v_p_lpf_freq, float __pyx_v_p_lpf_ramp, float __pyx_v_p_hpf_freq, float __pyx_v_p_hpf_ramp, float __pyx_v_p_pha_offset, float __pyx_v_p_pha_ramp, float __pyx_v_p_repeat_speed, float __pyx_v_p_arp_speed, float __pyx_v_p_arp_mod); /* proto */
static PyObject *__pyx_pf_6_pyfxr_4pluck(CYTHON_UNUSED PyObject *__pyx_self, float __pyx_v_duration, float __pyx_v_pitch, float __pyx_v_release); /* proto */
static PyObject *__pyx_pf_6_pyfxr_6__pyx_unpickle_Wavetable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6_pyfxr_8__pyx_unpickle_PygletSource(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6_pyfxr_Wavetable(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6_pyfxr_SoundBuffer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6_pyfxr_CachedSound(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6_pyfxr_PygletSource(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_16;
static PyObject *__pyx_int_512;
static PyObject *__pyx_int_93127629;
static PyObject *__pyx_int_116404587;
static PyObject *__pyx_int_4294967296;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_codeobj__10;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__14;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
/* Late includes */

/* "_pyfxr.pyx":16
 * 
 * 
 * cdef int16_t samp(float v) nogil:             # <<<<<<<<<<<<<<
 *     """Convert a float in [-1, 1] to an int16_t sample."""
 *     return <int16_t> floor(v * AMPLITUDE)
 */

static int16_t __pyx_f_6_pyfxr_samp(float __pyx_v_v) {
  int16_t __pyx_r;

  /* "_pyfxr.pyx":18
 * cdef int16_t samp(float v) nogil:
 *     """Convert a float in [-1, 1] to an int16_t sample."""
 *     return <int16_t> floor(v * AMPLITUDE)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((int16_t)floor((__pyx_v_v * __pyx_v_6_pyfxr_AMPLITUDE)));
  goto __pyx_L0;

  /* "_pyfxr.pyx":16
 * 
 * 
 * cdef int16_t samp(float v) nogil:             # <<<<<<<<<<<<<<
 *     """Convert a float in [-1, 1] to an int16_t sample."""
 *     return <int16_t> floor(v * AMPLITUDE)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "_pyfxr.pyx":24
 *     cdef int16_t[1024] wavetable
 * 
 *     def __init__(self, gen):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i, val in enumerate(gen):
 */

/* Python wrapper */
static int __pyx_pw_6_pyfxr_9Wavetable_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6_pyfxr_9Wavetable_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_gen = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gen,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gen)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 24, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_gen = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 24, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.Wavetable.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6_pyfxr_9Wavetable___init__(((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_v_self), __pyx_v_gen);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6_pyfxr_9Wavetable___init__(struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self, PyObject *__pyx_v_gen) {
  int __pyx_v_i;
  PyObject *__pyx_v_val = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  float __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "_pyfxr.pyx":26
 *     def __init__(self, gen):
 *         cdef int i
 *         for i, val in enumerate(gen):             # <<<<<<<<<<<<<<
 *             self.wavetable[i] = samp(val)
 *             if i == 1023:
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_gen)) || PyTuple_CheckExact(__pyx_v_gen)) {
    __pyx_t_2 = __pyx_v_gen; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_gen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 26, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 26, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 26, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 26, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 26, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 26, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_i = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "_pyfxr.pyx":27
 *         cdef int i
 *         for i, val in enumerate(gen):
 *             self.wavetable[i] = samp(val)             # <<<<<<<<<<<<<<
 *             if i == 1023:
 *                 return
 */
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_v_val); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 27, __pyx_L1_error)
    (__pyx_v_self->wavetable[__pyx_v_i]) = __pyx_f_6_pyfxr_samp(__pyx_t_6);

    /* "_pyfxr.pyx":28
 *         for i, val in enumerate(gen):
 *             self.wavetable[i] = samp(val)
 *             if i == 1023:             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
    __pyx_t_7 = ((__pyx_v_i == 0x3FF) != 0);
    if (__pyx_t_7) {

      /* "_pyfxr.pyx":29
 *             self.wavetable[i] = samp(val)
 *             if i == 1023:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         if i != 1023:
 */
      __pyx_r = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "_pyfxr.pyx":28
 *         for i, val in enumerate(gen):
 *             self.wavetable[i] = samp(val)
 *             if i == 1023:             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
    }

    /* "_pyfxr.pyx":26
 *     def __init__(self, gen):
 *         cdef int i
 *         for i, val in enumerate(gen):             # <<<<<<<<<<<<<<
 *             self.wavetable[i] = samp(val)
 *             if i == 1023:
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pyfxr.pyx":31
 *                 return
 * 
 *         if i != 1023:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "Wavetable generator generated too few values."
 */
  __pyx_t_7 = ((__pyx_v_i != 0x3FF) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "_pyfxr.pyx":32
 * 
 *         if i != 1023:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 "Wavetable generator generated too few values."
 *             )
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 32, __pyx_L1_error)

    /* "_pyfxr.pyx":31
 *                 return
 * 
 *         if i != 1023:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "Wavetable generator generated too few values."
 */
  }

  /* "_pyfxr.pyx":24
 *     cdef int16_t[1024] wavetable
 * 
 *     def __init__(self, gen):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i, val in enumerate(gen):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_pyfxr.Wavetable.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":37
 * 
 *     @staticmethod
 *     def from_function(f):             # <<<<<<<<<<<<<<
 *         """Generate a wavetable by calling a function f.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_3from_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_9Wavetable_2from_function[] = "Wavetable.from_function(f)\nGenerate a wavetable by calling a function f.\n\n        f should take a single float argument between 0 and tau (pi * 2) and\n        return values in [-1, 1].\n        ";
static PyMethodDef __pyx_mdef_6_pyfxr_9Wavetable_3from_function = {"from_function", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_3from_function, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_2from_function};
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_3from_function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_f = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("from_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_f,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_f)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "from_function") < 0)) __PYX_ERR(0, 37, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_f = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("from_function", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 37, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.Wavetable.from_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6_pyfxr_9Wavetable_2from_function(__pyx_v_f);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_9Wavetable_2from_function(PyObject *__pyx_v_f) {
  struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_w = 0;
  double __pyx_v_t;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  float __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_function", 0);

  /* "_pyfxr.pyx":46
 *         cdef double t
 * 
 *         w = Wavetable.__new__(Wavetable)             # <<<<<<<<<<<<<<
 *         for i in range(1024):
 *             t = i * pi / 512
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6_pyfxr_Wavetable(((PyTypeObject *)__pyx_ptype_6_pyfxr_Wavetable), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_w = ((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "_pyfxr.pyx":47
 * 
 *         w = Wavetable.__new__(Wavetable)
 *         for i in range(1024):             # <<<<<<<<<<<<<<
 *             t = i * pi / 512
 *             w.wavetable[i] = samp(f(t))
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 0x400; __pyx_t_2+=1) {
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "_pyfxr.pyx":48
 *         w = Wavetable.__new__(Wavetable)
 *         for i in range(1024):
 *             t = i * pi / 512             # <<<<<<<<<<<<<<
 *             w.wavetable[i] = samp(f(t))
 *         return w
 */
    __pyx_t_1 = PyFloat_FromDouble(M_PI); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyNumber_Multiply(__pyx_v_i, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_TrueDivideObjC(__pyx_t_3, __pyx_int_512, 0x200, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_t = __pyx_t_4;

    /* "_pyfxr.pyx":49
 *         for i in range(1024):
 *             t = i * pi / 512
 *             w.wavetable[i] = samp(f(t))             # <<<<<<<<<<<<<<
 *         return w
 * 
 */
    __pyx_t_3 = PyFloat_FromDouble(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_f);
    __pyx_t_5 = __pyx_v_f; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_7 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L1_error)
    (__pyx_v_w->wavetable[__pyx_t_8]) = __pyx_f_6_pyfxr_samp(__pyx_t_7);
  }

  /* "_pyfxr.pyx":50
 *             t = i * pi / 512
 *             w.wavetable[i] = samp(f(t))
 *         return w             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_w));
  __pyx_r = ((PyObject *)__pyx_v_w);
  goto __pyx_L0;

  /* "_pyfxr.pyx":37
 * 
 *     @staticmethod
 *     def from_function(f):             # <<<<<<<<<<<<<<
 *         """Generate a wavetable by calling a function f.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("_pyfxr.Wavetable.from_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_w);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":53
 * 
 *     @staticmethod
 *     def sine():             # <<<<<<<<<<<<<<
 *         """Construct a sine waveform."""
 *         cdef Wavetable w
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_5sine(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_9Wavetable_4sine[] = "Wavetable.sine()\nConstruct a sine waveform.";
static PyMethodDef __pyx_mdef_6_pyfxr_9Wavetable_5sine = {"sine", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_5sine, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_4sine};
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_5sine(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sine (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("sine", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "sine", 0))) return NULL;
  __pyx_r = __pyx_pf_6_pyfxr_9Wavetable_4sine();

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_9Wavetable_4sine(void) {
  struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_w = 0;
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sine", 0);

  /* "_pyfxr.pyx":57
 *         cdef Wavetable w
 *         cdef size_t i
 *         w = Wavetable.__new__(Wavetable)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             for i in range(1024):
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6_pyfxr_Wavetable(((PyTypeObject *)__pyx_ptype_6_pyfxr_Wavetable), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_w = ((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "_pyfxr.pyx":58
 *         cdef size_t i
 *         w = Wavetable.__new__(Wavetable)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(1024):
 *                 w.wavetable[i] = samp(
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "_pyfxr.pyx":59
 *         w = Wavetable.__new__(Wavetable)
 *         with nogil:
 *             for i in range(1024):             # <<<<<<<<<<<<<<
 *                 w.wavetable[i] = samp(
 *                     sin(pi * 2.0 * i / 1023.0)
 */
        for (__pyx_t_2 = 0; __pyx_t_2 < 0x400; __pyx_t_2+=1) {
          __pyx_v_i = __pyx_t_2;

          /* "_pyfxr.pyx":60
 *         with nogil:
 *             for i in range(1024):
 *                 w.wavetable[i] = samp(             # <<<<<<<<<<<<<<
 *                     sin(pi * 2.0 * i / 1023.0)
 *                 )
 */
          (__pyx_v_w->wavetable[__pyx_v_i]) = __pyx_f_6_pyfxr_samp(sin((((M_PI * 2.0) * __pyx_v_i) / 1023.0)));
        }
      }

      /* "_pyfxr.pyx":58
 *         cdef size_t i
 *         w = Wavetable.__new__(Wavetable)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(1024):
 *                 w.wavetable[i] = samp(
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "_pyfxr.pyx":63
 *                     sin(pi * 2.0 * i / 1023.0)
 *                 )
 *         return w             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_w));
  __pyx_r = ((PyObject *)__pyx_v_w);
  goto __pyx_L0;

  /* "_pyfxr.pyx":53
 * 
 *     @staticmethod
 *     def sine():             # <<<<<<<<<<<<<<
 *         """Construct a sine waveform."""
 *         cdef Wavetable w
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.Wavetable.sine", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_w);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":66
 * 
 *     @staticmethod
 *     def triangle():             # <<<<<<<<<<<<<<
 *         """Construct a triangle waveform."""
 *         cdef Wavetable w
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_7triangle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_9Wavetable_6triangle[] = "Wavetable.triangle()\nConstruct a triangle waveform.";
static PyMethodDef __pyx_mdef_6_pyfxr_9Wavetable_7triangle = {"triangle", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_7triangle, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_6triangle};
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_7triangle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("triangle (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("triangle", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "triangle", 0))) return NULL;
  __pyx_r = __pyx_pf_6_pyfxr_9Wavetable_6triangle();

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_9Wavetable_6triangle(void) {
  struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_w = 0;
  size_t __pyx_v_i;
  float __pyx_v_v;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("triangle", 0);

  /* "_pyfxr.pyx":71
 *         cdef size_t i
 *         cdef float v
 *         w = Wavetable.__new__(Wavetable)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             for i in range(1024):
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6_pyfxr_Wavetable(((PyTypeObject *)__pyx_ptype_6_pyfxr_Wavetable), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_w = ((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "_pyfxr.pyx":72
 *         cdef float v
 *         w = Wavetable.__new__(Wavetable)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(1024):
 *                 if i < 256:
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "_pyfxr.pyx":73
 *         w = Wavetable.__new__(Wavetable)
 *         with nogil:
 *             for i in range(1024):             # <<<<<<<<<<<<<<
 *                 if i < 256:
 *                     v = i / 256
 */
        for (__pyx_t_2 = 0; __pyx_t_2 < 0x400; __pyx_t_2+=1) {
          __pyx_v_i = __pyx_t_2;

          /* "_pyfxr.pyx":74
 *         with nogil:
 *             for i in range(1024):
 *                 if i < 256:             # <<<<<<<<<<<<<<
 *                     v = i / 256
 *                 elif i < 768:
 */
          __pyx_t_3 = ((__pyx_v_i < 0x100) != 0);
          if (__pyx_t_3) {

            /* "_pyfxr.pyx":75
 *             for i in range(1024):
 *                 if i < 256:
 *                     v = i / 256             # <<<<<<<<<<<<<<
 *                 elif i < 768:
 *                     v = 1.0 - (i - 255) / 256
 */
            __pyx_v_v = (((double)__pyx_v_i) / 256.0);

            /* "_pyfxr.pyx":74
 *         with nogil:
 *             for i in range(1024):
 *                 if i < 256:             # <<<<<<<<<<<<<<
 *                     v = i / 256
 *                 elif i < 768:
 */
            goto __pyx_L8;
          }

          /* "_pyfxr.pyx":76
 *                 if i < 256:
 *                     v = i / 256
 *                 elif i < 768:             # <<<<<<<<<<<<<<
 *                     v = 1.0 - (i - 255) / 256
 *                 else:
 */
          __pyx_t_3 = ((__pyx_v_i < 0x300) != 0);
          if (__pyx_t_3) {

            /* "_pyfxr.pyx":77
 *                     v = i / 256
 *                 elif i < 768:
 *                     v = 1.0 - (i - 255) / 256             # <<<<<<<<<<<<<<
 *                 else:
 *                     v = (i - 768) / 256 - 1.0
 */
            __pyx_v_v = (1.0 - (((double)(__pyx_v_i - 0xFF)) / 256.0));

            /* "_pyfxr.pyx":76
 *                 if i < 256:
 *                     v = i / 256
 *                 elif i < 768:             # <<<<<<<<<<<<<<
 *                     v = 1.0 - (i - 255) / 256
 *                 else:
 */
            goto __pyx_L8;
          }

          /* "_pyfxr.pyx":79
 *                     v = 1.0 - (i - 255) / 256
 *                 else:
 *                     v = (i - 768) / 256 - 1.0             # <<<<<<<<<<<<<<
 *                 w.wavetable[i] = samp(v)
 *         return w
 */
          /*else*/ {
            __pyx_v_v = ((((double)(__pyx_v_i - 0x300)) / 256.0) - 1.0);
          }
          __pyx_L8:;

          /* "_pyfxr.pyx":80
 *                 else:
 *                     v = (i - 768) / 256 - 1.0
 *                 w.wavetable[i] = samp(v)             # <<<<<<<<<<<<<<
 *         return w
 * 
 */
          (__pyx_v_w->wavetable[__pyx_v_i]) = __pyx_f_6_pyfxr_samp(__pyx_v_v);
        }
      }

      /* "_pyfxr.pyx":72
 *         cdef float v
 *         w = Wavetable.__new__(Wavetable)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(1024):
 *                 if i < 256:
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "_pyfxr.pyx":81
 *                     v = (i - 768) / 256 - 1.0
 *                 w.wavetable[i] = samp(v)
 *         return w             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_w));
  __pyx_r = ((PyObject *)__pyx_v_w);
  goto __pyx_L0;

  /* "_pyfxr.pyx":66
 * 
 *     @staticmethod
 *     def triangle():             # <<<<<<<<<<<<<<
 *         """Construct a triangle waveform."""
 *         cdef Wavetable w
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.Wavetable.triangle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_w);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":84
 * 
 *     @staticmethod
 *     def saw():             # <<<<<<<<<<<<<<
 *         """Construct a saw waveform."""
 *         cdef Wavetable w
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_9saw(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_9Wavetable_8saw[] = "Wavetable.saw()\nConstruct a saw waveform.";
static PyMethodDef __pyx_mdef_6_pyfxr_9Wavetable_9saw = {"saw", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_9saw, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_8saw};
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_9saw(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("saw (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("saw", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "saw", 0))) return NULL;
  __pyx_r = __pyx_pf_6_pyfxr_9Wavetable_8saw();

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_9Wavetable_8saw(void) {
  struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_w = 0;
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("saw", 0);

  /* "_pyfxr.pyx":89
 *         cdef size_t i
 *         cdef float v
 *         w = Wavetable.__new__(Wavetable)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             for i in range(1024):
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6_pyfxr_Wavetable(((PyTypeObject *)__pyx_ptype_6_pyfxr_Wavetable), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_w = ((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "_pyfxr.pyx":90
 *         cdef float v
 *         w = Wavetable.__new__(Wavetable)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(1024):
 *                 w.wavetable[(i + 512) % 1024] = samp(i / 512.0 - 1.0)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "_pyfxr.pyx":91
 *         w = Wavetable.__new__(Wavetable)
 *         with nogil:
 *             for i in range(1024):             # <<<<<<<<<<<<<<
 *                 w.wavetable[(i + 512) % 1024] = samp(i / 512.0 - 1.0)
 *         return w
 */
        for (__pyx_t_2 = 0; __pyx_t_2 < 0x400; __pyx_t_2+=1) {
          __pyx_v_i = __pyx_t_2;

          /* "_pyfxr.pyx":92
 *         with nogil:
 *             for i in range(1024):
 *                 w.wavetable[(i + 512) % 1024] = samp(i / 512.0 - 1.0)             # <<<<<<<<<<<<<<
 *         return w
 * 
 */
          (__pyx_v_w->wavetable[((__pyx_v_i + 0x200) % 0x400)]) = __pyx_f_6_pyfxr_samp(((((double)__pyx_v_i) / 512.0) - 1.0));
        }
      }

      /* "_pyfxr.pyx":90
 *         cdef float v
 *         w = Wavetable.__new__(Wavetable)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(1024):
 *                 w.wavetable[(i + 512) % 1024] = samp(i / 512.0 - 1.0)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "_pyfxr.pyx":93
 *             for i in range(1024):
 *                 w.wavetable[(i + 512) % 1024] = samp(i / 512.0 - 1.0)
 *         return w             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_w));
  __pyx_r = ((PyObject *)__pyx_v_w);
  goto __pyx_L0;

  /* "_pyfxr.pyx":84
 * 
 *     @staticmethod
 *     def saw():             # <<<<<<<<<<<<<<
 *         """Construct a saw waveform."""
 *         cdef Wavetable w
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.Wavetable.saw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_w);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":96
 * 
 *     @staticmethod
 *     def square(float duty_cycle=0.5):             # <<<<<<<<<<<<<<
 *         """Generate a square-wave waveform.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_11square(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_9Wavetable_10square[] = "Wavetable.square(float duty_cycle=0.5)\nGenerate a square-wave waveform.\n\n        duty_cycle is the fraction of the period during which the waveform is\n        greater than zero.\n\n        ";
static PyMethodDef __pyx_mdef_6_pyfxr_9Wavetable_11square = {"square", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_11square, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_10square};
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_11square(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  float __pyx_v_duty_cycle;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("square (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_duty_cycle,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_duty_cycle);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "square") < 0)) __PYX_ERR(0, 96, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_duty_cycle = __pyx_PyFloat_AsFloat(values[0]); if (unlikely((__pyx_v_duty_cycle == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 96, __pyx_L3_error)
    } else {
      __pyx_v_duty_cycle = ((float)0.5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("square", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 96, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.Wavetable.square", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6_pyfxr_9Wavetable_10square(__pyx_v_duty_cycle);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_9Wavetable_10square(float __pyx_v_duty_cycle) {
  struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_w = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_split;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("square", 0);

  /* "_pyfxr.pyx":106
 *         cdef size_t i, split
 *         cdef float v
 *         w = Wavetable.__new__(Wavetable)             # <<<<<<<<<<<<<<
 * 
 *         if not (0.0 < duty_cycle < 1.0):
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6_pyfxr_Wavetable(((PyTypeObject *)__pyx_ptype_6_pyfxr_Wavetable), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_w = ((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "_pyfxr.pyx":108
 *         w = Wavetable.__new__(Wavetable)
 * 
 *         if not (0.0 < duty_cycle < 1.0):             # <<<<<<<<<<<<<<
 *             raise ValueError("duty_cycle must be between 0 and 1")
 * 
 */
  __pyx_t_2 = (0.0 < __pyx_v_duty_cycle);
  if (__pyx_t_2) {
    __pyx_t_2 = (__pyx_v_duty_cycle < 1.0);
  }
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "_pyfxr.pyx":109
 * 
 *         if not (0.0 < duty_cycle < 1.0):
 *             raise ValueError("duty_cycle must be between 0 and 1")             # <<<<<<<<<<<<<<
 * 
 *         split = round(duty_cycle * 1024)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 109, __pyx_L1_error)

    /* "_pyfxr.pyx":108
 *         w = Wavetable.__new__(Wavetable)
 * 
 *         if not (0.0 < duty_cycle < 1.0):             # <<<<<<<<<<<<<<
 *             raise ValueError("duty_cycle must be between 0 and 1")
 * 
 */
  }

  /* "_pyfxr.pyx":111
 *             raise ValueError("duty_cycle must be between 0 and 1")
 * 
 *         split = round(duty_cycle * 1024)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             for i in range(split):
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_duty_cycle * 1024.0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_round, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_split = __pyx_t_5;

  /* "_pyfxr.pyx":112
 * 
 *         split = round(duty_cycle * 1024)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(split):
 *                 w.wavetable[i] = 32767
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "_pyfxr.pyx":113
 *         split = round(duty_cycle * 1024)
 *         with nogil:
 *             for i in range(split):             # <<<<<<<<<<<<<<
 *                 w.wavetable[i] = 32767
 *             for i in range(split, 1024):
 */
        __pyx_t_5 = __pyx_v_split;
        __pyx_t_6 = __pyx_t_5;
        for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
          __pyx_v_i = __pyx_t_7;

          /* "_pyfxr.pyx":114
 *         with nogil:
 *             for i in range(split):
 *                 w.wavetable[i] = 32767             # <<<<<<<<<<<<<<
 *             for i in range(split, 1024):
 *                 w.wavetable[i] = -32768
 */
          (__pyx_v_w->wavetable[__pyx_v_i]) = 0x7FFF;
        }

        /* "_pyfxr.pyx":115
 *             for i in range(split):
 *                 w.wavetable[i] = 32767
 *             for i in range(split, 1024):             # <<<<<<<<<<<<<<
 *                 w.wavetable[i] = -32768
 *         return w
 */
        for (__pyx_t_5 = __pyx_v_split; __pyx_t_5 < 0x400; __pyx_t_5+=1) {
          __pyx_v_i = __pyx_t_5;

          /* "_pyfxr.pyx":116
 *                 w.wavetable[i] = 32767
 *             for i in range(split, 1024):
 *                 w.wavetable[i] = -32768             # <<<<<<<<<<<<<<
 *         return w
 * 
 */
          (__pyx_v_w->wavetable[__pyx_v_i]) = -32768;
        }
      }

      /* "_pyfxr.pyx":112
 * 
 *         split = round(duty_cycle * 1024)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for i in range(split):
 *                 w.wavetable[i] = 32767
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "_pyfxr.pyx":117
 *             for i in range(split, 1024):
 *                 w.wavetable[i] = -32768
 *         return w             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_w));
  __pyx_r = ((PyObject *)__pyx_v_w);
  goto __pyx_L0;

  /* "_pyfxr.pyx":96
 * 
 *     @staticmethod
 *     def square(float duty_cycle=0.5):             # <<<<<<<<<<<<<<
 *         """Generate a square-wave waveform.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("_pyfxr.Wavetable.square", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_w);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":119
 *         return w
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t itemsize = sizeof(int16_t)
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_6_pyfxr_9Wavetable_13__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_6_pyfxr_9Wavetable_13__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_9Wavetable_12__getbuffer__(((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6_pyfxr_9Wavetable_12__getbuffer__(struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int16_t *__pyx_t_1;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "_pyfxr.pyx":120
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 *         cdef Py_ssize_t itemsize = sizeof(int16_t)             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = self.wavetable
 */
  __pyx_v_itemsize = (sizeof(int16_t));

  /* "_pyfxr.pyx":122
 *         cdef Py_ssize_t itemsize = sizeof(int16_t)
 * 
 *         buffer.buf = self.wavetable             # <<<<<<<<<<<<<<
 *         buffer.format = 'h'                     # double
 *         buffer.internal = NULL                  # see References
 */
  __pyx_t_1 = __pyx_v_self->wavetable;
  __pyx_v_buffer->buf = __pyx_t_1;

  /* "_pyfxr.pyx":123
 * 
 *         buffer.buf = self.wavetable
 *         buffer.format = 'h'                     # double             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"h");

  /* "_pyfxr.pyx":124
 *         buffer.buf = self.wavetable
 *         buffer.format = 'h'                     # double
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = sizeof(self.wavetable)
 */
  __pyx_v_buffer->internal = NULL;

  /* "_pyfxr.pyx":125
 *         buffer.format = 'h'                     # double
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = sizeof(self.wavetable)
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "_pyfxr.pyx":126
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = sizeof(self.wavetable)             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_v_buffer->len = (sizeof(__pyx_v_self->wavetable));

  /* "_pyfxr.pyx":127
 *         buffer.itemsize = itemsize
 *         buffer.len = sizeof(self.wavetable)
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 0
 */
  __pyx_v_buffer->ndim = 1;

  /* "_pyfxr.pyx":128
 *         buffer.len = sizeof(self.wavetable)
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 0
 *         buffer.shape = NULL
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "_pyfxr.pyx":129
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 0             # <<<<<<<<<<<<<<
 *         buffer.shape = NULL
 *         buffer.strides = NULL
 */
  __pyx_v_buffer->readonly = 0;

  /* "_pyfxr.pyx":130
 *         buffer.obj = self
 *         buffer.readonly = 0
 *         buffer.shape = NULL             # <<<<<<<<<<<<<<
 *         buffer.strides = NULL
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_v_buffer->shape = NULL;

  /* "_pyfxr.pyx":131
 *         buffer.readonly = 0
 *         buffer.shape = NULL
 *         buffer.strides = NULL             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_v_buffer->strides = NULL;

  /* "_pyfxr.pyx":132
 *         buffer.shape = NULL
 *         buffer.strides = NULL
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "_pyfxr.pyx":119
 *         return w
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t itemsize = sizeof(int16_t)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":134
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_6_pyfxr_9Wavetable_15__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_6_pyfxr_9Wavetable_15__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_6_pyfxr_9Wavetable_14__releasebuffer__(((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6_pyfxr_9Wavetable_14__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6_pyfxr_9Wavetable_16__reduce_cython__[] = "Wavetable.__reduce_cython__(self)";
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_9Wavetable_16__reduce_cython__(((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_9Wavetable_16__reduce_cython__(struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.wavetable,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_carray_to_py_int16_t(__pyx_v_self->wavetable, 0x400); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.wavetable,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.wavetable,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.wavetable,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, None), state
 *     else:
 */
  __pyx_t_4 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Wavetable); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_116404587);
    __Pyx_GIVEREF(__pyx_int_116404587);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_116404587);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, None), state
 *     else:
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Wavetable__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Wavetable); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_116404587);
    __Pyx_GIVEREF(__pyx_int_116404587);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_116404587);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_pyfxr.Wavetable.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Wavetable__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6_pyfxr_9Wavetable_18__setstate_cython__[] = "Wavetable.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6_pyfxr_9Wavetable_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_9Wavetable_18__setstate_cython__(((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_9Wavetable_18__setstate_cython__(struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Wavetable__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6_pyfxr___pyx_unpickle_Wavetable__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Wavetable, (type(self), 0x6f0316b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Wavetable__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.Wavetable.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":145
 *     channels: int = 1
 * 
 *     def __cinit__(self, size_t n_samples):             # <<<<<<<<<<<<<<
 *         self.samples = <int16_t*> PyMem_Malloc(n_samples * sizeof(int16_t))
 *         self.n_samples = n_samples
 */

/* Python wrapper */
static int __pyx_pw_6_pyfxr_11SoundBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6_pyfxr_11SoundBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_n_samples;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n_samples,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_samples)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 145, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_n_samples = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_n_samples == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 145, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.SoundBuffer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6_pyfxr_11SoundBuffer___cinit__(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self), __pyx_v_n_samples);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6_pyfxr_11SoundBuffer___cinit__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, size_t __pyx_v_n_samples) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "_pyfxr.pyx":146
 * 
 *     def __cinit__(self, size_t n_samples):
 *         self.samples = <int16_t*> PyMem_Malloc(n_samples * sizeof(int16_t))             # <<<<<<<<<<<<<<
 *         self.n_samples = n_samples
 *         if not self.samples:
 */
  __pyx_v_self->samples = ((int16_t *)PyMem_Malloc((__pyx_v_n_samples * (sizeof(int16_t)))));

  /* "_pyfxr.pyx":147
 *     def __cinit__(self, size_t n_samples):
 *         self.samples = <int16_t*> PyMem_Malloc(n_samples * sizeof(int16_t))
 *         self.n_samples = n_samples             # <<<<<<<<<<<<<<
 *         if not self.samples:
 *             raise MemoryError()
 */
  __pyx_v_self->n_samples = __pyx_v_n_samples;

  /* "_pyfxr.pyx":148
 *         self.samples = <int16_t*> PyMem_Malloc(n_samples * sizeof(int16_t))
 *         self.n_samples = n_samples
 *         if not self.samples:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         for i in range(n_samples):
 */
  __pyx_t_1 = ((!(__pyx_v_self->samples != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "_pyfxr.pyx":149
 *         self.n_samples = n_samples
 *         if not self.samples:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         for i in range(n_samples):
 *             self.samples[i] = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 149, __pyx_L1_error)

    /* "_pyfxr.pyx":148
 *         self.samples = <int16_t*> PyMem_Malloc(n_samples * sizeof(int16_t))
 *         self.n_samples = n_samples
 *         if not self.samples:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         for i in range(n_samples):
 */
  }

  /* "_pyfxr.pyx":150
 *         if not self.samples:
 *             raise MemoryError()
 *         for i in range(n_samples):             # <<<<<<<<<<<<<<
 *             self.samples[i] = 0
 * 
 */
  __pyx_t_2 = __pyx_v_n_samples;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "_pyfxr.pyx":151
 *             raise MemoryError()
 *         for i in range(n_samples):
 *             self.samples[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    (__pyx_v_self->samples[__pyx_v_i]) = 0;
  }

  /* "_pyfxr.pyx":145
 *     channels: int = 1
 * 
 *     def __cinit__(self, size_t n_samples):             # <<<<<<<<<<<<<<
 *         self.samples = <int16_t*> PyMem_Malloc(n_samples * sizeof(int16_t))
 *         self.n_samples = n_samples
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("_pyfxr.SoundBuffer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":153
 *             self.samples[i] = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.samples)
 * 
 */

/* Python wrapper */
static void __pyx_pw_6_pyfxr_11SoundBuffer_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6_pyfxr_11SoundBuffer_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6_pyfxr_11SoundBuffer_2__dealloc__(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6_pyfxr_11SoundBuffer_2__dealloc__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "_pyfxr.pyx":154
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self.samples)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  PyMem_Free(__pyx_v_self->samples);

  /* "_pyfxr.pyx":153
 *             self.samples[i] = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.samples)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "_pyfxr.pyx":156
 *         PyMem_Free(self.samples)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.n_samples
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6_pyfxr_11SoundBuffer_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6_pyfxr_11SoundBuffer_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11SoundBuffer_4__len__(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6_pyfxr_11SoundBuffer_4__len__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "_pyfxr.pyx":157
 * 
 *     def __len__(self):
 *         return self.n_samples             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, ssize_t i):
 */
  __pyx_r = __pyx_v_self->n_samples;
  goto __pyx_L0;

  /* "_pyfxr.pyx":156
 *         PyMem_Free(self.samples)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.n_samples
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":159
 *         return self.n_samples
 * 
 *     def __getitem__(self, ssize_t i):             # <<<<<<<<<<<<<<
 *         if i >= 0:
 *             if i >= self.n_samples:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  Py_ssize_t __pyx_v_i;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = PyInt_AsSsize_t(__pyx_arg_i); if (unlikely((__pyx_v_i == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 159, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.SoundBuffer.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6_pyfxr_11SoundBuffer_6__getitem__(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self), ((Py_ssize_t)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_6__getitem__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, Py_ssize_t __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "_pyfxr.pyx":160
 * 
 *     def __getitem__(self, ssize_t i):
 *         if i >= 0:             # <<<<<<<<<<<<<<
 *             if i >= self.n_samples:
 *                 raise IndexError("index out of range")
 */
  __pyx_t_1 = ((__pyx_v_i >= 0) != 0);
  if (__pyx_t_1) {

    /* "_pyfxr.pyx":161
 *     def __getitem__(self, ssize_t i):
 *         if i >= 0:
 *             if i >= self.n_samples:             # <<<<<<<<<<<<<<
 *                 raise IndexError("index out of range")
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_i >= __pyx_v_self->n_samples) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "_pyfxr.pyx":162
 *         if i >= 0:
 *             if i >= self.n_samples:
 *                 raise IndexError("index out of range")             # <<<<<<<<<<<<<<
 *         else:
 *             i = self.n_samples + i
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 162, __pyx_L1_error)

      /* "_pyfxr.pyx":161
 *     def __getitem__(self, ssize_t i):
 *         if i >= 0:
 *             if i >= self.n_samples:             # <<<<<<<<<<<<<<
 *                 raise IndexError("index out of range")
 *         else:
 */
    }

    /* "_pyfxr.pyx":160
 * 
 *     def __getitem__(self, ssize_t i):
 *         if i >= 0:             # <<<<<<<<<<<<<<
 *             if i >= self.n_samples:
 *                 raise IndexError("index out of range")
 */
    goto __pyx_L3;
  }

  /* "_pyfxr.pyx":164
 *                 raise IndexError("index out of range")
 *         else:
 *             i = self.n_samples + i             # <<<<<<<<<<<<<<
 *             if i < 0:
 *                 raise IndexError("index out of range")
 */
  /*else*/ {
    __pyx_v_i = (__pyx_v_self->n_samples + __pyx_v_i);

    /* "_pyfxr.pyx":165
 *         else:
 *             i = self.n_samples + i
 *             if i < 0:             # <<<<<<<<<<<<<<
 *                 raise IndexError("index out of range")
 *         return self.samples[i]
 */
    __pyx_t_1 = ((__pyx_v_i < 0) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "_pyfxr.pyx":166
 *             i = self.n_samples + i
 *             if i < 0:
 *                 raise IndexError("index out of range")             # <<<<<<<<<<<<<<
 *         return self.samples[i]
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 166, __pyx_L1_error)

      /* "_pyfxr.pyx":165
 *         else:
 *             i = self.n_samples + i
 *             if i < 0:             # <<<<<<<<<<<<<<
 *                 raise IndexError("index out of range")
 *         return self.samples[i]
 */
    }
  }
  __pyx_L3:;

  /* "_pyfxr.pyx":167
 *             if i < 0:
 *                 raise IndexError("index out of range")
 *         return self.samples[i]             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int16_t((__pyx_v_self->samples[__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_pyfxr.pyx":159
 *         return self.n_samples
 * 
 *     def __getitem__(self, ssize_t i):             # <<<<<<<<<<<<<<
 *         if i >= 0:
 *             if i >= self.n_samples:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("_pyfxr.SoundBuffer.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":170
 * 
 *     @property
 *     def duration(SoundBuffer self) -> float:             # <<<<<<<<<<<<<<
 *         """Get the duration of this sound in seconds, as a float."""
 *         return self.n_samples / <float> SAMPLE_RATE
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_8duration_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_8duration_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11SoundBuffer_8duration___get__(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_8duration___get__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "_pyfxr.pyx":172
 *     def duration(SoundBuffer self) -> float:
 *         """Get the duration of this sound in seconds, as a float."""
 *         return self.n_samples / <float> SAMPLE_RATE             # <<<<<<<<<<<<<<
 * 
 *     def save(self, filename: str):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(((float)__pyx_v_6_pyfxr_SAMPLE_RATE) == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 172, __pyx_L1_error)
  }
  __pyx_t_1 = PyFloat_FromDouble((((float)__pyx_v_self->n_samples) / ((float)__pyx_v_6_pyfxr_SAMPLE_RATE))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "_pyfxr.pyx":170
 * 
 *     @property
 *     def duration(SoundBuffer self) -> float:             # <<<<<<<<<<<<<<
 *         """Get the duration of this sound in seconds, as a float."""
 *         return self.n_samples / <float> SAMPLE_RATE
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.SoundBuffer.duration.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":174
 *         return self.n_samples / <float> SAMPLE_RATE
 * 
 *     def save(self, filename: str):             # <<<<<<<<<<<<<<
 *         """Save this sound to a .wav file."""
 *         import wave
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_9save(PyObject *__pyx_v_self, PyObject *__pyx_v_filename); /*proto*/
static char __pyx_doc_6_pyfxr_11SoundBuffer_8save[] = "SoundBuffer.save(self, unicode filename: str)\nSave this sound to a .wav file.";
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_9save(PyObject *__pyx_v_self, PyObject *__pyx_v_filename) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("save (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_filename), (&PyUnicode_Type), 1, "filename", 1))) __PYX_ERR(0, 174, __pyx_L1_error)
  __pyx_r = __pyx_pf_6_pyfxr_11SoundBuffer_8save(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self), ((PyObject*)__pyx_v_filename));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_8save(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_wave = NULL;
  PyObject *__pyx_v_wav = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("save", 0);

  /* "_pyfxr.pyx":176
 *     def save(self, filename: str):
 *         """Save this sound to a .wav file."""
 *         import wave             # <<<<<<<<<<<<<<
 * 
 *         with wave.open(filename, 'wb') as wav:
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_wave, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wave = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pyfxr.pyx":178
 *         import wave
 * 
 *         with wave.open(filename, 'wb') as wav:             # <<<<<<<<<<<<<<
 *             wav.setframerate(SAMPLE_RATE)
 *             wav.setnchannels(1)
 */
  /*with:*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_wave, __pyx_n_s_open); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_filename, __pyx_n_u_wb};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_filename, __pyx_n_u_wb};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 178, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_filename);
      __Pyx_GIVEREF(__pyx_v_filename);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_filename);
      __Pyx_INCREF(__pyx_n_u_wb);
      __Pyx_GIVEREF(__pyx_n_u_wb);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_n_u_wb);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 178, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {
          __pyx_v_wav = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "_pyfxr.pyx":179
 * 
 *         with wave.open(filename, 'wb') as wav:
 *             wav.setframerate(SAMPLE_RATE)             # <<<<<<<<<<<<<<
 *             wav.setnchannels(1)
 *             wav.setnframes(self.n_samples)
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_wav, __pyx_n_s_setframerate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_6_pyfxr_SAMPLE_RATE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 179, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "_pyfxr.pyx":180
 *         with wave.open(filename, 'wb') as wav:
 *             wav.setframerate(SAMPLE_RATE)
 *             wav.setnchannels(1)             # <<<<<<<<<<<<<<
 *             wav.setnframes(self.n_samples)
 *             wav.setsampwidth(2)
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_wav, __pyx_n_s_setnchannels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_int_1) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_int_1);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 180, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "_pyfxr.pyx":181
 *             wav.setframerate(SAMPLE_RATE)
 *             wav.setnchannels(1)
 *             wav.setnframes(self.n_samples)             # <<<<<<<<<<<<<<
 *             wav.setsampwidth(2)
 *             wav.writeframesraw(self)
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_wav, __pyx_n_s_setnframes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->n_samples); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 181, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "_pyfxr.pyx":182
 *             wav.setnchannels(1)
 *             wav.setnframes(self.n_samples)
 *             wav.setsampwidth(2)             # <<<<<<<<<<<<<<
 *             wav.writeframesraw(self)
 * 
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_wav, __pyx_n_s_setsampwidth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_int_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_int_2);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 182, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "_pyfxr.pyx":183
 *             wav.setnframes(self.n_samples)
 *             wav.setsampwidth(2)
 *             wav.writeframesraw(self)             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_wav, __pyx_n_s_writeframesraw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, ((PyObject *)__pyx_v_self)) : __Pyx_PyObject_CallOneArg(__pyx_t_1, ((PyObject *)__pyx_v_self));
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 183, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "_pyfxr.pyx":178
 *         import wave
 * 
 *         with wave.open(filename, 'wb') as wav:             # <<<<<<<<<<<<<<
 *             wav.setframerate(SAMPLE_RATE)
 *             wav.setnchannels(1)
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("_pyfxr.SoundBuffer.save", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 178, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = PyTuple_Pack(3, __pyx_t_5, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, NULL);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 178, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__pyx_t_11 < 0) __PYX_ERR(0, 178, __pyx_L9_except_error)
          __pyx_t_12 = ((!(__pyx_t_11 != 0)) != 0);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_XGIVEREF(__pyx_t_2);
            __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_1, __pyx_t_2);
            __pyx_t_5 = 0; __pyx_t_1 = 0; __pyx_t_2 = 0; 
            __PYX_ERR(0, 178, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_6) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__4, NULL);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 178, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L16;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L1_error;
    __pyx_L16:;
  }

  /* "_pyfxr.pyx":174
 *         return self.n_samples / <float> SAMPLE_RATE
 * 
 *     def save(self, filename: str):             # <<<<<<<<<<<<<<
 *         """Save this sound to a .wav file."""
 *         import wave
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_pyfxr.SoundBuffer.save", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wave);
  __Pyx_XDECREF(__pyx_v_wav);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":185
 *             wav.writeframesraw(self)
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t itemsize = sizeof(int16_t)
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_6_pyfxr_11SoundBuffer_11__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_6_pyfxr_11SoundBuffer_11__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11SoundBuffer_10__getbuffer__(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6_pyfxr_11SoundBuffer_10__getbuffer__(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int16_t *__pyx_t_1;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "_pyfxr.pyx":186
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 *         cdef Py_ssize_t itemsize = sizeof(int16_t)             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = self.samples
 */
  __pyx_v_itemsize = (sizeof(int16_t));

  /* "_pyfxr.pyx":188
 *         cdef Py_ssize_t itemsize = sizeof(int16_t)
 * 
 *         buffer.buf = self.samples             # <<<<<<<<<<<<<<
 *         buffer.format = 'h'                     # double
 *         buffer.internal = NULL                  # see References
 */
  __pyx_t_1 = __pyx_v_self->samples;
  __pyx_v_buffer->buf = __pyx_t_1;

  /* "_pyfxr.pyx":189
 * 
 *         buffer.buf = self.samples
 *         buffer.format = 'h'                     # double             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"h");

  /* "_pyfxr.pyx":190
 *         buffer.buf = self.samples
 *         buffer.format = 'h'                     # double
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = sizeof(int16_t) * self.n_samples
 */
  __pyx_v_buffer->internal = NULL;

  /* "_pyfxr.pyx":191
 *         buffer.format = 'h'                     # double
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = sizeof(int16_t) * self.n_samples
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "_pyfxr.pyx":192
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = sizeof(int16_t) * self.n_samples             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_v_buffer->len = ((sizeof(int16_t)) * __pyx_v_self->n_samples);

  /* "_pyfxr.pyx":193
 *         buffer.itemsize = itemsize
 *         buffer.len = sizeof(int16_t) * self.n_samples
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 0
 */
  __pyx_v_buffer->ndim = 1;

  /* "_pyfxr.pyx":194
 *         buffer.len = sizeof(int16_t) * self.n_samples
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 0
 *         buffer.shape = NULL
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "_pyfxr.pyx":195
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 0             # <<<<<<<<<<<<<<
 *         buffer.shape = NULL
 *         buffer.strides = NULL
 */
  __pyx_v_buffer->readonly = 0;

  /* "_pyfxr.pyx":196
 *         buffer.obj = self
 *         buffer.readonly = 0
 *         buffer.shape = NULL             # <<<<<<<<<<<<<<
 *         buffer.strides = NULL
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_v_buffer->shape = NULL;

  /* "_pyfxr.pyx":197
 *         buffer.readonly = 0
 *         buffer.shape = NULL
 *         buffer.strides = NULL             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_v_buffer->strides = NULL;

  /* "_pyfxr.pyx":198
 *         buffer.shape = NULL
 *         buffer.strides = NULL
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "_pyfxr.pyx":185
 *             wav.writeframesraw(self)
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t itemsize = sizeof(int16_t)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":200
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_6_pyfxr_11SoundBuffer_13__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_6_pyfxr_11SoundBuffer_13__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_6_pyfxr_11SoundBuffer_12__releasebuffer__(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6_pyfxr_11SoundBuffer_12__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "_pyfxr.pyx":203
 *         pass
 * 
 *     def get_queue_source(self):             # <<<<<<<<<<<<<<
 *         """Duck type as a pyglet.media.Source."""
 *         return PygletSource(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_15get_queue_source(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6_pyfxr_11SoundBuffer_14get_queue_source[] = "SoundBuffer.get_queue_source(self)\nDuck type as a pyglet.media.Source.";
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_15get_queue_source(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_queue_source (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11SoundBuffer_14get_queue_source(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_14get_queue_source(struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_queue_source", 0);

  /* "_pyfxr.pyx":205
 *     def get_queue_source(self):
 *         """Duck type as a pyglet.media.Source."""
 *         return PygletSource(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_6_pyfxr_PygletSource), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "_pyfxr.pyx":203
 *         pass
 * 
 *     def get_queue_source(self):             # <<<<<<<<<<<<<<
 *         """Duck type as a pyglet.media.Source."""
 *         return PygletSource(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.SoundBuffer.get_queue_source", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6_pyfxr_11SoundBuffer_16__reduce_cython__[] = "SoundBuffer.__reduce_cython__(self)";
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11SoundBuffer_16__reduce_cython__(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.SoundBuffer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6_pyfxr_11SoundBuffer_18__setstate_cython__[] = "SoundBuffer.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6_pyfxr_11SoundBuffer_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11SoundBuffer_18__setstate_cython__(((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11SoundBuffer_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.SoundBuffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":211
 *     cdef SoundBuffer buf
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.buf = None
 * 
 */

/* Python wrapper */
static int __pyx_pw_6_pyfxr_11CachedSound_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6_pyfxr_11CachedSound_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6_pyfxr_11CachedSound___cinit__(((struct __pyx_obj_6_pyfxr_CachedSound *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6_pyfxr_11CachedSound___cinit__(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "_pyfxr.pyx":212
 * 
 *     def __cinit__(self):
 *         self.buf = None             # <<<<<<<<<<<<<<
 * 
 *     def _clear(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->buf);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->buf));
  __pyx_v_self->buf = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)Py_None);

  /* "_pyfxr.pyx":211
 *     cdef SoundBuffer buf
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.buf = None
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":214
 *         self.buf = None
 * 
 *     def _clear(self):             # <<<<<<<<<<<<<<
 *         self.buf = None
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_3_clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6_pyfxr_11CachedSound_2_clear[] = "CachedSound._clear(self)";
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_3_clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_clear (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11CachedSound_2_clear(((struct __pyx_obj_6_pyfxr_CachedSound *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11CachedSound_2_clear(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_clear", 0);

  /* "_pyfxr.pyx":215
 * 
 *     def _clear(self):
 *         self.buf = None             # <<<<<<<<<<<<<<
 * 
 *     def _set(self, SoundBuffer sound):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->buf);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->buf));
  __pyx_v_self->buf = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)Py_None);

  /* "_pyfxr.pyx":214
 *         self.buf = None
 * 
 *     def _clear(self):             # <<<<<<<<<<<<<<
 *         self.buf = None
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":217
 *         self.buf = None
 * 
 *     def _set(self, SoundBuffer sound):             # <<<<<<<<<<<<<<
 *         self.buf = sound
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_5_set(PyObject *__pyx_v_self, PyObject *__pyx_v_sound); /*proto*/
static char __pyx_doc_6_pyfxr_11CachedSound_4_set[] = "CachedSound._set(self, SoundBuffer sound)";
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_5_set(PyObject *__pyx_v_self, PyObject *__pyx_v_sound) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sound), __pyx_ptype_6_pyfxr_SoundBuffer, 1, "sound", 0))) __PYX_ERR(0, 217, __pyx_L1_error)
  __pyx_r = __pyx_pf_6_pyfxr_11CachedSound_4_set(((struct __pyx_obj_6_pyfxr_CachedSound *)__pyx_v_self), ((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_v_sound));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11CachedSound_4_set(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self, struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_sound) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set", 0);

  /* "_pyfxr.pyx":218
 * 
 *     def _set(self, SoundBuffer sound):
 *         self.buf = sound             # <<<<<<<<<<<<<<
 * 
 *     def _get(self):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_sound));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_sound));
  __Pyx_GOTREF(__pyx_v_self->buf);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->buf));
  __pyx_v_self->buf = __pyx_v_sound;

  /* "_pyfxr.pyx":217
 *         self.buf = None
 * 
 *     def _set(self, SoundBuffer sound):             # <<<<<<<<<<<<<<
 *         self.buf = sound
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":220
 *         self.buf = sound
 * 
 *     def _get(self):             # <<<<<<<<<<<<<<
 *         if self.buf is None:
 *             self.buf = <SoundBuffer?> self._build()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_7_get(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6_pyfxr_11CachedSound_6_get[] = "CachedSound._get(self)";
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_7_get(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11CachedSound_6_get(((struct __pyx_obj_6_pyfxr_CachedSound *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11CachedSound_6_get(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get", 0);

  /* "_pyfxr.pyx":221
 * 
 *     def _get(self):
 *         if self.buf is None:             # <<<<<<<<<<<<<<
 *             self.buf = <SoundBuffer?> self._build()
 *         return self.buf
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->buf) == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "_pyfxr.pyx":222
 *     def _get(self):
 *         if self.buf is None:
 *             self.buf = <SoundBuffer?> self._build()             # <<<<<<<<<<<<<<
 *         return self.buf
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_build); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6_pyfxr_SoundBuffer)))) __PYX_ERR(0, 222, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->buf);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->buf));
    __pyx_v_self->buf = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "_pyfxr.pyx":221
 * 
 *     def _get(self):
 *         if self.buf is None:             # <<<<<<<<<<<<<<
 *             self.buf = <SoundBuffer?> self._build()
 *         return self.buf
 */
  }

  /* "_pyfxr.pyx":223
 *         if self.buf is None:
 *             self.buf = <SoundBuffer?> self._build()
 *         return self.buf             # <<<<<<<<<<<<<<
 * 
 *     def _build(self) -> SoundBuffer:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->buf));
  __pyx_r = ((PyObject *)__pyx_v_self->buf);
  goto __pyx_L0;

  /* "_pyfxr.pyx":220
 *         self.buf = sound
 * 
 *     def _get(self):             # <<<<<<<<<<<<<<
 *         if self.buf is None:
 *             self.buf = <SoundBuffer?> self._build()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_pyfxr.CachedSound._get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":225
 *         return self.buf
 * 
 *     def _build(self) -> SoundBuffer:             # <<<<<<<<<<<<<<
 *         """Override this to build the sound.
 * 
 */

/* Python wrapper */
static struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_pw_6_pyfxr_11CachedSound_9_build(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6_pyfxr_11CachedSound_8_build[] = "CachedSound._build(self) -> SoundBuffer\nOverride this to build the sound.\n\n        Must return a SoundBuffer.\n        ";
static struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_pw_6_pyfxr_11CachedSound_9_build(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_build (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11CachedSound_8_build(((struct __pyx_obj_6_pyfxr_CachedSound *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_pf_6_pyfxr_11CachedSound_8_build(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self) {
  struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_build", 0);

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)Py_None); __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":231
 *         """
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 *         """Delegate to our cached sound.
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_6_pyfxr_11CachedSound_11__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_6_pyfxr_11CachedSound_10__getbuffer__;
#endif
static CYTHON_UNUSED int __pyx_pw_6_pyfxr_11CachedSound_11__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11CachedSound_10__getbuffer__(((struct __pyx_obj_6_pyfxr_CachedSound *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6_pyfxr_11CachedSound_10__getbuffer__(struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "_pyfxr.pyx":237
 * 
 *         """
 *         PyObject_GetBuffer(self._get(), buffer, flags)             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = PyObject_GetBuffer(__pyx_t_1, __pyx_v_buffer, __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_pyfxr.pyx":231
 *         """
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 *         """Delegate to our cached sound.
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_pyfxr.CachedSound.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_buffer->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":239
 *         PyObject_GetBuffer(self._get(), buffer, flags)
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_6_pyfxr_11CachedSound_13__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_6_pyfxr_11CachedSound_13__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_6_pyfxr_11CachedSound_12__releasebuffer__(((struct __pyx_obj_6_pyfxr_CachedSound *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6_pyfxr_11CachedSound_12__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6_pyfxr_11CachedSound_14__reduce_cython__[] = "CachedSound.__reduce_cython__(self)";
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11CachedSound_14__reduce_cython__(((struct __pyx_obj_6_pyfxr_CachedSound *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11CachedSound_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.CachedSound.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6_pyfxr_11CachedSound_16__setstate_cython__[] = "CachedSound.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6_pyfxr_11CachedSound_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_11CachedSound_16__setstate_cython__(((struct __pyx_obj_6_pyfxr_CachedSound *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_11CachedSound_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6_pyfxr_CachedSound *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.CachedSound.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":247
 *     cdef size_t pos
 * 
 *     def __init__(self, SoundBuffer buf):             # <<<<<<<<<<<<<<
 *         self.buf = buf
 *         self.pos = 0
 */

/* Python wrapper */
static int __pyx_pw_6_pyfxr_12PygletSource_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6_pyfxr_12PygletSource_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_buf = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_buf,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buf)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 247, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_buf = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 247, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.PygletSource.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_buf), __pyx_ptype_6_pyfxr_SoundBuffer, 1, "buf", 0))) __PYX_ERR(0, 247, __pyx_L1_error)
  __pyx_r = __pyx_pf_6_pyfxr_12PygletSource___init__(((struct __pyx_obj_6_pyfxr_PygletSource *)__pyx_v_self), __pyx_v_buf);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6_pyfxr_12PygletSource___init__(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self, struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_buf) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "_pyfxr.pyx":248
 * 
 *     def __init__(self, SoundBuffer buf):
 *         self.buf = buf             # <<<<<<<<<<<<<<
 *         self.pos = 0
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_buf));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_buf));
  __Pyx_GOTREF(__pyx_v_self->buf);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->buf));
  __pyx_v_self->buf = __pyx_v_buf;

  /* "_pyfxr.pyx":249
 *     def __init__(self, SoundBuffer buf):
 *         self.buf = buf
 *         self.pos = 0             # <<<<<<<<<<<<<<
 * 
 *     video_format = None
 */
  __pyx_v_self->pos = 0;

  /* "_pyfxr.pyx":247
 *     cdef size_t pos
 * 
 *     def __init__(self, SoundBuffer buf):             # <<<<<<<<<<<<<<
 *         self.buf = buf
 *         self.pos = 0
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":254
 * 
 *     @property
 *     def audio_format(self):             # <<<<<<<<<<<<<<
 *         from pyglet.media.codecs import AudioFormat
 *         return AudioFormat(
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_12PygletSource_12audio_format_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6_pyfxr_12PygletSource_12audio_format_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_12PygletSource_12audio_format___get__(((struct __pyx_obj_6_pyfxr_PygletSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_12PygletSource_12audio_format___get__(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self) {
  PyObject *__pyx_v_AudioFormat = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "_pyfxr.pyx":255
 *     @property
 *     def audio_format(self):
 *         from pyglet.media.codecs import AudioFormat             # <<<<<<<<<<<<<<
 *         return AudioFormat(
 *             channels=1,
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_AudioFormat);
  __Pyx_GIVEREF(__pyx_n_s_AudioFormat);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_AudioFormat);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyglet_media_codecs, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_AudioFormat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_AudioFormat = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pyfxr.pyx":256
 *     def audio_format(self):
 *         from pyglet.media.codecs import AudioFormat
 *         return AudioFormat(             # <<<<<<<<<<<<<<
 *             channels=1,
 *             sample_size=16,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "_pyfxr.pyx":257
 *         from pyglet.media.codecs import AudioFormat
 *         return AudioFormat(
 *             channels=1,             # <<<<<<<<<<<<<<
 *             sample_size=16,
 *             sample_rate=self.buf.sample_rate
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_channels, __pyx_int_1) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_sample_size, __pyx_int_16) < 0) __PYX_ERR(0, 257, __pyx_L1_error)

  /* "_pyfxr.pyx":259
 *             channels=1,
 *             sample_size=16,
 *             sample_rate=self.buf.sample_rate             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->buf), __pyx_n_s_sample_rate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_sample_rate, __pyx_t_1) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_pyfxr.pyx":256
 *     def audio_format(self):
 *         from pyglet.media.codecs import AudioFormat
 *         return AudioFormat(             # <<<<<<<<<<<<<<
 *             channels=1,
 *             sample_size=16,
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_AudioFormat, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "_pyfxr.pyx":254
 * 
 *     @property
 *     def audio_format(self):             # <<<<<<<<<<<<<<
 *         from pyglet.media.codecs import AudioFormat
 *         return AudioFormat(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("_pyfxr.PygletSource.audio_format.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_AudioFormat);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":262
 *         )
 * 
 *     def get_audio_data(self, length):             # <<<<<<<<<<<<<<
 *         from pyglet.media.codecs import AudioData
 *         if self.pos != 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_12PygletSource_3get_audio_data(PyObject *__pyx_v_self, PyObject *__pyx_v_length); /*proto*/
static char __pyx_doc_6_pyfxr_12PygletSource_2get_audio_data[] = "PygletSource.get_audio_data(self, length)";
static PyObject *__pyx_pw_6_pyfxr_12PygletSource_3get_audio_data(PyObject *__pyx_v_self, PyObject *__pyx_v_length) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_audio_data (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_12PygletSource_2get_audio_data(((struct __pyx_obj_6_pyfxr_PygletSource *)__pyx_v_self), ((PyObject *)__pyx_v_length));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_12PygletSource_2get_audio_data(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_length) {
  PyObject *__pyx_v_AudioData = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_audio_data", 0);

  /* "_pyfxr.pyx":263
 * 
 *     def get_audio_data(self, length):
 *         from pyglet.media.codecs import AudioData             # <<<<<<<<<<<<<<
 *         if self.pos != 0:
 *             return None
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_AudioData);
  __Pyx_GIVEREF(__pyx_n_s_AudioData);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_AudioData);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyglet_media_codecs, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_AudioData); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_AudioData = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pyfxr.pyx":264
 *     def get_audio_data(self, length):
 *         from pyglet.media.codecs import AudioData
 *         if self.pos != 0:             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  __pyx_t_3 = ((__pyx_v_self->pos != 0) != 0);
  if (__pyx_t_3) {

    /* "_pyfxr.pyx":265
 *         from pyglet.media.codecs import AudioData
 *         if self.pos != 0:
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         self.pos = self.buf.n_samples * sizeof(int16_t)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pyfxr.pyx":264
 *     def get_audio_data(self, length):
 *         from pyglet.media.codecs import AudioData
 *         if self.pos != 0:             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  }

  /* "_pyfxr.pyx":267
 *             return None
 * 
 *         self.pos = self.buf.n_samples * sizeof(int16_t)             # <<<<<<<<<<<<<<
 *         return AudioData(
 *             self.buf,
 */
  __pyx_v_self->pos = (__pyx_v_self->buf->n_samples * (sizeof(int16_t)));

  /* "_pyfxr.pyx":268
 * 
 *         self.pos = self.buf.n_samples * sizeof(int16_t)
 *         return AudioData(             # <<<<<<<<<<<<<<
 *             self.buf,
 *             self.pos,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "_pyfxr.pyx":270
 *         return AudioData(
 *             self.buf,
 *             self.pos,             # <<<<<<<<<<<<<<
 *             self.buf.n_samples / <float> SAMPLE_RATE,
 *             self.buf.duration,
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "_pyfxr.pyx":271
 *             self.buf,
 *             self.pos,
 *             self.buf.n_samples / <float> SAMPLE_RATE,             # <<<<<<<<<<<<<<
 *             self.buf.duration,
 *             ()
 */
  if (unlikely(((float)__pyx_v_6_pyfxr_SAMPLE_RATE) == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 271, __pyx_L1_error)
  }
  __pyx_t_4 = PyFloat_FromDouble((((float)__pyx_v_self->buf->n_samples) / ((float)__pyx_v_6_pyfxr_SAMPLE_RATE))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "_pyfxr.pyx":272
 *             self.pos,
 *             self.buf.n_samples / <float> SAMPLE_RATE,
 *             self.buf.duration,             # <<<<<<<<<<<<<<
 *             ()
 *         )
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->buf), __pyx_n_s_duration); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "_pyfxr.pyx":273
 *             self.buf.n_samples / <float> SAMPLE_RATE,
 *             self.buf.duration,
 *             ()             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __Pyx_INCREF(__pyx_v_AudioData);
  __pyx_t_6 = __pyx_v_AudioData; __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[6] = {__pyx_t_7, ((PyObject *)__pyx_v_self->buf), __pyx_t_1, __pyx_t_4, __pyx_t_5, __pyx_empty_tuple};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[6] = {__pyx_t_7, ((PyObject *)__pyx_v_self->buf), __pyx_t_1, __pyx_t_4, __pyx_t_5, __pyx_empty_tuple};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(5+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_self->buf));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->buf));
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, ((PyObject *)__pyx_v_self->buf));
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_5);
    __Pyx_INCREF(__pyx_empty_tuple);
    __Pyx_GIVEREF(__pyx_empty_tuple);
    PyTuple_SET_ITEM(__pyx_t_9, 4+__pyx_t_8, __pyx_empty_tuple);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_pyfxr.pyx":262
 *         )
 * 
 *     def get_audio_data(self, length):             # <<<<<<<<<<<<<<
 *         from pyglet.media.codecs import AudioData
 *         if self.pos != 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("_pyfxr.PygletSource.get_audio_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_AudioData);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_12PygletSource_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6_pyfxr_12PygletSource_4__reduce_cython__[] = "PygletSource.__reduce_cython__(self)";
static PyObject *__pyx_pw_6_pyfxr_12PygletSource_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_12PygletSource_4__reduce_cython__(((struct __pyx_obj_6_pyfxr_PygletSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_12PygletSource_4__reduce_cython__(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.buf, self.pos)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->buf));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->buf));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->buf));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.buf, self.pos)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.buf, self.pos)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.buf is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.buf, self.pos)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.buf is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->buf) != Py_None);
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.buf is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, None), state
 *     else:
 */
  __pyx_t_4 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":13
 *         use_setstate = self.buf is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_PygletSource); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_93127629);
    __Pyx_GIVEREF(__pyx_int_93127629);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_93127629);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.buf is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, None), state
 *     else:
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PygletSource__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_PygletSource); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_93127629);
    __Pyx_GIVEREF(__pyx_int_93127629);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_93127629);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_pyfxr.PygletSource.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PygletSource__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_12PygletSource_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_6_pyfxr_12PygletSource_6__setstate_cython__[] = "PygletSource.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_6_pyfxr_12PygletSource_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6_pyfxr_12PygletSource_6__setstate_cython__(((struct __pyx_obj_6_pyfxr_PygletSource *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_12PygletSource_6__setstate_cython__(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PygletSource__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6_pyfxr___pyx_unpickle_PygletSource__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PygletSource, (type(self), 0x58d03cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PygletSource__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pyfxr.PygletSource.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":277
 * 
 * 
 * def tone(             # <<<<<<<<<<<<<<
 *     Wavetable wavetable,
 *     double pitch=440.0,  # Hz, default = A
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_1tone(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_tone[] = "tone(Wavetable wavetable, double pitch=440.0, uint32_t attack=4000, uint32_t decay=4000, uint32_t sustain=30000, uint32_t release=20000)";
static PyMethodDef __pyx_mdef_6_pyfxr_1tone = {"tone", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_1tone, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_tone};
static PyObject *__pyx_pw_6_pyfxr_1tone(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_wavetable = 0;
  double __pyx_v_pitch;
  uint32_t __pyx_v_attack;
  uint32_t __pyx_v_decay;
  uint32_t __pyx_v_sustain;
  uint32_t __pyx_v_release;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("tone (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_wavetable,&__pyx_n_s_pitch,&__pyx_n_s_attack,&__pyx_n_s_decay,&__pyx_n_s_sustain,&__pyx_n_s_release,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_wavetable)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pitch);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attack);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_decay);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sustain);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_release);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "tone") < 0)) __PYX_ERR(0, 277, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_wavetable = ((struct __pyx_obj_6_pyfxr_Wavetable *)values[0]);
    if (values[1]) {
      __pyx_v_pitch = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_pitch == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 279, __pyx_L3_error)
    } else {
      __pyx_v_pitch = ((double)440.0);
    }
    if (values[2]) {
      __pyx_v_attack = __Pyx_PyInt_As_uint32_t(values[2]); if (unlikely((__pyx_v_attack == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 280, __pyx_L3_error)
    } else {
      __pyx_v_attack = ((uint32_t)0xFA0);
    }
    if (values[3]) {
      __pyx_v_decay = __Pyx_PyInt_As_uint32_t(values[3]); if (unlikely((__pyx_v_decay == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 281, __pyx_L3_error)
    } else {
      __pyx_v_decay = ((uint32_t)0xFA0);
    }
    if (values[4]) {
      __pyx_v_sustain = __Pyx_PyInt_As_uint32_t(values[4]); if (unlikely((__pyx_v_sustain == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 282, __pyx_L3_error)
    } else {
      __pyx_v_sustain = ((uint32_t)0x7530);
    }
    if (values[5]) {
      __pyx_v_release = __Pyx_PyInt_As_uint32_t(values[5]); if (unlikely((__pyx_v_release == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 283, __pyx_L3_error)
    } else {
      __pyx_v_release = ((uint32_t)0x4E20);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("tone", 0, 1, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 277, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.tone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wavetable), __pyx_ptype_6_pyfxr_Wavetable, 1, "wavetable", 0))) __PYX_ERR(0, 278, __pyx_L1_error)
  __pyx_r = __pyx_pf_6_pyfxr_tone(__pyx_self, __pyx_v_wavetable, __pyx_v_pitch, __pyx_v_attack, __pyx_v_decay, __pyx_v_sustain, __pyx_v_release);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_tone(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v_wavetable, double __pyx_v_pitch, uint32_t __pyx_v_attack, uint32_t __pyx_v_decay, uint32_t __pyx_v_sustain, uint32_t __pyx_v_release) {
  uint64_t __pyx_v_time;
  size_t __pyx_v_n_samples;
  size_t __pyx_v_i;
  struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_t = 0;
  int16_t *__pyx_v_samples;
  uint64_t __pyx_v_omega;
  int16_t __pyx_v_v;
  float __pyx_v_amplitude;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint64_t __pyx_t_4;
  int16_t *__pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_t_9;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tone", 0);

  /* "_pyfxr.pyx":285
 *     uint32_t release=20000
 * ):
 *     cdef uint64_t time = 0             # <<<<<<<<<<<<<<
 *     cdef size_t n_samples, i
 *     cdef SoundBuffer t
 */
  __pyx_v_time = 0;

  /* "_pyfxr.pyx":298
 *     # High accuracy is needed because single-bit rounding errors add up
 *     # over tens of thousands of samples.
 *     time = 0             # <<<<<<<<<<<<<<
 *     omega = <uint64_t> (pitch * 1024.0 / SAMPLE_RATE * (1 << 32))
 * 
 */
  __pyx_v_time = 0;

  /* "_pyfxr.pyx":299
 *     # over tens of thousands of samples.
 *     time = 0
 *     omega = <uint64_t> (pitch * 1024.0 / SAMPLE_RATE * (1 << 32))             # <<<<<<<<<<<<<<
 * 
 *     n_samples = attack + decay + sustain + release
 */
  __pyx_t_1 = (__pyx_v_pitch * 1024.0);
  if (unlikely(__pyx_v_6_pyfxr_SAMPLE_RATE == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 299, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble((__pyx_t_1 / ((double)__pyx_v_6_pyfxr_SAMPLE_RATE))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_int_4294967296); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_uint64_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_omega = ((uint64_t)__pyx_t_4);

  /* "_pyfxr.pyx":301
 *     omega = <uint64_t> (pitch * 1024.0 / SAMPLE_RATE * (1 << 32))
 * 
 *     n_samples = attack + decay + sustain + release             # <<<<<<<<<<<<<<
 *     t = SoundBuffer(n_samples)
 *     samples = t.samples
 */
  __pyx_v_n_samples = (((__pyx_v_attack + __pyx_v_decay) + __pyx_v_sustain) + __pyx_v_release);

  /* "_pyfxr.pyx":302
 * 
 *     n_samples = attack + decay + sustain + release
 *     t = SoundBuffer(n_samples)             # <<<<<<<<<<<<<<
 *     samples = t.samples
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_6_pyfxr_SoundBuffer), __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_t = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "_pyfxr.pyx":303
 *     n_samples = attack + decay + sustain + release
 *     t = SoundBuffer(n_samples)
 *     samples = t.samples             # <<<<<<<<<<<<<<
 * 
 *     with nogil:
 */
  __pyx_t_5 = __pyx_v_t->samples;
  __pyx_v_samples = __pyx_t_5;

  /* "_pyfxr.pyx":305
 *     samples = t.samples
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         for i in range(n_samples):
 *             time += omega
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "_pyfxr.pyx":306
 * 
 *     with nogil:
 *         for i in range(n_samples):             # <<<<<<<<<<<<<<
 *             time += omega
 *             v = wavetable.wavetable[(time >> 32) & 1023]
 */
        __pyx_t_6 = __pyx_v_n_samples;
        __pyx_t_7 = __pyx_t_6;
        for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
          __pyx_v_i = __pyx_t_8;

          /* "_pyfxr.pyx":307
 *     with nogil:
 *         for i in range(n_samples):
 *             time += omega             # <<<<<<<<<<<<<<
 *             v = wavetable.wavetable[(time >> 32) & 1023]
 * 
 */
          __pyx_v_time = (__pyx_v_time + __pyx_v_omega);

          /* "_pyfxr.pyx":308
 *         for i in range(n_samples):
 *             time += omega
 *             v = wavetable.wavetable[(time >> 32) & 1023]             # <<<<<<<<<<<<<<
 * 
 *             if i < attack:
 */
          __pyx_v_v = (__pyx_v_wavetable->wavetable[((__pyx_v_time >> 32) & 0x3FF)]);

          /* "_pyfxr.pyx":310
 *             v = wavetable.wavetable[(time >> 32) & 1023]
 * 
 *             if i < attack:             # <<<<<<<<<<<<<<
 *                 amplitude = (i / <float> attack)
 *             elif i < attack + decay:
 */
          __pyx_t_9 = ((__pyx_v_i < __pyx_v_attack) != 0);
          if (__pyx_t_9) {

            /* "_pyfxr.pyx":311
 * 
 *             if i < attack:
 *                 amplitude = (i / <float> attack)             # <<<<<<<<<<<<<<
 *             elif i < attack + decay:
 *                 amplitude = (1.0 - (i - attack) / <float> decay * 0.3)
 */
            if (unlikely(((float)__pyx_v_attack) == 0)) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              PyErr_SetString(PyExc_ZeroDivisionError, "float division");
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 311, __pyx_L4_error)
            }
            __pyx_v_amplitude = (((float)__pyx_v_i) / ((float)__pyx_v_attack));

            /* "_pyfxr.pyx":310
 *             v = wavetable.wavetable[(time >> 32) & 1023]
 * 
 *             if i < attack:             # <<<<<<<<<<<<<<
 *                 amplitude = (i / <float> attack)
 *             elif i < attack + decay:
 */
            goto __pyx_L8;
          }

          /* "_pyfxr.pyx":312
 *             if i < attack:
 *                 amplitude = (i / <float> attack)
 *             elif i < attack + decay:             # <<<<<<<<<<<<<<
 *                 amplitude = (1.0 - (i - attack) / <float> decay * 0.3)
 *             elif i < attack + decay + sustain:
 */
          __pyx_t_9 = ((__pyx_v_i < (__pyx_v_attack + __pyx_v_decay)) != 0);
          if (__pyx_t_9) {

            /* "_pyfxr.pyx":313
 *                 amplitude = (i / <float> attack)
 *             elif i < attack + decay:
 *                 amplitude = (1.0 - (i - attack) / <float> decay * 0.3)             # <<<<<<<<<<<<<<
 *             elif i < attack + decay + sustain:
 *                 amplitude = 0.7
 */
            __pyx_t_10 = (__pyx_v_i - __pyx_v_attack);
            if (unlikely(((float)__pyx_v_decay) == 0)) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              PyErr_SetString(PyExc_ZeroDivisionError, "float division");
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 313, __pyx_L4_error)
            }
            __pyx_v_amplitude = (1.0 - ((((float)__pyx_t_10) / ((float)__pyx_v_decay)) * 0.3));

            /* "_pyfxr.pyx":312
 *             if i < attack:
 *                 amplitude = (i / <float> attack)
 *             elif i < attack + decay:             # <<<<<<<<<<<<<<
 *                 amplitude = (1.0 - (i - attack) / <float> decay * 0.3)
 *             elif i < attack + decay + sustain:
 */
            goto __pyx_L8;
          }

          /* "_pyfxr.pyx":314
 *             elif i < attack + decay:
 *                 amplitude = (1.0 - (i - attack) / <float> decay * 0.3)
 *             elif i < attack + decay + sustain:             # <<<<<<<<<<<<<<
 *                 amplitude = 0.7
 *             else:
 */
          __pyx_t_9 = ((__pyx_v_i < ((__pyx_v_attack + __pyx_v_decay) + __pyx_v_sustain)) != 0);
          if (__pyx_t_9) {

            /* "_pyfxr.pyx":315
 *                 amplitude = (1.0 - (i - attack) / <float> decay * 0.3)
 *             elif i < attack + decay + sustain:
 *                 amplitude = 0.7             # <<<<<<<<<<<<<<
 *             else:
 *                 amplitude = (n_samples - i) / release * 0.7
 */
            __pyx_v_amplitude = 0.7;

            /* "_pyfxr.pyx":314
 *             elif i < attack + decay:
 *                 amplitude = (1.0 - (i - attack) / <float> decay * 0.3)
 *             elif i < attack + decay + sustain:             # <<<<<<<<<<<<<<
 *                 amplitude = 0.7
 *             else:
 */
            goto __pyx_L8;
          }

          /* "_pyfxr.pyx":317
 *                 amplitude = 0.7
 *             else:
 *                 amplitude = (n_samples - i) / release * 0.7             # <<<<<<<<<<<<<<
 * 
 *             samples[i] = <int16_t> (amplitude * v)
 */
          /*else*/ {
            __pyx_t_10 = (__pyx_v_n_samples - __pyx_v_i);
            if (unlikely(__pyx_v_release == 0)) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              PyErr_SetString(PyExc_ZeroDivisionError, "float division");
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 317, __pyx_L4_error)
            }
            __pyx_v_amplitude = ((((double)__pyx_t_10) / ((double)__pyx_v_release)) * 0.7);
          }
          __pyx_L8:;

          /* "_pyfxr.pyx":319
 *                 amplitude = (n_samples - i) / release * 0.7
 * 
 *             samples[i] = <int16_t> (amplitude * v)             # <<<<<<<<<<<<<<
 * 
 *     return t
 */
          (__pyx_v_samples[__pyx_v_i]) = ((int16_t)(__pyx_v_amplitude * __pyx_v_v));
        }
      }

      /* "_pyfxr.pyx":305
 *     samples = t.samples
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         for i in range(n_samples):
 *             time += omega
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "_pyfxr.pyx":321
 *             samples[i] = <int16_t> (amplitude * v)
 * 
 *     return t             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_t));
  __pyx_r = ((PyObject *)__pyx_v_t);
  goto __pyx_L0;

  /* "_pyfxr.pyx":277
 * 
 * 
 * def tone(             # <<<<<<<<<<<<<<
 *     Wavetable wavetable,
 *     double pitch=440.0,  # Hz, default = A
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_pyfxr.tone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":328
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef void reset_sample(             # <<<<<<<<<<<<<<
 *     float p_base_freq,
 *     float p_freq_limit,
 */

static void __pyx_f_6_pyfxr_reset_sample(float __pyx_v_p_base_freq, float __pyx_v_p_freq_limit, float __pyx_v_p_freq_ramp, float __pyx_v_p_freq_dramp, float __pyx_v_p_duty, float __pyx_v_p_duty_ramp, float __pyx_v_p_arp_mod, float __pyx_v_p_arp_speed, double *__pyx_v_fperiod, int *__pyx_v_period, double *__pyx_v_fmaxperiod, double *__pyx_v_fslide, double *__pyx_v_fdslide, float *__pyx_v_square_duty, float *__pyx_v_square_slide, double *__pyx_v_arp_mod, int *__pyx_v_arp_time, int *__pyx_v_arp_limit) {
  int __pyx_t_1;

  /* "_pyfxr.pyx":348
 *     int *arp_limit
 * ) nogil:
 *     fperiod[0] = 100.0 / (p_base_freq * p_base_freq + 0.001)             # <<<<<<<<<<<<<<
 *     period[0] = <int> fperiod[0]
 *     fmaxperiod[0] = 100.0 / (p_freq_limit * p_freq_limit + 0.001)
 */
  (__pyx_v_fperiod[0]) = (100.0 / ((__pyx_v_p_base_freq * __pyx_v_p_base_freq) + 0.001));

  /* "_pyfxr.pyx":349
 * ) nogil:
 *     fperiod[0] = 100.0 / (p_base_freq * p_base_freq + 0.001)
 *     period[0] = <int> fperiod[0]             # <<<<<<<<<<<<<<
 *     fmaxperiod[0] = 100.0 / (p_freq_limit * p_freq_limit + 0.001)
 *     fslide[0] = 1.0 - p_freq_ramp ** 3.0 * 0.01
 */
  (__pyx_v_period[0]) = ((int)(__pyx_v_fperiod[0]));

  /* "_pyfxr.pyx":350
 *     fperiod[0] = 100.0 / (p_base_freq * p_base_freq + 0.001)
 *     period[0] = <int> fperiod[0]
 *     fmaxperiod[0] = 100.0 / (p_freq_limit * p_freq_limit + 0.001)             # <<<<<<<<<<<<<<
 *     fslide[0] = 1.0 - p_freq_ramp ** 3.0 * 0.01
 *     fdslide[0] = p_freq_dramp ** 3.0 * -0.000001;
 */
  (__pyx_v_fmaxperiod[0]) = (100.0 / ((__pyx_v_p_freq_limit * __pyx_v_p_freq_limit) + 0.001));

  /* "_pyfxr.pyx":351
 *     period[0] = <int> fperiod[0]
 *     fmaxperiod[0] = 100.0 / (p_freq_limit * p_freq_limit + 0.001)
 *     fslide[0] = 1.0 - p_freq_ramp ** 3.0 * 0.01             # <<<<<<<<<<<<<<
 *     fdslide[0] = p_freq_dramp ** 3.0 * -0.000001;
 *     square_duty[0] = 0.5 - p_duty * 0.5;
 */
  (__pyx_v_fslide[0]) = (1.0 - (pow(((double)__pyx_v_p_freq_ramp), 3.0) * 0.01));

  /* "_pyfxr.pyx":352
 *     fmaxperiod[0] = 100.0 / (p_freq_limit * p_freq_limit + 0.001)
 *     fslide[0] = 1.0 - p_freq_ramp ** 3.0 * 0.01
 *     fdslide[0] = p_freq_dramp ** 3.0 * -0.000001;             # <<<<<<<<<<<<<<
 *     square_duty[0] = 0.5 - p_duty * 0.5;
 *     square_slide[0] = p_duty_ramp * -0.00005;
 */
  (__pyx_v_fdslide[0]) = (pow(((double)__pyx_v_p_freq_dramp), 3.0) * -0.000001);

  /* "_pyfxr.pyx":353
 *     fslide[0] = 1.0 - p_freq_ramp ** 3.0 * 0.01
 *     fdslide[0] = p_freq_dramp ** 3.0 * -0.000001;
 *     square_duty[0] = 0.5 - p_duty * 0.5;             # <<<<<<<<<<<<<<
 *     square_slide[0] = p_duty_ramp * -0.00005;
 *     if p_arp_mod >= 0.0:
 */
  (__pyx_v_square_duty[0]) = (0.5 - (__pyx_v_p_duty * 0.5));

  /* "_pyfxr.pyx":354
 *     fdslide[0] = p_freq_dramp ** 3.0 * -0.000001;
 *     square_duty[0] = 0.5 - p_duty * 0.5;
 *     square_slide[0] = p_duty_ramp * -0.00005;             # <<<<<<<<<<<<<<
 *     if p_arp_mod >= 0.0:
 *         arp_mod[0] = 1.0 - p_arp_mod ** 2.0 * 0.9
 */
  (__pyx_v_square_slide[0]) = (__pyx_v_p_duty_ramp * -0.00005);

  /* "_pyfxr.pyx":355
 *     square_duty[0] = 0.5 - p_duty * 0.5;
 *     square_slide[0] = p_duty_ramp * -0.00005;
 *     if p_arp_mod >= 0.0:             # <<<<<<<<<<<<<<
 *         arp_mod[0] = 1.0 - p_arp_mod ** 2.0 * 0.9
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_p_arp_mod >= 0.0) != 0);
  if (__pyx_t_1) {

    /* "_pyfxr.pyx":356
 *     square_slide[0] = p_duty_ramp * -0.00005;
 *     if p_arp_mod >= 0.0:
 *         arp_mod[0] = 1.0 - p_arp_mod ** 2.0 * 0.9             # <<<<<<<<<<<<<<
 *     else:
 *         arp_mod[0] = 1.0 + p_arp_mod ** 2.0 * 10.0
 */
    (__pyx_v_arp_mod[0]) = (1.0 - (pow(((double)__pyx_v_p_arp_mod), 2.0) * 0.9));

    /* "_pyfxr.pyx":355
 *     square_duty[0] = 0.5 - p_duty * 0.5;
 *     square_slide[0] = p_duty_ramp * -0.00005;
 *     if p_arp_mod >= 0.0:             # <<<<<<<<<<<<<<
 *         arp_mod[0] = 1.0 - p_arp_mod ** 2.0 * 0.9
 *     else:
 */
    goto __pyx_L3;
  }

  /* "_pyfxr.pyx":358
 *         arp_mod[0] = 1.0 - p_arp_mod ** 2.0 * 0.9
 *     else:
 *         arp_mod[0] = 1.0 + p_arp_mod ** 2.0 * 10.0             # <<<<<<<<<<<<<<
 *     arp_time[0] = 0
 *     arp_limit[0] = <int> (((1.0 - p_arp_speed) ** 2.0) * 20000 + 32)
 */
  /*else*/ {
    (__pyx_v_arp_mod[0]) = (1.0 + (pow(((double)__pyx_v_p_arp_mod), 2.0) * 10.0));
  }
  __pyx_L3:;

  /* "_pyfxr.pyx":359
 *     else:
 *         arp_mod[0] = 1.0 + p_arp_mod ** 2.0 * 10.0
 *     arp_time[0] = 0             # <<<<<<<<<<<<<<
 *     arp_limit[0] = <int> (((1.0 - p_arp_speed) ** 2.0) * 20000 + 32)
 *     if p_arp_speed == 1.0:
 */
  (__pyx_v_arp_time[0]) = 0;

  /* "_pyfxr.pyx":360
 *         arp_mod[0] = 1.0 + p_arp_mod ** 2.0 * 10.0
 *     arp_time[0] = 0
 *     arp_limit[0] = <int> (((1.0 - p_arp_speed) ** 2.0) * 20000 + 32)             # <<<<<<<<<<<<<<
 *     if p_arp_speed == 1.0:
 *         arp_limit[0] = 0
 */
  (__pyx_v_arp_limit[0]) = ((int)((pow((1.0 - __pyx_v_p_arp_speed), 2.0) * 20000.0) + 32.0));

  /* "_pyfxr.pyx":361
 *     arp_time[0] = 0
 *     arp_limit[0] = <int> (((1.0 - p_arp_speed) ** 2.0) * 20000 + 32)
 *     if p_arp_speed == 1.0:             # <<<<<<<<<<<<<<
 *         arp_limit[0] = 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_p_arp_speed == 1.0) != 0);
  if (__pyx_t_1) {

    /* "_pyfxr.pyx":362
 *     arp_limit[0] = <int> (((1.0 - p_arp_speed) ** 2.0) * 20000 + 32)
 *     if p_arp_speed == 1.0:
 *         arp_limit[0] = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_arp_limit[0]) = 0;

    /* "_pyfxr.pyx":361
 *     arp_time[0] = 0
 *     arp_limit[0] = <int> (((1.0 - p_arp_speed) ** 2.0) * 20000 + 32)
 *     if p_arp_speed == 1.0:             # <<<<<<<<<<<<<<
 *         arp_limit[0] = 0
 * 
 */
  }

  /* "_pyfxr.pyx":328
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef void reset_sample(             # <<<<<<<<<<<<<<
 *     float p_base_freq,
 *     float p_freq_limit,
 */

  /* function exit code */
}

/* "_pyfxr.pyx":368
 * 
 * 
 * cdef void clamp(floating *v, floating min, floating max) nogil:             # <<<<<<<<<<<<<<
 *     """Clamp the given value v to between min and max."""
 *     if v[0] < min:
 */

static void __pyx_fuse_0__pyx_f_6_pyfxr_clamp(float *__pyx_v_v, float __pyx_v_min, float __pyx_v_max) {
  int __pyx_t_1;

  /* "_pyfxr.pyx":370
 * cdef void clamp(floating *v, floating min, floating max) nogil:
 *     """Clamp the given value v to between min and max."""
 *     if v[0] < min:             # <<<<<<<<<<<<<<
 *         v[0] = min
 *     elif v[0] > max:
 */
  __pyx_t_1 = (((__pyx_v_v[0]) < __pyx_v_min) != 0);
  if (__pyx_t_1) {

    /* "_pyfxr.pyx":371
 *     """Clamp the given value v to between min and max."""
 *     if v[0] < min:
 *         v[0] = min             # <<<<<<<<<<<<<<
 *     elif v[0] > max:
 *         v[0] = max
 */
    (__pyx_v_v[0]) = __pyx_v_min;

    /* "_pyfxr.pyx":370
 * cdef void clamp(floating *v, floating min, floating max) nogil:
 *     """Clamp the given value v to between min and max."""
 *     if v[0] < min:             # <<<<<<<<<<<<<<
 *         v[0] = min
 *     elif v[0] > max:
 */
    goto __pyx_L3;
  }

  /* "_pyfxr.pyx":372
 *     if v[0] < min:
 *         v[0] = min
 *     elif v[0] > max:             # <<<<<<<<<<<<<<
 *         v[0] = max
 * 
 */
  __pyx_t_1 = (((__pyx_v_v[0]) > __pyx_v_max) != 0);
  if (__pyx_t_1) {

    /* "_pyfxr.pyx":373
 *         v[0] = min
 *     elif v[0] > max:
 *         v[0] = max             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_v[0]) = __pyx_v_max;

    /* "_pyfxr.pyx":372
 *     if v[0] < min:
 *         v[0] = min
 *     elif v[0] > max:             # <<<<<<<<<<<<<<
 *         v[0] = max
 * 
 */
  }
  __pyx_L3:;

  /* "_pyfxr.pyx":368
 * 
 * 
 * cdef void clamp(floating *v, floating min, floating max) nogil:             # <<<<<<<<<<<<<<
 *     """Clamp the given value v to between min and max."""
 *     if v[0] < min:
 */

  /* function exit code */
}

/* "_pyfxr.pyx":377
 * 
 * @cython.boundscheck(False)
 * cdef void fill_noise(float *noise_buffer) nogil:             # <<<<<<<<<<<<<<
 *     for i in range(32):
 *         noise_buffer[i] = frnd(2.0) - 1.0
 */

static void __pyx_f_6_pyfxr_fill_noise(float *__pyx_v_noise_buffer) {
  long __pyx_v_i;
  long __pyx_t_1;

  /* "_pyfxr.pyx":378
 * @cython.boundscheck(False)
 * cdef void fill_noise(float *noise_buffer) nogil:
 *     for i in range(32):             # <<<<<<<<<<<<<<
 *         noise_buffer[i] = frnd(2.0) - 1.0
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 32; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "_pyfxr.pyx":379
 * cdef void fill_noise(float *noise_buffer) nogil:
 *     for i in range(32):
 *         noise_buffer[i] = frnd(2.0) - 1.0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_noise_buffer[__pyx_v_i]) = (__pyx_f_6_pyfxr_frnd(2.0) - 1.0);
  }

  /* "_pyfxr.pyx":377
 * 
 * @cython.boundscheck(False)
 * cdef void fill_noise(float *noise_buffer) nogil:             # <<<<<<<<<<<<<<
 *     for i in range(32):
 *         noise_buffer[i] = frnd(2.0) - 1.0
 */

  /* function exit code */
}

/* "_pyfxr.pyx":383
 * 
 * @cython.cdivision(True)
 * cdef float frnd(float range_) nogil:             # <<<<<<<<<<<<<<
 *     return <float> (rand() % 10001) / 10000 * range_
 * 
 */

static float __pyx_f_6_pyfxr_frnd(float __pyx_v_range_) {
  float __pyx_r;

  /* "_pyfxr.pyx":384
 * @cython.cdivision(True)
 * cdef float frnd(float range_) nogil:
 *     return <float> (rand() % 10001) / 10000 * range_             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((((float)(rand() % 0x2711)) / 10000.0) * __pyx_v_range_);
  goto __pyx_L0;

  /* "_pyfxr.pyx":383
 * 
 * @cython.cdivision(True)
 * cdef float frnd(float range_) nogil:             # <<<<<<<<<<<<<<
 *     return <float> (rand() % 10001) / 10000 * range_
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "_pyfxr.pyx":391
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * def sfx(             # <<<<<<<<<<<<<<
 *     int wave_type=0,
 *     float p_base_freq=0.3,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_3sfx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_2sfx[] = "sfx(int wave_type=0, float p_base_freq=0.3, float p_freq_limit=0.0, float p_freq_ramp=0.0, float p_freq_dramp=0.0, float p_duty=0.0, float p_duty_ramp=0.0, float p_vib_strength=0.0, float p_vib_speed=0.0, float p_vib_delay=0.0, float p_env_attack=0.0, float p_env_sustain=0.3, float p_env_decay=0.4, float p_env_punch=0.0, float p_lpf_resonance=0.0, float p_lpf_freq=1.0, float p_lpf_ramp=0.0, float p_hpf_freq=0.0, float p_hpf_ramp=0.0, float p_pha_offset=0.0, float p_pha_ramp=0.0, float p_repeat_speed=0.0, float p_arp_speed=0.0, float p_arp_mod=0.0)";
static PyMethodDef __pyx_mdef_6_pyfxr_3sfx = {"sfx", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_3sfx, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_2sfx};
static PyObject *__pyx_pw_6_pyfxr_3sfx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_wave_type;
  float __pyx_v_p_base_freq;
  float __pyx_v_p_freq_limit;
  float __pyx_v_p_freq_ramp;
  float __pyx_v_p_freq_dramp;
  float __pyx_v_p_duty;
  float __pyx_v_p_duty_ramp;
  float __pyx_v_p_vib_strength;
  float __pyx_v_p_vib_speed;
  CYTHON_UNUSED float __pyx_v_p_vib_delay;
  float __pyx_v_p_env_attack;
  float __pyx_v_p_env_sustain;
  float __pyx_v_p_env_decay;
  float __pyx_v_p_env_punch;
  float __pyx_v_p_lpf_resonance;
  float __pyx_v_p_lpf_freq;
  float __pyx_v_p_lpf_ramp;
  float __pyx_v_p_hpf_freq;
  float __pyx_v_p_hpf_ramp;
  float __pyx_v_p_pha_offset;
  float __pyx_v_p_pha_ramp;
  float __pyx_v_p_repeat_speed;
  float __pyx_v_p_arp_speed;
  float __pyx_v_p_arp_mod;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sfx (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_wave_type,&__pyx_n_s_p_base_freq,&__pyx_n_s_p_freq_limit,&__pyx_n_s_p_freq_ramp,&__pyx_n_s_p_freq_dramp,&__pyx_n_s_p_duty,&__pyx_n_s_p_duty_ramp,&__pyx_n_s_p_vib_strength,&__pyx_n_s_p_vib_speed,&__pyx_n_s_p_vib_delay,&__pyx_n_s_p_env_attack,&__pyx_n_s_p_env_sustain,&__pyx_n_s_p_env_decay,&__pyx_n_s_p_env_punch,&__pyx_n_s_p_lpf_resonance,&__pyx_n_s_p_lpf_freq,&__pyx_n_s_p_lpf_ramp,&__pyx_n_s_p_hpf_freq,&__pyx_n_s_p_hpf_ramp,&__pyx_n_s_p_pha_offset,&__pyx_n_s_p_pha_ramp,&__pyx_n_s_p_repeat_speed,&__pyx_n_s_p_arp_speed,&__pyx_n_s_p_arp_mod,0};
    PyObject* values[24] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 24: values[23] = PyTuple_GET_ITEM(__pyx_args, 23);
        CYTHON_FALLTHROUGH;
        case 23: values[22] = PyTuple_GET_ITEM(__pyx_args, 22);
        CYTHON_FALLTHROUGH;
        case 22: values[21] = PyTuple_GET_ITEM(__pyx_args, 21);
        CYTHON_FALLTHROUGH;
        case 21: values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
        CYTHON_FALLTHROUGH;
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_wave_type);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_base_freq);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_freq_limit);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_freq_ramp);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_freq_dramp);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_duty);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_duty_ramp);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_vib_strength);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_vib_speed);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_vib_delay);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_env_attack);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_env_sustain);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_env_decay);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_env_punch);
          if (value) { values[13] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_lpf_resonance);
          if (value) { values[14] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_lpf_freq);
          if (value) { values[15] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_lpf_ramp);
          if (value) { values[16] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_hpf_freq);
          if (value) { values[17] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_hpf_ramp);
          if (value) { values[18] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_pha_offset);
          if (value) { values[19] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 20:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_pha_ramp);
          if (value) { values[20] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 21:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_repeat_speed);
          if (value) { values[21] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 22:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_arp_speed);
          if (value) { values[22] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 23:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p_arp_mod);
          if (value) { values[23] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sfx") < 0)) __PYX_ERR(0, 391, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 24: values[23] = PyTuple_GET_ITEM(__pyx_args, 23);
        CYTHON_FALLTHROUGH;
        case 23: values[22] = PyTuple_GET_ITEM(__pyx_args, 22);
        CYTHON_FALLTHROUGH;
        case 22: values[21] = PyTuple_GET_ITEM(__pyx_args, 21);
        CYTHON_FALLTHROUGH;
        case 21: values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
        CYTHON_FALLTHROUGH;
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_wave_type = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_wave_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 392, __pyx_L3_error)
    } else {
      __pyx_v_wave_type = ((int)0);
    }
    if (values[1]) {
      __pyx_v_p_base_freq = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_p_base_freq == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 393, __pyx_L3_error)
    } else {
      __pyx_v_p_base_freq = ((float)0.3);
    }
    if (values[2]) {
      __pyx_v_p_freq_limit = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_p_freq_limit == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 394, __pyx_L3_error)
    } else {
      __pyx_v_p_freq_limit = ((float)0.0);
    }
    if (values[3]) {
      __pyx_v_p_freq_ramp = __pyx_PyFloat_AsFloat(values[3]); if (unlikely((__pyx_v_p_freq_ramp == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L3_error)
    } else {
      __pyx_v_p_freq_ramp = ((float)0.0);
    }
    if (values[4]) {
      __pyx_v_p_freq_dramp = __pyx_PyFloat_AsFloat(values[4]); if (unlikely((__pyx_v_p_freq_dramp == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 396, __pyx_L3_error)
    } else {
      __pyx_v_p_freq_dramp = ((float)0.0);
    }
    if (values[5]) {
      __pyx_v_p_duty = __pyx_PyFloat_AsFloat(values[5]); if (unlikely((__pyx_v_p_duty == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 397, __pyx_L3_error)
    } else {
      __pyx_v_p_duty = ((float)0.0);
    }
    if (values[6]) {
      __pyx_v_p_duty_ramp = __pyx_PyFloat_AsFloat(values[6]); if (unlikely((__pyx_v_p_duty_ramp == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 398, __pyx_L3_error)
    } else {
      __pyx_v_p_duty_ramp = ((float)0.0);
    }
    if (values[7]) {
      __pyx_v_p_vib_strength = __pyx_PyFloat_AsFloat(values[7]); if (unlikely((__pyx_v_p_vib_strength == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    } else {
      __pyx_v_p_vib_strength = ((float)0.0);
    }
    if (values[8]) {
      __pyx_v_p_vib_speed = __pyx_PyFloat_AsFloat(values[8]); if (unlikely((__pyx_v_p_vib_speed == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 400, __pyx_L3_error)
    } else {
      __pyx_v_p_vib_speed = ((float)0.0);
    }
    if (values[9]) {
      __pyx_v_p_vib_delay = __pyx_PyFloat_AsFloat(values[9]); if (unlikely((__pyx_v_p_vib_delay == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 401, __pyx_L3_error)
    } else {
      __pyx_v_p_vib_delay = ((float)0.0);
    }
    if (values[10]) {
      __pyx_v_p_env_attack = __pyx_PyFloat_AsFloat(values[10]); if (unlikely((__pyx_v_p_env_attack == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L3_error)
    } else {
      __pyx_v_p_env_attack = ((float)0.0);
    }
    if (values[11]) {
      __pyx_v_p_env_sustain = __pyx_PyFloat_AsFloat(values[11]); if (unlikely((__pyx_v_p_env_sustain == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 403, __pyx_L3_error)
    } else {
      __pyx_v_p_env_sustain = ((float)0.3);
    }
    if (values[12]) {
      __pyx_v_p_env_decay = __pyx_PyFloat_AsFloat(values[12]); if (unlikely((__pyx_v_p_env_decay == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 404, __pyx_L3_error)
    } else {
      __pyx_v_p_env_decay = ((float)0.4);
    }
    if (values[13]) {
      __pyx_v_p_env_punch = __pyx_PyFloat_AsFloat(values[13]); if (unlikely((__pyx_v_p_env_punch == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 405, __pyx_L3_error)
    } else {
      __pyx_v_p_env_punch = ((float)0.0);
    }
    if (values[14]) {
      __pyx_v_p_lpf_resonance = __pyx_PyFloat_AsFloat(values[14]); if (unlikely((__pyx_v_p_lpf_resonance == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 406, __pyx_L3_error)
    } else {
      __pyx_v_p_lpf_resonance = ((float)0.0);
    }
    if (values[15]) {
      __pyx_v_p_lpf_freq = __pyx_PyFloat_AsFloat(values[15]); if (unlikely((__pyx_v_p_lpf_freq == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 407, __pyx_L3_error)
    } else {
      __pyx_v_p_lpf_freq = ((float)1.0);
    }
    if (values[16]) {
      __pyx_v_p_lpf_ramp = __pyx_PyFloat_AsFloat(values[16]); if (unlikely((__pyx_v_p_lpf_ramp == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 408, __pyx_L3_error)
    } else {
      __pyx_v_p_lpf_ramp = ((float)0.0);
    }
    if (values[17]) {
      __pyx_v_p_hpf_freq = __pyx_PyFloat_AsFloat(values[17]); if (unlikely((__pyx_v_p_hpf_freq == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 409, __pyx_L3_error)
    } else {
      __pyx_v_p_hpf_freq = ((float)0.0);
    }
    if (values[18]) {
      __pyx_v_p_hpf_ramp = __pyx_PyFloat_AsFloat(values[18]); if (unlikely((__pyx_v_p_hpf_ramp == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 410, __pyx_L3_error)
    } else {
      __pyx_v_p_hpf_ramp = ((float)0.0);
    }
    if (values[19]) {
      __pyx_v_p_pha_offset = __pyx_PyFloat_AsFloat(values[19]); if (unlikely((__pyx_v_p_pha_offset == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 411, __pyx_L3_error)
    } else {
      __pyx_v_p_pha_offset = ((float)0.0);
    }
    if (values[20]) {
      __pyx_v_p_pha_ramp = __pyx_PyFloat_AsFloat(values[20]); if (unlikely((__pyx_v_p_pha_ramp == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
    } else {
      __pyx_v_p_pha_ramp = ((float)0.0);
    }
    if (values[21]) {
      __pyx_v_p_repeat_speed = __pyx_PyFloat_AsFloat(values[21]); if (unlikely((__pyx_v_p_repeat_speed == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L3_error)
    } else {
      __pyx_v_p_repeat_speed = ((float)0.0);
    }
    if (values[22]) {
      __pyx_v_p_arp_speed = __pyx_PyFloat_AsFloat(values[22]); if (unlikely((__pyx_v_p_arp_speed == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 414, __pyx_L3_error)
    } else {
      __pyx_v_p_arp_speed = ((float)0.0);
    }
    if (values[23]) {
      __pyx_v_p_arp_mod = __pyx_PyFloat_AsFloat(values[23]); if (unlikely((__pyx_v_p_arp_mod == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 415, __pyx_L3_error)
    } else {
      __pyx_v_p_arp_mod = ((float)0.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sfx", 0, 0, 24, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 391, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.sfx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6_pyfxr_2sfx(__pyx_self, __pyx_v_wave_type, __pyx_v_p_base_freq, __pyx_v_p_freq_limit, __pyx_v_p_freq_ramp, __pyx_v_p_freq_dramp, __pyx_v_p_duty, __pyx_v_p_duty_ramp, __pyx_v_p_vib_strength, __pyx_v_p_vib_speed, __pyx_v_p_vib_delay, __pyx_v_p_env_attack, __pyx_v_p_env_sustain, __pyx_v_p_env_decay, __pyx_v_p_env_punch, __pyx_v_p_lpf_resonance, __pyx_v_p_lpf_freq, __pyx_v_p_lpf_ramp, __pyx_v_p_hpf_freq, __pyx_v_p_hpf_ramp, __pyx_v_p_pha_offset, __pyx_v_p_pha_ramp, __pyx_v_p_repeat_speed, __pyx_v_p_arp_speed, __pyx_v_p_arp_mod);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_2sfx(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_wave_type, float __pyx_v_p_base_freq, float __pyx_v_p_freq_limit, float __pyx_v_p_freq_ramp, float __pyx_v_p_freq_dramp, float __pyx_v_p_duty, float __pyx_v_p_duty_ramp, float __pyx_v_p_vib_strength, float __pyx_v_p_vib_speed, CYTHON_UNUSED float __pyx_v_p_vib_delay, float __pyx_v_p_env_attack, float __pyx_v_p_env_sustain, float __pyx_v_p_env_decay, float __pyx_v_p_env_punch, float __pyx_v_p_lpf_resonance, float __pyx_v_p_lpf_freq, float __pyx_v_p_lpf_ramp, float __pyx_v_p_hpf_freq, float __pyx_v_p_hpf_ramp, float __pyx_v_p_pha_offset, float __pyx_v_p_pha_ramp, float __pyx_v_p_repeat_speed, float __pyx_v_p_arp_speed, float __pyx_v_p_arp_mod) {
  int __pyx_v_phase;
  int __pyx_v_period;
  int __pyx_v_env_stage;
  int __pyx_v_env_time;
  int __pyx_v_iphase;
  int __pyx_v_ipp;
  double __pyx_v_fperiod;
  double __pyx_v_fmaxperiod;
  double __pyx_v_fslide;
  double __pyx_v_fdslide;
  double __pyx_v_arp_mod;
  float __pyx_v_square_duty;
  float __pyx_v_square_slide;
  float __pyx_v_env_vol;
  float __pyx_v_fphase;
  float __pyx_v_fdphase;
  int __pyx_v_env_length[3];
  float __pyx_v_phaser_buffer[0x400];
  float __pyx_v_noise_buffer[32];
  float __pyx_v_fltp;
  float __pyx_v_fltdp;
  float __pyx_v_fltphp;
  float __pyx_v_fltw;
  float __pyx_v_fltw_d;
  float __pyx_v_fltdmp;
  float __pyx_v_flthp;
  float __pyx_v_flthp_d;
  float __pyx_v_vib_phase;
  float __pyx_v_vib_speed;
  float __pyx_v_vib_amp;
  int __pyx_v_rep_time;
  int __pyx_v_rep_limit;
  int __pyx_v_arp_time;
  int __pyx_v_arp_limit;
  float __pyx_v_rfperiod;
  float __pyx_v_ssample;
  float __pyx_v_sample;
  float __pyx_v_fp;
  float __pyx_v_pp;
  CYTHON_UNUSED int __pyx_v_si;
  size_t __pyx_v_n_samples;
  struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_s = 0;
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  double __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sfx", 0);

  /* "_pyfxr.pyx":417
 *     float p_arp_mod=0.0,
 * ):
 *     cdef int phase=0, period, env_stage, env_time, iphase, ipp             # <<<<<<<<<<<<<<
 *     cdef double fperiod, fmaxperiod, fslide, fdslide, arp_mod
 *     cdef float square_duty, square_slide, env_vol, fphase, fdphase
 */
  __pyx_v_phase = 0;

  /* "_pyfxr.pyx":423
 *     cdef float phaser_buffer[1024]
 *     cdef float noise_buffer[32]
 *     cdef float fltp = 0.0, fltdp = 0.0, fltphp = 0.0             # <<<<<<<<<<<<<<
 *     cdef float fltw, fltw_d, fltdmp, flthp, flthp_d, vib_phase, vib_speed, vib_amp
 *     cdef int rep_time, rep_limit, arp_time, arp_limit
 */
  __pyx_v_fltp = 0.0;
  __pyx_v_fltdp = 0.0;
  __pyx_v_fltphp = 0.0;

  /* "_pyfxr.pyx":427
 *     cdef int rep_time, rep_limit, arp_time, arp_limit
 *     cdef float rfperiod
 *     cdef float ssample = 0.0, sample = 0.0, fp, pp             # <<<<<<<<<<<<<<
 *     cdef int si, ni
 * 
 */
  __pyx_v_ssample = 0.0;
  __pyx_v_sample = 0.0;

  /* "_pyfxr.pyx":430
 *     cdef int si, ni
 * 
 *     reset_sample(             # <<<<<<<<<<<<<<
 *         p_base_freq,
 *         p_freq_limit,
 */
  __pyx_f_6_pyfxr_reset_sample(__pyx_v_p_base_freq, __pyx_v_p_freq_limit, __pyx_v_p_freq_ramp, __pyx_v_p_freq_dramp, __pyx_v_p_duty, __pyx_v_p_duty_ramp, __pyx_v_p_arp_mod, __pyx_v_p_arp_speed, (&__pyx_v_fperiod), (&__pyx_v_period), (&__pyx_v_fmaxperiod), (&__pyx_v_fslide), (&__pyx_v_fdslide), (&__pyx_v_square_duty), (&__pyx_v_square_slide), (&__pyx_v_arp_mod), (&__pyx_v_arp_time), (&__pyx_v_arp_limit));

  /* "_pyfxr.pyx":453
 * 
 *     # reset filter
 *     fltw = 0.1 * p_lpf_freq ** 3.0             # <<<<<<<<<<<<<<
 *     fltw_d = 1.0 + p_lpf_ramp * 0.0001
 *     fltdmp = 5.0 / (1.0 + p_lpf_resonance ** 2.0 * 20.0) * (0.01 + fltw)
 */
  __pyx_v_fltw = (0.1 * pow(((double)__pyx_v_p_lpf_freq), 3.0));

  /* "_pyfxr.pyx":454
 *     # reset filter
 *     fltw = 0.1 * p_lpf_freq ** 3.0
 *     fltw_d = 1.0 + p_lpf_ramp * 0.0001             # <<<<<<<<<<<<<<
 *     fltdmp = 5.0 / (1.0 + p_lpf_resonance ** 2.0 * 20.0) * (0.01 + fltw)
 *     clamp(&fltdmp, 0.0, 0.8)
 */
  __pyx_v_fltw_d = (1.0 + (__pyx_v_p_lpf_ramp * 0.0001));

  /* "_pyfxr.pyx":455
 *     fltw = 0.1 * p_lpf_freq ** 3.0
 *     fltw_d = 1.0 + p_lpf_ramp * 0.0001
 *     fltdmp = 5.0 / (1.0 + p_lpf_resonance ** 2.0 * 20.0) * (0.01 + fltw)             # <<<<<<<<<<<<<<
 *     clamp(&fltdmp, 0.0, 0.8)
 * 
 */
  __pyx_v_fltdmp = ((5.0 / (1.0 + (pow(((double)__pyx_v_p_lpf_resonance), 2.0) * 20.0))) * (0.01 + __pyx_v_fltw));

  /* "_pyfxr.pyx":456
 *     fltw_d = 1.0 + p_lpf_ramp * 0.0001
 *     fltdmp = 5.0 / (1.0 + p_lpf_resonance ** 2.0 * 20.0) * (0.01 + fltw)
 *     clamp(&fltdmp, 0.0, 0.8)             # <<<<<<<<<<<<<<
 * 
 *     flthp = 0.1 * p_hpf_freq ** 2.0
 */
  __pyx_fuse_0__pyx_f_6_pyfxr_clamp((&__pyx_v_fltdmp), 0.0, 0.8);

  /* "_pyfxr.pyx":458
 *     clamp(&fltdmp, 0.0, 0.8)
 * 
 *     flthp = 0.1 * p_hpf_freq ** 2.0             # <<<<<<<<<<<<<<
 *     flthp_d = 1.0 + p_hpf_ramp * 0.0003;
 * 
 */
  __pyx_v_flthp = (0.1 * pow(((double)__pyx_v_p_hpf_freq), 2.0));

  /* "_pyfxr.pyx":459
 * 
 *     flthp = 0.1 * p_hpf_freq ** 2.0
 *     flthp_d = 1.0 + p_hpf_ramp * 0.0003;             # <<<<<<<<<<<<<<
 * 
 *     # reset vibrato
 */
  __pyx_v_flthp_d = (1.0 + (__pyx_v_p_hpf_ramp * 0.0003));

  /* "_pyfxr.pyx":462
 * 
 *     # reset vibrato
 *     vib_phase = 0.0;             # <<<<<<<<<<<<<<
 *     vib_speed = p_vib_speed ** 2.0 * 0.01;
 *     vib_amp = p_vib_strength * 0.5;
 */
  __pyx_v_vib_phase = 0.0;

  /* "_pyfxr.pyx":463
 *     # reset vibrato
 *     vib_phase = 0.0;
 *     vib_speed = p_vib_speed ** 2.0 * 0.01;             # <<<<<<<<<<<<<<
 *     vib_amp = p_vib_strength * 0.5;
 * 
 */
  __pyx_v_vib_speed = (pow(((double)__pyx_v_p_vib_speed), 2.0) * 0.01);

  /* "_pyfxr.pyx":464
 *     vib_phase = 0.0;
 *     vib_speed = p_vib_speed ** 2.0 * 0.01;
 *     vib_amp = p_vib_strength * 0.5;             # <<<<<<<<<<<<<<
 * 
 *     # reset envelope
 */
  __pyx_v_vib_amp = (__pyx_v_p_vib_strength * 0.5);

  /* "_pyfxr.pyx":467
 * 
 *     # reset envelope
 *     env_vol = 0.0;             # <<<<<<<<<<<<<<
 *     env_stage = 0;
 *     env_time = 0;
 */
  __pyx_v_env_vol = 0.0;

  /* "_pyfxr.pyx":468
 *     # reset envelope
 *     env_vol = 0.0;
 *     env_stage = 0;             # <<<<<<<<<<<<<<
 *     env_time = 0;
 *     env_length[0] = <int> (p_env_attack * p_env_attack * 100000.0);
 */
  __pyx_v_env_stage = 0;

  /* "_pyfxr.pyx":469
 *     env_vol = 0.0;
 *     env_stage = 0;
 *     env_time = 0;             # <<<<<<<<<<<<<<
 *     env_length[0] = <int> (p_env_attack * p_env_attack * 100000.0);
 *     env_length[1] = <int> (p_env_sustain * p_env_sustain * 100000.0);
 */
  __pyx_v_env_time = 0;

  /* "_pyfxr.pyx":470
 *     env_stage = 0;
 *     env_time = 0;
 *     env_length[0] = <int> (p_env_attack * p_env_attack * 100000.0);             # <<<<<<<<<<<<<<
 *     env_length[1] = <int> (p_env_sustain * p_env_sustain * 100000.0);
 *     env_length[2] = <int> (p_env_decay * p_env_decay * 100000.0);
 */
  (__pyx_v_env_length[0]) = ((int)((__pyx_v_p_env_attack * __pyx_v_p_env_attack) * 100000.0));

  /* "_pyfxr.pyx":471
 *     env_time = 0;
 *     env_length[0] = <int> (p_env_attack * p_env_attack * 100000.0);
 *     env_length[1] = <int> (p_env_sustain * p_env_sustain * 100000.0);             # <<<<<<<<<<<<<<
 *     env_length[2] = <int> (p_env_decay * p_env_decay * 100000.0);
 * 
 */
  (__pyx_v_env_length[1]) = ((int)((__pyx_v_p_env_sustain * __pyx_v_p_env_sustain) * 100000.0));

  /* "_pyfxr.pyx":472
 *     env_length[0] = <int> (p_env_attack * p_env_attack * 100000.0);
 *     env_length[1] = <int> (p_env_sustain * p_env_sustain * 100000.0);
 *     env_length[2] = <int> (p_env_decay * p_env_decay * 100000.0);             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t n_samples = env_length[0] + env_length[1] + env_length[2];
 */
  (__pyx_v_env_length[2]) = ((int)((__pyx_v_p_env_decay * __pyx_v_p_env_decay) * 100000.0));

  /* "_pyfxr.pyx":474
 *     env_length[2] = <int> (p_env_decay * p_env_decay * 100000.0);
 * 
 *     cdef size_t n_samples = env_length[0] + env_length[1] + env_length[2];             # <<<<<<<<<<<<<<
 *     cdef SoundBuffer s = SoundBuffer(n_samples)
 * 
 */
  __pyx_v_n_samples = (((__pyx_v_env_length[0]) + (__pyx_v_env_length[1])) + (__pyx_v_env_length[2]));

  /* "_pyfxr.pyx":475
 * 
 *     cdef size_t n_samples = env_length[0] + env_length[1] + env_length[2];
 *     cdef SoundBuffer s = SoundBuffer(n_samples)             # <<<<<<<<<<<<<<
 * 
 *     with nogil:
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_6_pyfxr_SoundBuffer), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "_pyfxr.pyx":477
 *     cdef SoundBuffer s = SoundBuffer(n_samples)
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         fphase = p_pha_offset ** 2.0 * 1020.0;
 *         if p_pha_offset < 0.0:
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "_pyfxr.pyx":478
 * 
 *     with nogil:
 *         fphase = p_pha_offset ** 2.0 * 1020.0;             # <<<<<<<<<<<<<<
 *         if p_pha_offset < 0.0:
 *             fphase = -fphase
 */
        __pyx_v_fphase = (pow(((double)__pyx_v_p_pha_offset), 2.0) * 1020.0);

        /* "_pyfxr.pyx":479
 *     with nogil:
 *         fphase = p_pha_offset ** 2.0 * 1020.0;
 *         if p_pha_offset < 0.0:             # <<<<<<<<<<<<<<
 *             fphase = -fphase
 *         fdphase = p_pha_ramp ** 2.0
 */
        __pyx_t_3 = ((__pyx_v_p_pha_offset < 0.0) != 0);
        if (__pyx_t_3) {

          /* "_pyfxr.pyx":480
 *         fphase = p_pha_offset ** 2.0 * 1020.0;
 *         if p_pha_offset < 0.0:
 *             fphase = -fphase             # <<<<<<<<<<<<<<
 *         fdphase = p_pha_ramp ** 2.0
 *         if p_pha_ramp < 0.0:
 */
          __pyx_v_fphase = (-__pyx_v_fphase);

          /* "_pyfxr.pyx":479
 *     with nogil:
 *         fphase = p_pha_offset ** 2.0 * 1020.0;
 *         if p_pha_offset < 0.0:             # <<<<<<<<<<<<<<
 *             fphase = -fphase
 *         fdphase = p_pha_ramp ** 2.0
 */
        }

        /* "_pyfxr.pyx":481
 *         if p_pha_offset < 0.0:
 *             fphase = -fphase
 *         fdphase = p_pha_ramp ** 2.0             # <<<<<<<<<<<<<<
 *         if p_pha_ramp < 0.0:
 *             fdphase = -fdphase
 */
        __pyx_v_fdphase = pow(((double)__pyx_v_p_pha_ramp), 2.0);

        /* "_pyfxr.pyx":482
 *             fphase = -fphase
 *         fdphase = p_pha_ramp ** 2.0
 *         if p_pha_ramp < 0.0:             # <<<<<<<<<<<<<<
 *             fdphase = -fdphase
 *         iphase = abs(<int> fphase)
 */
        __pyx_t_3 = ((__pyx_v_p_pha_ramp < 0.0) != 0);
        if (__pyx_t_3) {

          /* "_pyfxr.pyx":483
 *         fdphase = p_pha_ramp ** 2.0
 *         if p_pha_ramp < 0.0:
 *             fdphase = -fdphase             # <<<<<<<<<<<<<<
 *         iphase = abs(<int> fphase)
 *         ipp = 0
 */
          __pyx_v_fdphase = (-__pyx_v_fdphase);

          /* "_pyfxr.pyx":482
 *             fphase = -fphase
 *         fdphase = p_pha_ramp ** 2.0
 *         if p_pha_ramp < 0.0:             # <<<<<<<<<<<<<<
 *             fdphase = -fdphase
 *         iphase = abs(<int> fphase)
 */
        }

        /* "_pyfxr.pyx":484
 *         if p_pha_ramp < 0.0:
 *             fdphase = -fdphase
 *         iphase = abs(<int> fphase)             # <<<<<<<<<<<<<<
 *         ipp = 0
 *         for i in range(1024):
 */
        __pyx_v_iphase = abs(((int)__pyx_v_fphase));

        /* "_pyfxr.pyx":485
 *             fdphase = -fdphase
 *         iphase = abs(<int> fphase)
 *         ipp = 0             # <<<<<<<<<<<<<<
 *         for i in range(1024):
 *             phaser_buffer[i] = 0.0
 */
        __pyx_v_ipp = 0;

        /* "_pyfxr.pyx":486
 *         iphase = abs(<int> fphase)
 *         ipp = 0
 *         for i in range(1024):             # <<<<<<<<<<<<<<
 *             phaser_buffer[i] = 0.0
 * 
 */
        for (__pyx_t_4 = 0; __pyx_t_4 < 0x400; __pyx_t_4+=1) {
          __pyx_v_i = __pyx_t_4;

          /* "_pyfxr.pyx":487
 *         ipp = 0
 *         for i in range(1024):
 *             phaser_buffer[i] = 0.0             # <<<<<<<<<<<<<<
 * 
 *         # Fill noise buffer
 */
          (__pyx_v_phaser_buffer[__pyx_v_i]) = 0.0;
        }

        /* "_pyfxr.pyx":490
 * 
 *         # Fill noise buffer
 *         fill_noise(noise_buffer)             # <<<<<<<<<<<<<<
 * 
 *         # reset repeats
 */
        __pyx_f_6_pyfxr_fill_noise(__pyx_v_noise_buffer);

        /* "_pyfxr.pyx":493
 * 
 *         # reset repeats
 *         rep_time = 0             # <<<<<<<<<<<<<<
 *         rep_limit = <int> ((1.0 - p_repeat_speed) ** 2.0) * 20000 + 32
 *         if p_repeat_speed == 0.0:
 */
        __pyx_v_rep_time = 0;

        /* "_pyfxr.pyx":494
 *         # reset repeats
 *         rep_time = 0
 *         rep_limit = <int> ((1.0 - p_repeat_speed) ** 2.0) * 20000 + 32             # <<<<<<<<<<<<<<
 *         if p_repeat_speed == 0.0:
 *             rep_limit = 0
 */
        __pyx_v_rep_limit = ((((int)pow((1.0 - __pyx_v_p_repeat_speed), 2.0)) * 0x4E20) + 32);

        /* "_pyfxr.pyx":495
 *         rep_time = 0
 *         rep_limit = <int> ((1.0 - p_repeat_speed) ** 2.0) * 20000 + 32
 *         if p_repeat_speed == 0.0:             # <<<<<<<<<<<<<<
 *             rep_limit = 0
 * 
 */
        __pyx_t_3 = ((__pyx_v_p_repeat_speed == 0.0) != 0);
        if (__pyx_t_3) {

          /* "_pyfxr.pyx":496
 *         rep_limit = <int> ((1.0 - p_repeat_speed) ** 2.0) * 20000 + 32
 *         if p_repeat_speed == 0.0:
 *             rep_limit = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n_samples):
 */
          __pyx_v_rep_limit = 0;

          /* "_pyfxr.pyx":495
 *         rep_time = 0
 *         rep_limit = <int> ((1.0 - p_repeat_speed) ** 2.0) * 20000 + 32
 *         if p_repeat_speed == 0.0:             # <<<<<<<<<<<<<<
 *             rep_limit = 0
 * 
 */
        }

        /* "_pyfxr.pyx":498
 *             rep_limit = 0
 * 
 *         for i in range(n_samples):             # <<<<<<<<<<<<<<
 *             rep_time += 1
 * 
 */
        __pyx_t_4 = __pyx_v_n_samples;
        __pyx_t_5 = __pyx_t_4;
        for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
          __pyx_v_i = __pyx_t_6;

          /* "_pyfxr.pyx":499
 * 
 *         for i in range(n_samples):
 *             rep_time += 1             # <<<<<<<<<<<<<<
 * 
 *             if rep_limit and rep_time >= rep_limit:
 */
          __pyx_v_rep_time = (__pyx_v_rep_time + 1);

          /* "_pyfxr.pyx":501
 *             rep_time += 1
 * 
 *             if rep_limit and rep_time >= rep_limit:             # <<<<<<<<<<<<<<
 *                 rep_time = 0
 *                 reset_sample(
 */
          __pyx_t_7 = (__pyx_v_rep_limit != 0);
          if (__pyx_t_7) {
          } else {
            __pyx_t_3 = __pyx_t_7;
            goto __pyx_L14_bool_binop_done;
          }
          __pyx_t_7 = ((__pyx_v_rep_time >= __pyx_v_rep_limit) != 0);
          __pyx_t_3 = __pyx_t_7;
          __pyx_L14_bool_binop_done:;
          if (__pyx_t_3) {

            /* "_pyfxr.pyx":502
 * 
 *             if rep_limit and rep_time >= rep_limit:
 *                 rep_time = 0             # <<<<<<<<<<<<<<
 *                 reset_sample(
 *                     p_base_freq,
 */
            __pyx_v_rep_time = 0;

            /* "_pyfxr.pyx":503
 *             if rep_limit and rep_time >= rep_limit:
 *                 rep_time = 0
 *                 reset_sample(             # <<<<<<<<<<<<<<
 *                     p_base_freq,
 *                     p_freq_limit,
 */
            __pyx_f_6_pyfxr_reset_sample(__pyx_v_p_base_freq, __pyx_v_p_freq_limit, __pyx_v_p_freq_ramp, __pyx_v_p_freq_dramp, __pyx_v_p_duty, __pyx_v_p_duty_ramp, __pyx_v_p_arp_mod, __pyx_v_p_arp_speed, (&__pyx_v_fperiod), (&__pyx_v_period), (&__pyx_v_fmaxperiod), (&__pyx_v_fslide), (&__pyx_v_fdslide), (&__pyx_v_square_duty), (&__pyx_v_square_slide), (&__pyx_v_arp_mod), (&__pyx_v_arp_time), (&__pyx_v_arp_limit));

            /* "_pyfxr.pyx":501
 *             rep_time += 1
 * 
 *             if rep_limit and rep_time >= rep_limit:             # <<<<<<<<<<<<<<
 *                 rep_time = 0
 *                 reset_sample(
 */
          }

          /* "_pyfxr.pyx":526
 * 
 *             # frequency envelopes/arpeggios
 *             arp_time += 1             # <<<<<<<<<<<<<<
 *             if 0 != arp_limit < arp_time:
 *                 arp_limit = 0
 */
          __pyx_v_arp_time = (__pyx_v_arp_time + 1);

          /* "_pyfxr.pyx":527
 *             # frequency envelopes/arpeggios
 *             arp_time += 1
 *             if 0 != arp_limit < arp_time:             # <<<<<<<<<<<<<<
 *                 arp_limit = 0
 *                 fperiod *= arp_mod
 */
          __pyx_t_3 = (0 != __pyx_v_arp_limit);
          if (__pyx_t_3) {
            __pyx_t_3 = (__pyx_v_arp_limit < __pyx_v_arp_time);
          }
          __pyx_t_7 = (__pyx_t_3 != 0);
          if (__pyx_t_7) {

            /* "_pyfxr.pyx":528
 *             arp_time += 1
 *             if 0 != arp_limit < arp_time:
 *                 arp_limit = 0             # <<<<<<<<<<<<<<
 *                 fperiod *= arp_mod
 * 
 */
            __pyx_v_arp_limit = 0;

            /* "_pyfxr.pyx":529
 *             if 0 != arp_limit < arp_time:
 *                 arp_limit = 0
 *                 fperiod *= arp_mod             # <<<<<<<<<<<<<<
 * 
 *             fslide += fdslide
 */
            __pyx_v_fperiod = (__pyx_v_fperiod * __pyx_v_arp_mod);

            /* "_pyfxr.pyx":527
 *             # frequency envelopes/arpeggios
 *             arp_time += 1
 *             if 0 != arp_limit < arp_time:             # <<<<<<<<<<<<<<
 *                 arp_limit = 0
 *                 fperiod *= arp_mod
 */
          }

          /* "_pyfxr.pyx":531
 *                 fperiod *= arp_mod
 * 
 *             fslide += fdslide             # <<<<<<<<<<<<<<
 *             fperiod *= fslide
 *             if fperiod > fmaxperiod:
 */
          __pyx_v_fslide = (__pyx_v_fslide + __pyx_v_fdslide);

          /* "_pyfxr.pyx":532
 * 
 *             fslide += fdslide
 *             fperiod *= fslide             # <<<<<<<<<<<<<<
 *             if fperiod > fmaxperiod:
 *                 fperiod = fmaxperiod
 */
          __pyx_v_fperiod = (__pyx_v_fperiod * __pyx_v_fslide);

          /* "_pyfxr.pyx":533
 *             fslide += fdslide
 *             fperiod *= fslide
 *             if fperiod > fmaxperiod:             # <<<<<<<<<<<<<<
 *                 fperiod = fmaxperiod
 *                 if p_freq_limit > 0.0:
 */
          __pyx_t_7 = ((__pyx_v_fperiod > __pyx_v_fmaxperiod) != 0);
          if (__pyx_t_7) {

            /* "_pyfxr.pyx":534
 *             fperiod *= fslide
 *             if fperiod > fmaxperiod:
 *                 fperiod = fmaxperiod             # <<<<<<<<<<<<<<
 *                 if p_freq_limit > 0.0:
 *                     break
 */
            __pyx_v_fperiod = __pyx_v_fmaxperiod;

            /* "_pyfxr.pyx":535
 *             if fperiod > fmaxperiod:
 *                 fperiod = fmaxperiod
 *                 if p_freq_limit > 0.0:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
            __pyx_t_7 = ((__pyx_v_p_freq_limit > 0.0) != 0);
            if (__pyx_t_7) {

              /* "_pyfxr.pyx":536
 *                 fperiod = fmaxperiod
 *                 if p_freq_limit > 0.0:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             rfperiod = fperiod
 */
              goto __pyx_L12_break;

              /* "_pyfxr.pyx":535
 *             if fperiod > fmaxperiod:
 *                 fperiod = fmaxperiod
 *                 if p_freq_limit > 0.0:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
            }

            /* "_pyfxr.pyx":533
 *             fslide += fdslide
 *             fperiod *= fslide
 *             if fperiod > fmaxperiod:             # <<<<<<<<<<<<<<
 *                 fperiod = fmaxperiod
 *                 if p_freq_limit > 0.0:
 */
          }

          /* "_pyfxr.pyx":538
 *                     break
 * 
 *             rfperiod = fperiod             # <<<<<<<<<<<<<<
 *             if vib_amp > 0.0:
 *                 vib_phase += vib_speed
 */
          __pyx_v_rfperiod = __pyx_v_fperiod;

          /* "_pyfxr.pyx":539
 * 
 *             rfperiod = fperiod
 *             if vib_amp > 0.0:             # <<<<<<<<<<<<<<
 *                 vib_phase += vib_speed
 *                 rfperiod = fperiod * (1.0 + sin(vib_phase) * vib_amp)
 */
          __pyx_t_7 = ((__pyx_v_vib_amp > 0.0) != 0);
          if (__pyx_t_7) {

            /* "_pyfxr.pyx":540
 *             rfperiod = fperiod
 *             if vib_amp > 0.0:
 *                 vib_phase += vib_speed             # <<<<<<<<<<<<<<
 *                 rfperiod = fperiod * (1.0 + sin(vib_phase) * vib_amp)
 * 
 */
            __pyx_v_vib_phase = (__pyx_v_vib_phase + __pyx_v_vib_speed);

            /* "_pyfxr.pyx":541
 *             if vib_amp > 0.0:
 *                 vib_phase += vib_speed
 *                 rfperiod = fperiod * (1.0 + sin(vib_phase) * vib_amp)             # <<<<<<<<<<<<<<
 * 
 *             period = <int> rfperiod
 */
            __pyx_v_rfperiod = (__pyx_v_fperiod * (1.0 + (sin(__pyx_v_vib_phase) * __pyx_v_vib_amp)));

            /* "_pyfxr.pyx":539
 * 
 *             rfperiod = fperiod
 *             if vib_amp > 0.0:             # <<<<<<<<<<<<<<
 *                 vib_phase += vib_speed
 *                 rfperiod = fperiod * (1.0 + sin(vib_phase) * vib_amp)
 */
          }

          /* "_pyfxr.pyx":543
 *                 rfperiod = fperiod * (1.0 + sin(vib_phase) * vib_amp)
 * 
 *             period = <int> rfperiod             # <<<<<<<<<<<<<<
 *             if period < 8:
 *                 period = 8
 */
          __pyx_v_period = ((int)__pyx_v_rfperiod);

          /* "_pyfxr.pyx":544
 * 
 *             period = <int> rfperiod
 *             if period < 8:             # <<<<<<<<<<<<<<
 *                 period = 8
 *             square_duty += square_slide
 */
          __pyx_t_7 = ((__pyx_v_period < 8) != 0);
          if (__pyx_t_7) {

            /* "_pyfxr.pyx":545
 *             period = <int> rfperiod
 *             if period < 8:
 *                 period = 8             # <<<<<<<<<<<<<<
 *             square_duty += square_slide
 *             clamp(&square_duty, 0.0, 0.5)
 */
            __pyx_v_period = 8;

            /* "_pyfxr.pyx":544
 * 
 *             period = <int> rfperiod
 *             if period < 8:             # <<<<<<<<<<<<<<
 *                 period = 8
 *             square_duty += square_slide
 */
          }

          /* "_pyfxr.pyx":546
 *             if period < 8:
 *                 period = 8
 *             square_duty += square_slide             # <<<<<<<<<<<<<<
 *             clamp(&square_duty, 0.0, 0.5)
 * 
 */
          __pyx_v_square_duty = (__pyx_v_square_duty + __pyx_v_square_slide);

          /* "_pyfxr.pyx":547
 *                 period = 8
 *             square_duty += square_slide
 *             clamp(&square_duty, 0.0, 0.5)             # <<<<<<<<<<<<<<
 * 
 *             # volume envelope
 */
          __pyx_fuse_0__pyx_f_6_pyfxr_clamp((&__pyx_v_square_duty), 0.0, 0.5);

          /* "_pyfxr.pyx":550
 * 
 *             # volume envelope
 *             env_time += 1             # <<<<<<<<<<<<<<
 *             if env_time > env_length[env_stage]:
 *                 env_time = 0
 */
          __pyx_v_env_time = (__pyx_v_env_time + 1);

          /* "_pyfxr.pyx":551
 *             # volume envelope
 *             env_time += 1
 *             if env_time > env_length[env_stage]:             # <<<<<<<<<<<<<<
 *                 env_time = 0
 *                 env_stage += 1
 */
          __pyx_t_7 = ((__pyx_v_env_time > (__pyx_v_env_length[__pyx_v_env_stage])) != 0);
          if (__pyx_t_7) {

            /* "_pyfxr.pyx":552
 *             env_time += 1
 *             if env_time > env_length[env_stage]:
 *                 env_time = 0             # <<<<<<<<<<<<<<
 *                 env_stage += 1
 *                 if env_stage == 3:
 */
            __pyx_v_env_time = 0;

            /* "_pyfxr.pyx":553
 *             if env_time > env_length[env_stage]:
 *                 env_time = 0
 *                 env_stage += 1             # <<<<<<<<<<<<<<
 *                 if env_stage == 3:
 *                     break
 */
            __pyx_v_env_stage = (__pyx_v_env_stage + 1);

            /* "_pyfxr.pyx":554
 *                 env_time = 0
 *                 env_stage += 1
 *                 if env_stage == 3:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
            __pyx_t_7 = ((__pyx_v_env_stage == 3) != 0);
            if (__pyx_t_7) {

              /* "_pyfxr.pyx":555
 *                 env_stage += 1
 *                 if env_stage == 3:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             if env_stage == 0:
 */
              goto __pyx_L12_break;

              /* "_pyfxr.pyx":554
 *                 env_time = 0
 *                 env_stage += 1
 *                 if env_stage == 3:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
            }

            /* "_pyfxr.pyx":551
 *             # volume envelope
 *             env_time += 1
 *             if env_time > env_length[env_stage]:             # <<<<<<<<<<<<<<
 *                 env_time = 0
 *                 env_stage += 1
 */
          }

          /* "_pyfxr.pyx":557
 *                     break
 * 
 *             if env_stage == 0:             # <<<<<<<<<<<<<<
 *                 env_vol = <float> env_time / env_length[0]
 *             elif env_stage == 1:
 */
          switch (__pyx_v_env_stage) {
            case 0:

            /* "_pyfxr.pyx":558
 * 
 *             if env_stage == 0:
 *                 env_vol = <float> env_time / env_length[0]             # <<<<<<<<<<<<<<
 *             elif env_stage == 1:
 *                 # TODO: removed a pow(1.0 - x / y, 1.0) here. Why?
 */
            __pyx_v_env_vol = (((float)__pyx_v_env_time) / ((float)(__pyx_v_env_length[0])));

            /* "_pyfxr.pyx":557
 *                     break
 * 
 *             if env_stage == 0:             # <<<<<<<<<<<<<<
 *                 env_vol = <float> env_time / env_length[0]
 *             elif env_stage == 1:
 */
            break;
            case 1:

            /* "_pyfxr.pyx":561
 *             elif env_stage == 1:
 *                 # TODO: removed a pow(1.0 - x / y, 1.0) here. Why?
 *                 env_vol = 1.0 + (1.0 - <float> env_time / env_length[1]) * 2.0 * p_env_punch             # <<<<<<<<<<<<<<
 *             elif env_stage == 2:
 *                 env_vol = 1.0 - <float> env_time / env_length[2]
 */
            __pyx_v_env_vol = (1.0 + (((1.0 - (((float)__pyx_v_env_time) / ((float)(__pyx_v_env_length[1])))) * 2.0) * __pyx_v_p_env_punch));

            /* "_pyfxr.pyx":559
 *             if env_stage == 0:
 *                 env_vol = <float> env_time / env_length[0]
 *             elif env_stage == 1:             # <<<<<<<<<<<<<<
 *                 # TODO: removed a pow(1.0 - x / y, 1.0) here. Why?
 *                 env_vol = 1.0 + (1.0 - <float> env_time / env_length[1]) * 2.0 * p_env_punch
 */
            break;
            case 2:

            /* "_pyfxr.pyx":563
 *                 env_vol = 1.0 + (1.0 - <float> env_time / env_length[1]) * 2.0 * p_env_punch
 *             elif env_stage == 2:
 *                 env_vol = 1.0 - <float> env_time / env_length[2]             # <<<<<<<<<<<<<<
 * 
 *             # phaser step
 */
            __pyx_v_env_vol = (1.0 - (((float)__pyx_v_env_time) / ((float)(__pyx_v_env_length[2]))));

            /* "_pyfxr.pyx":562
 *                 # TODO: removed a pow(1.0 - x / y, 1.0) here. Why?
 *                 env_vol = 1.0 + (1.0 - <float> env_time / env_length[1]) * 2.0 * p_env_punch
 *             elif env_stage == 2:             # <<<<<<<<<<<<<<
 *                 env_vol = 1.0 - <float> env_time / env_length[2]
 * 
 */
            break;
            default: break;
          }

          /* "_pyfxr.pyx":566
 * 
 *             # phaser step
 *             fphase += fdphase;             # <<<<<<<<<<<<<<
 *             iphase = abs(<int> fphase)
 *             if iphase > 1023:
 */
          __pyx_v_fphase = (__pyx_v_fphase + __pyx_v_fdphase);

          /* "_pyfxr.pyx":567
 *             # phaser step
 *             fphase += fdphase;
 *             iphase = abs(<int> fphase)             # <<<<<<<<<<<<<<
 *             if iphase > 1023:
 *                 iphase = 1023
 */
          __pyx_v_iphase = abs(((int)__pyx_v_fphase));

          /* "_pyfxr.pyx":568
 *             fphase += fdphase;
 *             iphase = abs(<int> fphase)
 *             if iphase > 1023:             # <<<<<<<<<<<<<<
 *                 iphase = 1023
 * 
 */
          __pyx_t_7 = ((__pyx_v_iphase > 0x3FF) != 0);
          if (__pyx_t_7) {

            /* "_pyfxr.pyx":569
 *             iphase = abs(<int> fphase)
 *             if iphase > 1023:
 *                 iphase = 1023             # <<<<<<<<<<<<<<
 * 
 *             if flthp_d != 0.0:
 */
            __pyx_v_iphase = 0x3FF;

            /* "_pyfxr.pyx":568
 *             fphase += fdphase;
 *             iphase = abs(<int> fphase)
 *             if iphase > 1023:             # <<<<<<<<<<<<<<
 *                 iphase = 1023
 * 
 */
          }

          /* "_pyfxr.pyx":571
 *                 iphase = 1023
 * 
 *             if flthp_d != 0.0:             # <<<<<<<<<<<<<<
 *                 flthp *= flthp_d
 *                 clamp(&flthp, 0.00001, 0.1)
 */
          __pyx_t_7 = ((__pyx_v_flthp_d != 0.0) != 0);
          if (__pyx_t_7) {

            /* "_pyfxr.pyx":572
 * 
 *             if flthp_d != 0.0:
 *                 flthp *= flthp_d             # <<<<<<<<<<<<<<
 *                 clamp(&flthp, 0.00001, 0.1)
 * 
 */
            __pyx_v_flthp = (__pyx_v_flthp * __pyx_v_flthp_d);

            /* "_pyfxr.pyx":573
 *             if flthp_d != 0.0:
 *                 flthp *= flthp_d
 *                 clamp(&flthp, 0.00001, 0.1)             # <<<<<<<<<<<<<<
 * 
 *             ssample = 0.0
 */
            __pyx_fuse_0__pyx_f_6_pyfxr_clamp((&__pyx_v_flthp), 0.00001, 0.1);

            /* "_pyfxr.pyx":571
 *                 iphase = 1023
 * 
 *             if flthp_d != 0.0:             # <<<<<<<<<<<<<<
 *                 flthp *= flthp_d
 *                 clamp(&flthp, 0.00001, 0.1)
 */
          }

          /* "_pyfxr.pyx":575
 *                 clamp(&flthp, 0.00001, 0.1)
 * 
 *             ssample = 0.0             # <<<<<<<<<<<<<<
 *             for si in range(8):  # 8x supersampling
 *                 sample = 0.0
 */
          __pyx_v_ssample = 0.0;

          /* "_pyfxr.pyx":576
 * 
 *             ssample = 0.0
 *             for si in range(8):  # 8x supersampling             # <<<<<<<<<<<<<<
 *                 sample = 0.0
 *                 phase += 1
 */
          for (__pyx_t_8 = 0; __pyx_t_8 < 8; __pyx_t_8+=1) {
            __pyx_v_si = __pyx_t_8;

            /* "_pyfxr.pyx":577
 *             ssample = 0.0
 *             for si in range(8):  # 8x supersampling
 *                 sample = 0.0             # <<<<<<<<<<<<<<
 *                 phase += 1
 *                 if phase >= period:
 */
            __pyx_v_sample = 0.0;

            /* "_pyfxr.pyx":578
 *             for si in range(8):  # 8x supersampling
 *                 sample = 0.0
 *                 phase += 1             # <<<<<<<<<<<<<<
 *                 if phase >= period:
 *                     phase %= period
 */
            __pyx_v_phase = (__pyx_v_phase + 1);

            /* "_pyfxr.pyx":579
 *                 sample = 0.0
 *                 phase += 1
 *                 if phase >= period:             # <<<<<<<<<<<<<<
 *                     phase %= period
 *                     if wave_type == 3:
 */
            __pyx_t_7 = ((__pyx_v_phase >= __pyx_v_period) != 0);
            if (__pyx_t_7) {

              /* "_pyfxr.pyx":580
 *                 phase += 1
 *                 if phase >= period:
 *                     phase %= period             # <<<<<<<<<<<<<<
 *                     if wave_type == 3:
 *                         fill_noise(noise_buffer)
 */
              __pyx_v_phase = (__pyx_v_phase % __pyx_v_period);

              /* "_pyfxr.pyx":581
 *                 if phase >= period:
 *                     phase %= period
 *                     if wave_type == 3:             # <<<<<<<<<<<<<<
 *                         fill_noise(noise_buffer)
 * 
 */
              __pyx_t_7 = ((__pyx_v_wave_type == 3) != 0);
              if (__pyx_t_7) {

                /* "_pyfxr.pyx":582
 *                     phase %= period
 *                     if wave_type == 3:
 *                         fill_noise(noise_buffer)             # <<<<<<<<<<<<<<
 * 
 *                 # base waveform
 */
                __pyx_f_6_pyfxr_fill_noise(__pyx_v_noise_buffer);

                /* "_pyfxr.pyx":581
 *                 if phase >= period:
 *                     phase %= period
 *                     if wave_type == 3:             # <<<<<<<<<<<<<<
 *                         fill_noise(noise_buffer)
 * 
 */
              }

              /* "_pyfxr.pyx":579
 *                 sample = 0.0
 *                 phase += 1
 *                 if phase >= period:             # <<<<<<<<<<<<<<
 *                     phase %= period
 *                     if wave_type == 3:
 */
            }

            /* "_pyfxr.pyx":585
 * 
 *                 # base waveform
 *                 fp = <float> phase / period;             # <<<<<<<<<<<<<<
 *                 if wave_type == 0:  # square
 *                     sample = 0.5 if fp < square_duty else -0.5
 */
            __pyx_v_fp = (((float)__pyx_v_phase) / ((float)__pyx_v_period));

            /* "_pyfxr.pyx":586
 *                 # base waveform
 *                 fp = <float> phase / period;
 *                 if wave_type == 0:  # square             # <<<<<<<<<<<<<<
 *                     sample = 0.5 if fp < square_duty else -0.5
 *                 elif wave_type == 1:  # sawtooth
 */
            switch (__pyx_v_wave_type) {
              case 0:

              /* "_pyfxr.pyx":587
 *                 fp = <float> phase / period;
 *                 if wave_type == 0:  # square
 *                     sample = 0.5 if fp < square_duty else -0.5             # <<<<<<<<<<<<<<
 *                 elif wave_type == 1:  # sawtooth
 *                     sample = 1.0 - fp * 2
 */
              if (((__pyx_v_fp < __pyx_v_square_duty) != 0)) {
                __pyx_t_9 = 0.5;
              } else {
                __pyx_t_9 = -0.5;
              }
              __pyx_v_sample = __pyx_t_9;

              /* "_pyfxr.pyx":586
 *                 # base waveform
 *                 fp = <float> phase / period;
 *                 if wave_type == 0:  # square             # <<<<<<<<<<<<<<
 *                     sample = 0.5 if fp < square_duty else -0.5
 *                 elif wave_type == 1:  # sawtooth
 */
              break;
              case 1:

              /* "_pyfxr.pyx":589
 *                     sample = 0.5 if fp < square_duty else -0.5
 *                 elif wave_type == 1:  # sawtooth
 *                     sample = 1.0 - fp * 2             # <<<<<<<<<<<<<<
 *                 elif wave_type == 2:  # sine
 *                     sample = <float> sin(fp * 2 * pi);
 */
              __pyx_v_sample = (1.0 - (__pyx_v_fp * 2.0));

              /* "_pyfxr.pyx":588
 *                 if wave_type == 0:  # square
 *                     sample = 0.5 if fp < square_duty else -0.5
 *                 elif wave_type == 1:  # sawtooth             # <<<<<<<<<<<<<<
 *                     sample = 1.0 - fp * 2
 *                 elif wave_type == 2:  # sine
 */
              break;
              case 2:

              /* "_pyfxr.pyx":591
 *                     sample = 1.0 - fp * 2
 *                 elif wave_type == 2:  # sine
 *                     sample = <float> sin(fp * 2 * pi);             # <<<<<<<<<<<<<<
 *                 elif wave_type == 3:  # noise
 *                     sample = noise_buffer[<size_t> (phase * 32 / period)]
 */
              __pyx_v_sample = ((float)sin(((__pyx_v_fp * 2.0) * M_PI)));

              /* "_pyfxr.pyx":590
 *                 elif wave_type == 1:  # sawtooth
 *                     sample = 1.0 - fp * 2
 *                 elif wave_type == 2:  # sine             # <<<<<<<<<<<<<<
 *                     sample = <float> sin(fp * 2 * pi);
 *                 elif wave_type == 3:  # noise
 */
              break;
              case 3:

              /* "_pyfxr.pyx":593
 *                     sample = <float> sin(fp * 2 * pi);
 *                 elif wave_type == 3:  # noise
 *                     sample = noise_buffer[<size_t> (phase * 32 / period)]             # <<<<<<<<<<<<<<
 * 
 *                 # lp filter
 */
              __pyx_v_sample = (__pyx_v_noise_buffer[((size_t)((__pyx_v_phase * 32) / ((long)__pyx_v_period)))]);

              /* "_pyfxr.pyx":592
 *                 elif wave_type == 2:  # sine
 *                     sample = <float> sin(fp * 2 * pi);
 *                 elif wave_type == 3:  # noise             # <<<<<<<<<<<<<<
 *                     sample = noise_buffer[<size_t> (phase * 32 / period)]
 * 
 */
              break;
              default: break;
            }

            /* "_pyfxr.pyx":596
 * 
 *                 # lp filter
 *                 pp = fltp             # <<<<<<<<<<<<<<
 *                 fltw *= fltw_d
 *                 clamp(&fltw, 0.0, 0.1)
 */
            __pyx_v_pp = __pyx_v_fltp;

            /* "_pyfxr.pyx":597
 *                 # lp filter
 *                 pp = fltp
 *                 fltw *= fltw_d             # <<<<<<<<<<<<<<
 *                 clamp(&fltw, 0.0, 0.1)
 *                 if p_lpf_freq != 1.0:
 */
            __pyx_v_fltw = (__pyx_v_fltw * __pyx_v_fltw_d);

            /* "_pyfxr.pyx":598
 *                 pp = fltp
 *                 fltw *= fltw_d
 *                 clamp(&fltw, 0.0, 0.1)             # <<<<<<<<<<<<<<
 *                 if p_lpf_freq != 1.0:
 *                     fltdp += (sample - fltp) * fltw;
 */
            __pyx_fuse_0__pyx_f_6_pyfxr_clamp((&__pyx_v_fltw), 0.0, 0.1);

            /* "_pyfxr.pyx":599
 *                 fltw *= fltw_d
 *                 clamp(&fltw, 0.0, 0.1)
 *                 if p_lpf_freq != 1.0:             # <<<<<<<<<<<<<<
 *                     fltdp += (sample - fltp) * fltw;
 *                     fltdp -= fltdp * fltdmp;
 */
            __pyx_t_7 = ((__pyx_v_p_lpf_freq != 1.0) != 0);
            if (__pyx_t_7) {

              /* "_pyfxr.pyx":600
 *                 clamp(&fltw, 0.0, 0.1)
 *                 if p_lpf_freq != 1.0:
 *                     fltdp += (sample - fltp) * fltw;             # <<<<<<<<<<<<<<
 *                     fltdp -= fltdp * fltdmp;
 *                     fltp += fltdp
 */
              __pyx_v_fltdp = (__pyx_v_fltdp + ((__pyx_v_sample - __pyx_v_fltp) * __pyx_v_fltw));

              /* "_pyfxr.pyx":601
 *                 if p_lpf_freq != 1.0:
 *                     fltdp += (sample - fltp) * fltw;
 *                     fltdp -= fltdp * fltdmp;             # <<<<<<<<<<<<<<
 *                     fltp += fltdp
 *                 else:
 */
              __pyx_v_fltdp = (__pyx_v_fltdp - (__pyx_v_fltdp * __pyx_v_fltdmp));

              /* "_pyfxr.pyx":602
 *                     fltdp += (sample - fltp) * fltw;
 *                     fltdp -= fltdp * fltdmp;
 *                     fltp += fltdp             # <<<<<<<<<<<<<<
 *                 else:
 *                     fltp = sample
 */
              __pyx_v_fltp = (__pyx_v_fltp + __pyx_v_fltdp);

              /* "_pyfxr.pyx":599
 *                 fltw *= fltw_d
 *                 clamp(&fltw, 0.0, 0.1)
 *                 if p_lpf_freq != 1.0:             # <<<<<<<<<<<<<<
 *                     fltdp += (sample - fltp) * fltw;
 *                     fltdp -= fltdp * fltdmp;
 */
              goto __pyx_L29;
            }

            /* "_pyfxr.pyx":604
 *                     fltp += fltdp
 *                 else:
 *                     fltp = sample             # <<<<<<<<<<<<<<
 *                     fltdp = 0.0
 * 
 */
            /*else*/ {
              __pyx_v_fltp = __pyx_v_sample;

              /* "_pyfxr.pyx":605
 *                 else:
 *                     fltp = sample
 *                     fltdp = 0.0             # <<<<<<<<<<<<<<
 * 
 *                 # hp filter
 */
              __pyx_v_fltdp = 0.0;
            }
            __pyx_L29:;

            /* "_pyfxr.pyx":608
 * 
 *                 # hp filter
 *                 fltphp += fltp - pp;             # <<<<<<<<<<<<<<
 *                 fltphp -= fltphp * flthp;
 *                 sample = fltphp;
 */
            __pyx_v_fltphp = (__pyx_v_fltphp + (__pyx_v_fltp - __pyx_v_pp));

            /* "_pyfxr.pyx":609
 *                 # hp filter
 *                 fltphp += fltp - pp;
 *                 fltphp -= fltphp * flthp;             # <<<<<<<<<<<<<<
 *                 sample = fltphp;
 * 
 */
            __pyx_v_fltphp = (__pyx_v_fltphp - (__pyx_v_fltphp * __pyx_v_flthp));

            /* "_pyfxr.pyx":610
 *                 fltphp += fltp - pp;
 *                 fltphp -= fltphp * flthp;
 *                 sample = fltphp;             # <<<<<<<<<<<<<<
 * 
 *                 # phaser
 */
            __pyx_v_sample = __pyx_v_fltphp;

            /* "_pyfxr.pyx":613
 * 
 *                 # phaser
 *                 phaser_buffer[ipp & 1023] = sample             # <<<<<<<<<<<<<<
 *                 sample += phaser_buffer[(ipp - iphase + 1024) & 1023]
 *                 ipp = (ipp + 1) & 1023
 */
            (__pyx_v_phaser_buffer[(__pyx_v_ipp & 0x3FF)]) = __pyx_v_sample;

            /* "_pyfxr.pyx":614
 *                 # phaser
 *                 phaser_buffer[ipp & 1023] = sample
 *                 sample += phaser_buffer[(ipp - iphase + 1024) & 1023]             # <<<<<<<<<<<<<<
 *                 ipp = (ipp + 1) & 1023
 * 
 */
            __pyx_v_sample = (__pyx_v_sample + (__pyx_v_phaser_buffer[(((__pyx_v_ipp - __pyx_v_iphase) + 0x400) & 0x3FF)]));

            /* "_pyfxr.pyx":615
 *                 phaser_buffer[ipp & 1023] = sample
 *                 sample += phaser_buffer[(ipp - iphase + 1024) & 1023]
 *                 ipp = (ipp + 1) & 1023             # <<<<<<<<<<<<<<
 * 
 *                 # final accumulation and envelope application
 */
            __pyx_v_ipp = ((__pyx_v_ipp + 1) & 0x3FF);

            /* "_pyfxr.pyx":618
 * 
 *                 # final accumulation and envelope application
 *                 ssample += sample * env_vol             # <<<<<<<<<<<<<<
 * 
 *             ssample /= 8
 */
            __pyx_v_ssample = (__pyx_v_ssample + (__pyx_v_sample * __pyx_v_env_vol));
          }

          /* "_pyfxr.pyx":620
 *                 ssample += sample * env_vol
 * 
 *             ssample /= 8             # <<<<<<<<<<<<<<
 *             clamp(&ssample, -1.0, 1.0)
 *             s.samples[i] = samp(ssample)
 */
          __pyx_v_ssample = (__pyx_v_ssample / 8.0);

          /* "_pyfxr.pyx":621
 * 
 *             ssample /= 8
 *             clamp(&ssample, -1.0, 1.0)             # <<<<<<<<<<<<<<
 *             s.samples[i] = samp(ssample)
 *     return s
 */
          __pyx_fuse_0__pyx_f_6_pyfxr_clamp((&__pyx_v_ssample), -1.0, 1.0);

          /* "_pyfxr.pyx":622
 *             ssample /= 8
 *             clamp(&ssample, -1.0, 1.0)
 *             s.samples[i] = samp(ssample)             # <<<<<<<<<<<<<<
 *     return s
 * 
 */
          (__pyx_v_s->samples[__pyx_v_i]) = __pyx_f_6_pyfxr_samp(__pyx_v_ssample);
        }
        __pyx_L12_break:;
      }

      /* "_pyfxr.pyx":477
 *     cdef SoundBuffer s = SoundBuffer(n_samples)
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         fphase = p_pha_offset ** 2.0 * 1020.0;
 *         if p_pha_offset < 0.0:
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "_pyfxr.pyx":623
 *             clamp(&ssample, -1.0, 1.0)
 *             s.samples[i] = samp(ssample)
 *     return s             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_s));
  __pyx_r = ((PyObject *)__pyx_v_s);
  goto __pyx_L0;

  /* "_pyfxr.pyx":391
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * def sfx(             # <<<<<<<<<<<<<<
 *     int wave_type=0,
 *     float p_base_freq=0.3,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("_pyfxr.sfx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pyfxr.pyx":630
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * def pluck(float duration, float pitch, float release=0.1):             # <<<<<<<<<<<<<<
 *     """Generate a pluck sound using the Karplus-Strong algorithm."""
 *     cdef size_t delay, i, n_samples, release_samples, pos
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_5pluck(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_4pluck[] = "pluck(float duration, float pitch, float release=0.1)\nGenerate a pluck sound using the Karplus-Strong algorithm.";
static PyMethodDef __pyx_mdef_6_pyfxr_5pluck = {"pluck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_5pluck, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_4pluck};
static PyObject *__pyx_pw_6_pyfxr_5pluck(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  float __pyx_v_duration;
  float __pyx_v_pitch;
  float __pyx_v_release;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pluck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_duration,&__pyx_n_s_pitch,&__pyx_n_s_release,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_duration)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pitch)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pluck", 0, 2, 3, 1); __PYX_ERR(0, 630, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_release);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pluck") < 0)) __PYX_ERR(0, 630, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_duration = __pyx_PyFloat_AsFloat(values[0]); if (unlikely((__pyx_v_duration == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    __pyx_v_pitch = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_pitch == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_release = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_release == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    } else {
      __pyx_v_release = ((float)0.1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pluck", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 630, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.pluck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6_pyfxr_4pluck(__pyx_self, __pyx_v_duration, __pyx_v_pitch, __pyx_v_release);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_4pluck(CYTHON_UNUSED PyObject *__pyx_self, float __pyx_v_duration, float __pyx_v_pitch, float __pyx_v_release) {
  size_t __pyx_v_delay;
  size_t __pyx_v_i;
  size_t __pyx_v_n_samples;
  size_t __pyx_v_release_samples;
  size_t __pyx_v_pos;
  int16_t __pyx_v_randval;
  int16_t __pyx_v_prev;
  float __pyx_v_fsample;
  struct __pyx_obj_6_pyfxr_SoundBuffer *__pyx_v_s = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int16_t __pyx_t_10;
  long __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pluck", 0);

  /* "_pyfxr.pyx":635
 *     cdef int16_t randval, prev
 *     cdef float fsample
 *     n_samples = <size_t> (SAMPLE_RATE * duration)             # <<<<<<<<<<<<<<
 *     release_samples = min(<size_t> (SAMPLE_RATE * release), n_samples)
 * 
 */
  __pyx_v_n_samples = ((size_t)(__pyx_v_6_pyfxr_SAMPLE_RATE * __pyx_v_duration));

  /* "_pyfxr.pyx":636
 *     cdef float fsample
 *     n_samples = <size_t> (SAMPLE_RATE * duration)
 *     release_samples = min(<size_t> (SAMPLE_RATE * release), n_samples)             # <<<<<<<<<<<<<<
 * 
 *     delay = <size_t> (SAMPLE_RATE / pitch)
 */
  __pyx_t_1 = __pyx_v_n_samples;
  __pyx_t_2 = ((size_t)(__pyx_v_6_pyfxr_SAMPLE_RATE * __pyx_v_release));
  if (((__pyx_t_1 < __pyx_t_2) != 0)) {
    __pyx_t_3 = __pyx_t_1;
  } else {
    __pyx_t_3 = __pyx_t_2;
  }
  __pyx_v_release_samples = __pyx_t_3;

  /* "_pyfxr.pyx":638
 *     release_samples = min(<size_t> (SAMPLE_RATE * release), n_samples)
 * 
 *     delay = <size_t> (SAMPLE_RATE / pitch)             # <<<<<<<<<<<<<<
 *     if n_samples < delay:
 *         raise ValueError(
 */
  __pyx_v_delay = ((size_t)(((float)__pyx_v_6_pyfxr_SAMPLE_RATE) / __pyx_v_pitch));

  /* "_pyfxr.pyx":639
 * 
 *     delay = <size_t> (SAMPLE_RATE / pitch)
 *     if n_samples < delay:             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             f"n_samples must be at least {delay} for pitch {pitch}"
 */
  __pyx_t_4 = ((__pyx_v_n_samples < __pyx_v_delay) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "_pyfxr.pyx":641
 *     if n_samples < delay:
 *         raise ValueError(
 *             f"n_samples must be at least {delay} for pitch {pitch}"             # <<<<<<<<<<<<<<
 *         )
 * 
 */
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 0;
    __pyx_t_7 = 127;
    __Pyx_INCREF(__pyx_kp_u_n_samples_must_be_at_least);
    __pyx_t_6 += 27;
    __Pyx_GIVEREF(__pyx_kp_u_n_samples_must_be_at_least);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_n_samples_must_be_at_least);
    __pyx_t_8 = __Pyx_PyUnicode_From_size_t(__pyx_v_delay, 0, ' ', 'd'); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_for_pitch);
    __pyx_t_6 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_for_pitch);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u_for_pitch);
    __pyx_t_8 = PyFloat_FromDouble(__pyx_v_pitch); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_5, 4, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "_pyfxr.pyx":640
 *     delay = <size_t> (SAMPLE_RATE / pitch)
 *     if n_samples < delay:
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             f"n_samples must be at least {delay} for pitch {pitch}"
 *         )
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 640, __pyx_L1_error)

    /* "_pyfxr.pyx":639
 * 
 *     delay = <size_t> (SAMPLE_RATE / pitch)
 *     if n_samples < delay:             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             f"n_samples must be at least {delay} for pitch {pitch}"
 */
  }

  /* "_pyfxr.pyx":644
 *         )
 * 
 *     cdef SoundBuffer s = SoundBuffer(n_samples)             # <<<<<<<<<<<<<<
 * 
 *     with nogil:
 */
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_n_samples); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_6_pyfxr_SoundBuffer), __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_s = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "_pyfxr.pyx":646
 *     cdef SoundBuffer s = SoundBuffer(n_samples)
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         prev = 0
 *         for i in range(delay):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "_pyfxr.pyx":647
 * 
 *     with nogil:
 *         prev = 0             # <<<<<<<<<<<<<<
 *         for i in range(delay):
 *             randval = 1 << 15 if rand() % 2 else -(1 << 15)
 */
        __pyx_v_prev = 0;

        /* "_pyfxr.pyx":648
 *     with nogil:
 *         prev = 0
 *         for i in range(delay):             # <<<<<<<<<<<<<<
 *             randval = 1 << 15 if rand() % 2 else -(1 << 15)
 *             prev = s.samples[i] = (randval >> 1) + (prev >> 1)
 */
        __pyx_t_3 = __pyx_v_delay;
        __pyx_t_1 = __pyx_t_3;
        for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
          __pyx_v_i = __pyx_t_2;

          /* "_pyfxr.pyx":649
 *         prev = 0
 *         for i in range(delay):
 *             randval = 1 << 15 if rand() % 2 else -(1 << 15)             # <<<<<<<<<<<<<<
 *             prev = s.samples[i] = (randval >> 1) + (prev >> 1)
 * 
 */
          if (((rand() % 2) != 0)) {
            __pyx_t_10 = 0x8000;
          } else {
            __pyx_t_10 = -32768;
          }
          __pyx_v_randval = __pyx_t_10;

          /* "_pyfxr.pyx":650
 *         for i in range(delay):
 *             randval = 1 << 15 if rand() % 2 else -(1 << 15)
 *             prev = s.samples[i] = (randval >> 1) + (prev >> 1)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(delay, n_samples):
 */
          __pyx_t_11 = ((__pyx_v_randval >> 1) + (__pyx_v_prev >> 1));
          __pyx_v_prev = __pyx_t_11;
          (__pyx_v_s->samples[__pyx_v_i]) = __pyx_t_11;
        }

        /* "_pyfxr.pyx":652
 *             prev = s.samples[i] = (randval >> 1) + (prev >> 1)
 * 
 *         for i in range(delay, n_samples):             # <<<<<<<<<<<<<<
 *             prev = s.samples[i] = (s.samples[i - delay] >> 1) + (prev >> 1)
 * 
 */
        __pyx_t_3 = __pyx_v_n_samples;
        __pyx_t_1 = __pyx_t_3;
        for (__pyx_t_2 = __pyx_v_delay; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
          __pyx_v_i = __pyx_t_2;

          /* "_pyfxr.pyx":653
 * 
 *         for i in range(delay, n_samples):
 *             prev = s.samples[i] = (s.samples[i - delay] >> 1) + (prev >> 1)             # <<<<<<<<<<<<<<
 * 
 *         # Apply attack and release envelopes
 */
          __pyx_t_11 = (((__pyx_v_s->samples[(__pyx_v_i - __pyx_v_delay)]) >> 1) + (__pyx_v_prev >> 1));
          __pyx_v_prev = __pyx_t_11;
          (__pyx_v_s->samples[__pyx_v_i]) = __pyx_t_11;
        }

        /* "_pyfxr.pyx":656
 * 
 *         # Apply attack and release envelopes
 *         for pos in range(delay):             # <<<<<<<<<<<<<<
 *             fsample = s.samples[pos] / <float> (1 << 15)
 *             s.samples[pos] = samp(i * fsample / delay)
 */
        __pyx_t_3 = __pyx_v_delay;
        __pyx_t_1 = __pyx_t_3;
        for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
          __pyx_v_pos = __pyx_t_2;

          /* "_pyfxr.pyx":657
 *         # Apply attack and release envelopes
 *         for pos in range(delay):
 *             fsample = s.samples[pos] / <float> (1 << 15)             # <<<<<<<<<<<<<<
 *             s.samples[pos] = samp(i * fsample / delay)
 *         for i in range(release_samples):
 */
          __pyx_v_fsample = (((float)(__pyx_v_s->samples[__pyx_v_pos])) / ((float)0x8000));

          /* "_pyfxr.pyx":658
 *         for pos in range(delay):
 *             fsample = s.samples[pos] / <float> (1 << 15)
 *             s.samples[pos] = samp(i * fsample / delay)             # <<<<<<<<<<<<<<
 *         for i in range(release_samples):
 *             pos = n_samples - i - 1
 */
          (__pyx_v_s->samples[__pyx_v_pos]) = __pyx_f_6_pyfxr_samp(((__pyx_v_i * __pyx_v_fsample) / ((float)__pyx_v_delay)));
        }

        /* "_pyfxr.pyx":659
 *             fsample = s.samples[pos] / <float> (1 << 15)
 *             s.samples[pos] = samp(i * fsample / delay)
 *         for i in range(release_samples):             # <<<<<<<<<<<<<<
 *             pos = n_samples - i - 1
 *             fsample = s.samples[pos] / <float> (1 << 15)
 */
        __pyx_t_3 = __pyx_v_release_samples;
        __pyx_t_1 = __pyx_t_3;
        for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
          __pyx_v_i = __pyx_t_2;

          /* "_pyfxr.pyx":660
 *             s.samples[pos] = samp(i * fsample / delay)
 *         for i in range(release_samples):
 *             pos = n_samples - i - 1             # <<<<<<<<<<<<<<
 *             fsample = s.samples[pos] / <float> (1 << 15)
 *             s.samples[pos] = samp(i * fsample / release_samples)
 */
          __pyx_v_pos = ((__pyx_v_n_samples - __pyx_v_i) - 1);

          /* "_pyfxr.pyx":661
 *         for i in range(release_samples):
 *             pos = n_samples - i - 1
 *             fsample = s.samples[pos] / <float> (1 << 15)             # <<<<<<<<<<<<<<
 *             s.samples[pos] = samp(i * fsample / release_samples)
 * 
 */
          __pyx_v_fsample = (((float)(__pyx_v_s->samples[__pyx_v_pos])) / ((float)0x8000));

          /* "_pyfxr.pyx":662
 *             pos = n_samples - i - 1
 *             fsample = s.samples[pos] / <float> (1 << 15)
 *             s.samples[pos] = samp(i * fsample / release_samples)             # <<<<<<<<<<<<<<
 * 
 *     return s
 */
          (__pyx_v_s->samples[__pyx_v_pos]) = __pyx_f_6_pyfxr_samp(((__pyx_v_i * __pyx_v_fsample) / ((float)__pyx_v_release_samples)));
        }
      }

      /* "_pyfxr.pyx":646
 *     cdef SoundBuffer s = SoundBuffer(n_samples)
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         prev = 0
 *         for i in range(delay):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "_pyfxr.pyx":664
 *             s.samples[pos] = samp(i * fsample / release_samples)
 * 
 *     return s             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_s));
  __pyx_r = ((PyObject *)__pyx_v_s);
  goto __pyx_L0;

  /* "_pyfxr.pyx":630
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * def pluck(float duration, float pitch, float release=0.1):             # <<<<<<<<<<<<<<
 *     """Generate a pluck sound using the Karplus-Strong algorithm."""
 *     cdef size_t delay, i, n_samples, release_samples, pos
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("_pyfxr.pluck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Wavetable(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_7__pyx_unpickle_Wavetable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_6__pyx_unpickle_Wavetable[] = "__pyx_unpickle_Wavetable(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_6_pyfxr_7__pyx_unpickle_Wavetable = {"__pyx_unpickle_Wavetable", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_7__pyx_unpickle_Wavetable, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_6__pyx_unpickle_Wavetable};
static PyObject *__pyx_pw_6_pyfxr_7__pyx_unpickle_Wavetable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Wavetable (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Wavetable", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Wavetable", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Wavetable") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Wavetable", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.__pyx_unpickle_Wavetable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6_pyfxr_6__pyx_unpickle_Wavetable(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_6__pyx_unpickle_Wavetable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Wavetable", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x6f0316b:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6f0316b = (wavetable))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x6f0316b) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x6f0316b:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6f0316b = (wavetable))" % __pyx_checksum)
 *     __pyx_result = Wavetable.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x6f0316b:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6f0316b = (wavetable))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Wavetable.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x6f, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x6f0316b:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6f0316b = (wavetable))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6f0316b = (wavetable))" % __pyx_checksum)
 *     __pyx_result = Wavetable.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Wavetable__set_state(<Wavetable> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6_pyfxr_Wavetable), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6f0316b = (wavetable))" % __pyx_checksum)
 *     __pyx_result = Wavetable.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Wavetable__set_state(<Wavetable> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Wavetable.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Wavetable__set_state(<Wavetable> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Wavetable__set_state(Wavetable __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6_pyfxr___pyx_unpickle_Wavetable__set_state(((struct __pyx_obj_6_pyfxr_Wavetable *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6f0316b = (wavetable))" % __pyx_checksum)
 *     __pyx_result = Wavetable.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Wavetable__set_state(<Wavetable> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Wavetable__set_state(<Wavetable> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Wavetable__set_state(Wavetable __pyx_result, tuple __pyx_state):
 *     __pyx_result.wavetable = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Wavetable(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_pyfxr.__pyx_unpickle_Wavetable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Wavetable__set_state(<Wavetable> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Wavetable__set_state(Wavetable __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.wavetable = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6_pyfxr___pyx_unpickle_Wavetable__set_state(struct __pyx_obj_6_pyfxr_Wavetable *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int16_t __pyx_t_2[0x400];
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Wavetable__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Wavetable__set_state(Wavetable __pyx_result, tuple __pyx_state):
 *     __pyx_result.wavetable = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__Pyx_carray_from_py_int16_t(__pyx_t_1, __pyx_t_2, 0x400) < 0)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely((0x400) != (0x400))) {
    PyErr_Format(PyExc_ValueError, "Assignment to slice of wrong length, expected %" CYTHON_FORMAT_SSIZE_T "d, got %" CYTHON_FORMAT_SSIZE_T "d", (Py_ssize_t)(0x400), (Py_ssize_t)(0x400));
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  memcpy(&(__pyx_v___pyx_result->wavetable[0]), __pyx_t_2, sizeof(__pyx_v___pyx_result->wavetable[0]) * (0x400));

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Wavetable__set_state(Wavetable __pyx_result, tuple __pyx_state):
 *     __pyx_result.wavetable = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 1) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.wavetable = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Wavetable__set_state(Wavetable __pyx_result, tuple __pyx_state):
 *     __pyx_result.wavetable = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Wavetable__set_state(<Wavetable> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Wavetable__set_state(Wavetable __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.wavetable = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("_pyfxr.__pyx_unpickle_Wavetable__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PygletSource(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6_pyfxr_9__pyx_unpickle_PygletSource(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6_pyfxr_8__pyx_unpickle_PygletSource[] = "__pyx_unpickle_PygletSource(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_6_pyfxr_9__pyx_unpickle_PygletSource = {"__pyx_unpickle_PygletSource", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9__pyx_unpickle_PygletSource, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_8__pyx_unpickle_PygletSource};
static PyObject *__pyx_pw_6_pyfxr_9__pyx_unpickle_PygletSource(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PygletSource (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PygletSource", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PygletSource", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_PygletSource") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PygletSource", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("_pyfxr.__pyx_unpickle_PygletSource", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6_pyfxr_8__pyx_unpickle_PygletSource(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6_pyfxr_8__pyx_unpickle_PygletSource(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PygletSource", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x58d03cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x58d03cd = (buf, pos))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x58d03cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x58d03cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x58d03cd = (buf, pos))" % __pyx_checksum)
 *     __pyx_result = PygletSource.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x58d03cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x58d03cd = (buf, pos))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = PygletSource.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x58, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x58d03cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x58d03cd = (buf, pos))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x58d03cd = (buf, pos))" % __pyx_checksum)
 *     __pyx_result = PygletSource.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PygletSource__set_state(<PygletSource> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6_pyfxr_PygletSource), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x58d03cd = (buf, pos))" % __pyx_checksum)
 *     __pyx_result = PygletSource.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PygletSource__set_state(<PygletSource> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = PygletSource.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PygletSource__set_state(<PygletSource> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PygletSource__set_state(PygletSource __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6_pyfxr___pyx_unpickle_PygletSource__set_state(((struct __pyx_obj_6_pyfxr_PygletSource *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x58d03cd = (buf, pos))" % __pyx_checksum)
 *     __pyx_result = PygletSource.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PygletSource__set_state(<PygletSource> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PygletSource__set_state(<PygletSource> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PygletSource__set_state(PygletSource __pyx_result, tuple __pyx_state):
 *     __pyx_result.buf = __pyx_state[0]; __pyx_result.pos = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PygletSource(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_pyfxr.__pyx_unpickle_PygletSource", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PygletSource__set_state(<PygletSource> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PygletSource__set_state(PygletSource __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.buf = __pyx_state[0]; __pyx_result.pos = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6_pyfxr___pyx_unpickle_PygletSource__set_state(struct __pyx_obj_6_pyfxr_PygletSource *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PygletSource__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PygletSource__set_state(PygletSource __pyx_result, tuple __pyx_state):
 *     __pyx_result.buf = __pyx_state[0]; __pyx_result.pos = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6_pyfxr_SoundBuffer))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->buf);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->buf));
  __pyx_v___pyx_result->buf = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pos = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PygletSource__set_state(PygletSource __pyx_result, tuple __pyx_state):
 *     __pyx_result.buf = __pyx_state[0]; __pyx_result.pos = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 2) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.buf = __pyx_state[0]; __pyx_result.pos = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PygletSource__set_state(PygletSource __pyx_result, tuple __pyx_state):
 *     __pyx_result.buf = __pyx_state[0]; __pyx_result.pos = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PygletSource__set_state(<PygletSource> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PygletSource__set_state(PygletSource __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.buf = __pyx_state[0]; __pyx_result.pos = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("_pyfxr.__pyx_unpickle_PygletSource__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "carray.to_py":112
 * 
 * @cname("__Pyx_carray_to_py_int16_t")
 * cdef inline list __Pyx_carray_to_py_int16_t(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

static CYTHON_INLINE PyObject *__Pyx_carray_to_py_int16_t(int16_t *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  size_t __pyx_v_i;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_carray_to_py_int16_t", 0);

  /* "carray.to_py":115
 *     cdef size_t i
 *     cdef object value
 *     l = PyList_New(length)             # <<<<<<<<<<<<<<
 *     for i in range(<size_t>length):
 *         value = v[i]
 */
  __pyx_t_1 = PyList_New(__pyx_v_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_l = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "carray.to_py":116
 *     cdef object value
 *     l = PyList_New(length)
 *     for i in range(<size_t>length):             # <<<<<<<<<<<<<<
 *         value = v[i]
 *         Py_INCREF(value)
 */
  __pyx_t_2 = ((size_t)__pyx_v_length);
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "carray.to_py":117
 *     l = PyList_New(length)
 *     for i in range(<size_t>length):
 *         value = v[i]             # <<<<<<<<<<<<<<
 *         Py_INCREF(value)
 *         PyList_SET_ITEM(l, i, value)
 */
    __pyx_t_1 = __Pyx_PyInt_From_int16_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "carray.to_py":118
 *     for i in range(<size_t>length):
 *         value = v[i]
 *         Py_INCREF(value)             # <<<<<<<<<<<<<<
 *         PyList_SET_ITEM(l, i, value)
 *     return l
 */
    Py_INCREF(__pyx_v_value);

    /* "carray.to_py":119
 *         value = v[i]
 *         Py_INCREF(value)
 *         PyList_SET_ITEM(l, i, value)             # <<<<<<<<<<<<<<
 *     return l
 * 
 */
    PyList_SET_ITEM(__pyx_v_l, __pyx_v_i, __pyx_v_value);
  }

  /* "carray.to_py":120
 *         Py_INCREF(value)
 *         PyList_SET_ITEM(l, i, value)
 *     return l             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_l);
  __pyx_r = __pyx_v_l;
  goto __pyx_L0;

  /* "carray.to_py":112
 * 
 * @cname("__Pyx_carray_to_py_int16_t")
 * cdef inline list __Pyx_carray_to_py_int16_t(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("carray.to_py.__Pyx_carray_to_py_int16_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "carray.to_py":124
 * 
 * @cname("__Pyx_carray_to_tuple_int16_t")
 * cdef inline tuple __Pyx_carray_to_tuple_int16_t(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

static CYTHON_INLINE PyObject *__Pyx_carray_to_tuple_int16_t(int16_t *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  size_t __pyx_v_i;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_carray_to_tuple_int16_t", 0);

  /* "carray.to_py":127
 *     cdef size_t i
 *     cdef object value
 *     t = PyTuple_New(length)             # <<<<<<<<<<<<<<
 *     for i in range(<size_t>length):
 *         value = v[i]
 */
  __pyx_t_1 = PyTuple_New(__pyx_v_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_t = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "carray.to_py":128
 *     cdef object value
 *     t = PyTuple_New(length)
 *     for i in range(<size_t>length):             # <<<<<<<<<<<<<<
 *         value = v[i]
 *         Py_INCREF(value)
 */
  __pyx_t_2 = ((size_t)__pyx_v_length);
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "carray.to_py":129
 *     t = PyTuple_New(length)
 *     for i in range(<size_t>length):
 *         value = v[i]             # <<<<<<<<<<<<<<
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(t, i, value)
 */
    __pyx_t_1 = __Pyx_PyInt_From_int16_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "carray.to_py":130
 *     for i in range(<size_t>length):
 *         value = v[i]
 *         Py_INCREF(value)             # <<<<<<<<<<<<<<
 *         PyTuple_SET_ITEM(t, i, value)
 *     return t
 */
    Py_INCREF(__pyx_v_value);

    /* "carray.to_py":131
 *         value = v[i]
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(t, i, value)             # <<<<<<<<<<<<<<
 *     return t
 */
    PyTuple_SET_ITEM(__pyx_v_t, __pyx_v_i, __pyx_v_value);
  }

  /* "carray.to_py":132
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(t, i, value)
 *     return t             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_t);
  __pyx_r = __pyx_v_t;
  goto __pyx_L0;

  /* "carray.to_py":124
 * 
 * @cname("__Pyx_carray_to_tuple_int16_t")
 * cdef inline tuple __Pyx_carray_to_tuple_int16_t(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("carray.to_py.__Pyx_carray_to_tuple_int16_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "carray.from_py":77
 * 
 * @cname("__Pyx_carray_from_py_int16_t")
 * cdef int __Pyx_carray_from_py_int16_t(object o, base_type *v, Py_ssize_t length) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i = length
 *     try:
 */

static int __Pyx_carray_from_py_int16_t(PyObject *__pyx_v_o, int16_t *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_item = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  int16_t __pyx_t_11;
  char const *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_carray_from_py_int16_t", 0);

  /* "carray.from_py":78
 * @cname("__Pyx_carray_from_py_int16_t")
 * cdef int __Pyx_carray_from_py_int16_t(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length             # <<<<<<<<<<<<<<
 *     try:
 *         i = len(o)
 */
  __pyx_v_i = __pyx_v_length;

  /* "carray.from_py":79
 * cdef int __Pyx_carray_from_py_int16_t(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "carray.from_py":80
 *     cdef Py_ssize_t i = length
 *     try:
 *         i = len(o)             # <<<<<<<<<<<<<<
 *     except (TypeError, OverflowError):
 *         pass
 */
      __pyx_t_4 = PyObject_Length(__pyx_v_o); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 80, __pyx_L3_error)
      __pyx_v_i = __pyx_t_4;

      /* "carray.from_py":79
 * cdef int __Pyx_carray_from_py_int16_t(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "carray.from_py":81
 *     try:
 *         i = len(o)
 *     except (TypeError, OverflowError):             # <<<<<<<<<<<<<<
 *         pass
 *     if i == length:
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OverflowError);
    if (__pyx_t_5) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "carray.from_py":79
 * cdef int __Pyx_carray_from_py_int16_t(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "carray.from_py":83
 *     except (TypeError, OverflowError):
 *         pass
 *     if i == length:             # <<<<<<<<<<<<<<
 *         for i, item in enumerate(o):
 *             if i >= length:
 */
  __pyx_t_6 = ((__pyx_v_i == __pyx_v_length) != 0);
  if (__pyx_t_6) {

    /* "carray.from_py":84
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */
    __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
      __pyx_t_7 = __pyx_v_o; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 84, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 84, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(1, 84, __pyx_L1_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 84, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(1, 84, __pyx_L1_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 84, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        }
      } else {
        __pyx_t_10 = __pyx_t_9(__pyx_t_7);
        if (unlikely(!__pyx_t_10)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(1, 84, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_10);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_10);
      __pyx_t_10 = 0;
      __pyx_v_i = __pyx_t_4;
      __pyx_t_4 = (__pyx_t_4 + 1);

      /* "carray.from_py":85
 *     if i == length:
 *         for i, item in enumerate(o):
 *             if i >= length:             # <<<<<<<<<<<<<<
 *                 break
 *             v[i] = item
 */
      __pyx_t_6 = ((__pyx_v_i >= __pyx_v_length) != 0);
      if (__pyx_t_6) {

        /* "carray.from_py":86
 *         for i, item in enumerate(o):
 *             if i >= length:
 *                 break             # <<<<<<<<<<<<<<
 *             v[i] = item
 *         else:
 */
        goto __pyx_L11_break;

        /* "carray.from_py":85
 *     if i == length:
 *         for i, item in enumerate(o):
 *             if i >= length:             # <<<<<<<<<<<<<<
 *                 break
 *             v[i] = item
 */
      }

      /* "carray.from_py":87
 *             if i >= length:
 *                 break
 *             v[i] = item             # <<<<<<<<<<<<<<
 *         else:
 *             i += 1  # convert index to length
 */
      __pyx_t_11 = __Pyx_PyInt_As_int16_t(__pyx_v_item); if (unlikely((__pyx_t_11 == ((int16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 87, __pyx_L1_error)
      (__pyx_v_v[__pyx_v_i]) = __pyx_t_11;

      /* "carray.from_py":84
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */
    }
    /*else*/ {

      /* "carray.from_py":89
 *             v[i] = item
 *         else:
 *             i += 1  # convert index to length             # <<<<<<<<<<<<<<
 *             if i == length:
 *                 return 0
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "carray.from_py":90
 *         else:
 *             i += 1  # convert index to length
 *             if i == length:             # <<<<<<<<<<<<<<
 *                 return 0
 * 
 */
      __pyx_t_6 = ((__pyx_v_i == __pyx_v_length) != 0);
      if (__pyx_t_6) {

        /* "carray.from_py":91
 *             i += 1  # convert index to length
 *             if i == length:
 *                 return 0             # <<<<<<<<<<<<<<
 * 
 *     PyErr_Format(
 */
        __pyx_r = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L0;

        /* "carray.from_py":90
 *         else:
 *             i += 1  # convert index to length
 *             if i == length:             # <<<<<<<<<<<<<<
 *                 return 0
 * 
 */
      }
    }

    /* "carray.from_py":84
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */
    __pyx_L11_break:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "carray.from_py":83
 *     except (TypeError, OverflowError):
 *         pass
 *     if i == length:             # <<<<<<<<<<<<<<
 *         for i, item in enumerate(o):
 *             if i >= length:
 */
  }

  /* "carray.from_py":96
 *         IndexError,
 *         ("too many values found during array assignment, expected %zd"
 *          if i >= length else             # <<<<<<<<<<<<<<
 *          "not enough values found during array assignment, expected %zd, got %zd"),
 *         length, i)
 */
  if (((__pyx_v_i >= __pyx_v_length) != 0)) {
    __pyx_t_12 = ((char const *)"too many values found during array assignment, expected %zd");
  } else {
    __pyx_t_12 = ((char const *)"not enough values found during array assignment, expected %zd, got %zd");
  }

  /* "carray.from_py":93
 *                 return 0
 * 
 *     PyErr_Format(             # <<<<<<<<<<<<<<
 *         IndexError,
 *         ("too many values found during array assignment, expected %zd"
 */
  __pyx_t_7 = PyErr_Format(__pyx_builtin_IndexError, __pyx_t_12, __pyx_v_length, __pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "carray.from_py":77
 * 
 * @cname("__Pyx_carray_from_py_int16_t")
 * cdef int __Pyx_carray_from_py_int16_t(object o, base_type *v, Py_ssize_t length) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i = length
 *     try:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("carray.from_py.__Pyx_carray_from_py_int16_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_6_pyfxr_Wavetable(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_6_pyfxr_Wavetable(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_6_pyfxr_Wavetable[] = {
  {"from_function", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_3from_function, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_2from_function},
  {"sine", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_5sine, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_4sine},
  {"triangle", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_7triangle, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_6triangle},
  {"saw", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_9saw, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_8saw},
  {"square", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6_pyfxr_9Wavetable_11square, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6_pyfxr_9Wavetable_10square},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6_pyfxr_9Wavetable_17__reduce_cython__, METH_NOARGS, __pyx_doc_6_pyfxr_9Wavetable_16__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6_pyfxr_9Wavetable_19__setstate_cython__, METH_O, __pyx_doc_6_pyfxr_9Wavetable_18__setstate_cython__},
  {0, 0, 0, 0}
};

static PyBufferProcs __pyx_tp_as_buffer_Wavetable = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_6_pyfxr_9Wavetable_13__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_6_pyfxr_9Wavetable_15__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_6_pyfxr_Wavetable = {
  PyVarObject_HEAD_INIT(0, 0)
  "_pyfxr.Wavetable", /*tp_name*/
  sizeof(struct __pyx_obj_6_pyfxr_Wavetable), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6_pyfxr_Wavetable, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_Wavetable, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Wavetable(gen)", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6_pyfxr_Wavetable, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6_pyfxr_9Wavetable_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6_pyfxr_Wavetable, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_6_pyfxr_SoundBuffer(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_6_pyfxr_11SoundBuffer_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6_pyfxr_SoundBuffer(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6_pyfxr_11SoundBuffer_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_6_pyfxr_SoundBuffer(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_6_pyfxr_11SoundBuffer_duration(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6_pyfxr_11SoundBuffer_8duration_1__get__(o);
}

static PyMethodDef __pyx_methods_6_pyfxr_SoundBuffer[] = {
  {"save", (PyCFunction)__pyx_pw_6_pyfxr_11SoundBuffer_9save, METH_O, __pyx_doc_6_pyfxr_11SoundBuffer_8save},
  {"get_queue_source", (PyCFunction)__pyx_pw_6_pyfxr_11SoundBuffer_15get_queue_source, METH_NOARGS, __pyx_doc_6_pyfxr_11SoundBuffer_14get_queue_source},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6_pyfxr_11SoundBuffer_17__reduce_cython__, METH_NOARGS, __pyx_doc_6_pyfxr_11SoundBuffer_16__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6_pyfxr_11SoundBuffer_19__setstate_cython__, METH_O, __pyx_doc_6_pyfxr_11SoundBuffer_18__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6_pyfxr_SoundBuffer[] = {
  {(char *)"duration", __pyx_getprop_6_pyfxr_11SoundBuffer_duration, 0, (char *)"Get the duration of this sound in seconds, as a float.", 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_SoundBuffer = {
  __pyx_pw_6_pyfxr_11SoundBuffer_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6_pyfxr_SoundBuffer, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_SoundBuffer = {
  __pyx_pw_6_pyfxr_11SoundBuffer_5__len__, /*mp_length*/
  __pyx_pw_6_pyfxr_11SoundBuffer_7__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_SoundBuffer = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_6_pyfxr_11SoundBuffer_11__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_6_pyfxr_11SoundBuffer_13__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_6_pyfxr_SoundBuffer = {
  PyVarObject_HEAD_INIT(0, 0)
  "_pyfxr.SoundBuffer", /*tp_name*/
  sizeof(struct __pyx_obj_6_pyfxr_SoundBuffer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6_pyfxr_SoundBuffer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_SoundBuffer, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_SoundBuffer, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_SoundBuffer, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6_pyfxr_SoundBuffer, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6_pyfxr_SoundBuffer, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6_pyfxr_SoundBuffer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_6_pyfxr_CachedSound(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6_pyfxr_CachedSound *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6_pyfxr_CachedSound *)o);
  p->buf = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6_pyfxr_11CachedSound_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6_pyfxr_CachedSound(PyObject *o) {
  struct __pyx_obj_6_pyfxr_CachedSound *p = (struct __pyx_obj_6_pyfxr_CachedSound *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->buf);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6_pyfxr_CachedSound(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6_pyfxr_CachedSound *p = (struct __pyx_obj_6_pyfxr_CachedSound *)o;
  if (p->buf) {
    e = (*v)(((PyObject *)p->buf), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6_pyfxr_CachedSound(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6_pyfxr_CachedSound *p = (struct __pyx_obj_6_pyfxr_CachedSound *)o;
  tmp = ((PyObject*)p->buf);
  p->buf = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6_pyfxr_CachedSound[] = {
  {"_clear", (PyCFunction)__pyx_pw_6_pyfxr_11CachedSound_3_clear, METH_NOARGS, __pyx_doc_6_pyfxr_11CachedSound_2_clear},
  {"_set", (PyCFunction)__pyx_pw_6_pyfxr_11CachedSound_5_set, METH_O, __pyx_doc_6_pyfxr_11CachedSound_4_set},
  {"_get", (PyCFunction)__pyx_pw_6_pyfxr_11CachedSound_7_get, METH_NOARGS, __pyx_doc_6_pyfxr_11CachedSound_6_get},
  {"_build", (PyCFunction)__pyx_pw_6_pyfxr_11CachedSound_9_build, METH_NOARGS, __pyx_doc_6_pyfxr_11CachedSound_8_build},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6_pyfxr_11CachedSound_15__reduce_cython__, METH_NOARGS, __pyx_doc_6_pyfxr_11CachedSound_14__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6_pyfxr_11CachedSound_17__setstate_cython__, METH_O, __pyx_doc_6_pyfxr_11CachedSound_16__setstate_cython__},
  {0, 0, 0, 0}
};

static PyBufferProcs __pyx_tp_as_buffer_CachedSound = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_6_pyfxr_11CachedSound_11__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_6_pyfxr_11CachedSound_13__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_6_pyfxr_CachedSound = {
  PyVarObject_HEAD_INIT(0, 0)
  "_pyfxr.CachedSound", /*tp_name*/
  sizeof(struct __pyx_obj_6_pyfxr_CachedSound), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6_pyfxr_CachedSound, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_CachedSound, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6_pyfxr_CachedSound, /*tp_traverse*/
  __pyx_tp_clear_6_pyfxr_CachedSound, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6_pyfxr_CachedSound, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6_pyfxr_CachedSound, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_6_pyfxr_PygletSource(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6_pyfxr_PygletSource *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6_pyfxr_PygletSource *)o);
  p->buf = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6_pyfxr_PygletSource(PyObject *o) {
  struct __pyx_obj_6_pyfxr_PygletSource *p = (struct __pyx_obj_6_pyfxr_PygletSource *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->buf);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6_pyfxr_PygletSource(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6_pyfxr_PygletSource *p = (struct __pyx_obj_6_pyfxr_PygletSource *)o;
  if (p->buf) {
    e = (*v)(((PyObject *)p->buf), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6_pyfxr_PygletSource(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6_pyfxr_PygletSource *p = (struct __pyx_obj_6_pyfxr_PygletSource *)o;
  tmp = ((PyObject*)p->buf);
  p->buf = ((struct __pyx_obj_6_pyfxr_SoundBuffer *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6_pyfxr_12PygletSource_audio_format(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6_pyfxr_12PygletSource_12audio_format_1__get__(o);
}

static PyMethodDef __pyx_methods_6_pyfxr_PygletSource[] = {
  {"get_audio_data", (PyCFunction)__pyx_pw_6_pyfxr_12PygletSource_3get_audio_data, METH_O, __pyx_doc_6_pyfxr_12PygletSource_2get_audio_data},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6_pyfxr_12PygletSource_5__reduce_cython__, METH_NOARGS, __pyx_doc_6_pyfxr_12PygletSource_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6_pyfxr_12PygletSource_7__setstate_cython__, METH_O, __pyx_doc_6_pyfxr_12PygletSource_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6_pyfxr_PygletSource[] = {
  {(char *)"audio_format", __pyx_getprop_6_pyfxr_12PygletSource_audio_format, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6_pyfxr_PygletSource = {
  PyVarObject_HEAD_INIT(0, 0)
  "_pyfxr.PygletSource", /*tp_name*/
  sizeof(struct __pyx_obj_6_pyfxr_PygletSource), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6_pyfxr_PygletSource, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "PygletSource(SoundBuffer buf)", /*tp_doc*/
  __pyx_tp_traverse_6_pyfxr_PygletSource, /*tp_traverse*/
  __pyx_tp_clear_6_pyfxr_PygletSource, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6_pyfxr_PygletSource, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6_pyfxr_PygletSource, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6_pyfxr_12PygletSource_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6_pyfxr_PygletSource, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__pyfxr(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__pyfxr},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_pyfxr",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AudioData, __pyx_k_AudioData, sizeof(__pyx_k_AudioData), 0, 0, 1, 1},
  {&__pyx_n_s_AudioFormat, __pyx_k_AudioFormat, sizeof(__pyx_k_AudioFormat), 0, 0, 1, 1},
  {&__pyx_n_s_CachedSound, __pyx_k_CachedSound, sizeof(__pyx_k_CachedSound), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x58, __pyx_k_Incompatible_checksums_s_vs_0x58, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x58), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x6f, __pyx_k_Incompatible_checksums_s_vs_0x6f, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x6f), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_OverflowError, __pyx_k_OverflowError, sizeof(__pyx_k_OverflowError), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_PygletSource, __pyx_k_PygletSource, sizeof(__pyx_k_PygletSource), 0, 0, 1, 1},
  {&__pyx_n_s_SoundBuffer, __pyx_k_SoundBuffer, sizeof(__pyx_k_SoundBuffer), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_Wavetable, __pyx_k_Wavetable, sizeof(__pyx_k_Wavetable), 0, 0, 1, 1},
  {&__pyx_kp_u_Wavetable_generator_generated_to, __pyx_k_Wavetable_generator_generated_to, sizeof(__pyx_k_Wavetable_generator_generated_to), 0, 1, 0, 0},
  {&__pyx_n_s_amplitude, __pyx_k_amplitude, sizeof(__pyx_k_amplitude), 0, 0, 1, 1},
  {&__pyx_n_s_arp_limit, __pyx_k_arp_limit, sizeof(__pyx_k_arp_limit), 0, 0, 1, 1},
  {&__pyx_n_s_arp_mod, __pyx_k_arp_mod, sizeof(__pyx_k_arp_mod), 0, 0, 1, 1},
  {&__pyx_n_s_arp_time, __pyx_k_arp_time, sizeof(__pyx_k_arp_time), 0, 0, 1, 1},
  {&__pyx_n_s_attack, __pyx_k_attack, sizeof(__pyx_k_attack), 0, 0, 1, 1},
  {&__pyx_n_s_buf, __pyx_k_buf, sizeof(__pyx_k_buf), 0, 0, 1, 1},
  {&__pyx_n_s_build, __pyx_k_build, sizeof(__pyx_k_build), 0, 0, 1, 1},
  {&__pyx_n_s_channels, __pyx_k_channels, sizeof(__pyx_k_channels), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_decay, __pyx_k_decay, sizeof(__pyx_k_decay), 0, 0, 1, 1},
  {&__pyx_n_s_delay, __pyx_k_delay, sizeof(__pyx_k_delay), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_duration, __pyx_k_duration, sizeof(__pyx_k_duration), 0, 0, 1, 1},
  {&__pyx_n_s_duty_cycle, __pyx_k_duty_cycle, sizeof(__pyx_k_duty_cycle), 0, 0, 1, 1},
  {&__pyx_kp_u_duty_cycle_must_be_between_0_and, __pyx_k_duty_cycle_must_be_between_0_and, sizeof(__pyx_k_duty_cycle_must_be_between_0_and), 0, 1, 0, 0},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_env_length, __pyx_k_env_length, sizeof(__pyx_k_env_length), 0, 0, 1, 1},
  {&__pyx_n_s_env_stage, __pyx_k_env_stage, sizeof(__pyx_k_env_stage), 0, 0, 1, 1},
  {&__pyx_n_s_env_time, __pyx_k_env_time, sizeof(__pyx_k_env_time), 0, 0, 1, 1},
  {&__pyx_n_s_env_vol, __pyx_k_env_vol, sizeof(__pyx_k_env_vol), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_fdphase, __pyx_k_fdphase, sizeof(__pyx_k_fdphase), 0, 0, 1, 1},
  {&__pyx_n_s_fdslide, __pyx_k_fdslide, sizeof(__pyx_k_fdslide), 0, 0, 1, 1},
  {&__pyx_n_s_fltdmp, __pyx_k_fltdmp, sizeof(__pyx_k_fltdmp), 0, 0, 1, 1},
  {&__pyx_n_s_fltdp, __pyx_k_fltdp, sizeof(__pyx_k_fltdp), 0, 0, 1, 1},
  {&__pyx_n_s_flthp, __pyx_k_flthp, sizeof(__pyx_k_flthp), 0, 0, 1, 1},
  {&__pyx_n_s_flthp_d, __pyx_k_flthp_d, sizeof(__pyx_k_flthp_d), 0, 0, 1, 1},
  {&__pyx_n_s_fltp, __pyx_k_fltp, sizeof(__pyx_k_fltp), 0, 0, 1, 1},
  {&__pyx_n_s_fltphp, __pyx_k_fltphp, sizeof(__pyx_k_fltphp), 0, 0, 1, 1},
  {&__pyx_n_s_fltw, __pyx_k_fltw, sizeof(__pyx_k_fltw), 0, 0, 1, 1},
  {&__pyx_n_s_fltw_d, __pyx_k_fltw_d, sizeof(__pyx_k_fltw_d), 0, 0, 1, 1},
  {&__pyx_n_s_fmaxperiod, __pyx_k_fmaxperiod, sizeof(__pyx_k_fmaxperiod), 0, 0, 1, 1},
  {&__pyx_kp_u_for_pitch, __pyx_k_for_pitch, sizeof(__pyx_k_for_pitch), 0, 1, 0, 0},
  {&__pyx_n_s_fp, __pyx_k_fp, sizeof(__pyx_k_fp), 0, 0, 1, 1},
  {&__pyx_n_s_fperiod, __pyx_k_fperiod, sizeof(__pyx_k_fperiod), 0, 0, 1, 1},
  {&__pyx_n_s_fphase, __pyx_k_fphase, sizeof(__pyx_k_fphase), 0, 0, 1, 1},
  {&__pyx_n_s_from_function, __pyx_k_from_function, sizeof(__pyx_k_from_function), 0, 0, 1, 1},
  {&__pyx_n_s_fsample, __pyx_k_fsample, sizeof(__pyx_k_fsample), 0, 0, 1, 1},
  {&__pyx_n_s_fslide, __pyx_k_fslide, sizeof(__pyx_k_fslide), 0, 0, 1, 1},
  {&__pyx_n_s_gen, __pyx_k_gen, sizeof(__pyx_k_gen), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_kp_u_index_out_of_range, __pyx_k_index_out_of_range, sizeof(__pyx_k_index_out_of_range), 0, 1, 0, 0},
  {&__pyx_n_s_iphase, __pyx_k_iphase, sizeof(__pyx_k_iphase), 0, 0, 1, 1},
  {&__pyx_n_s_ipp, __pyx_k_ipp, sizeof(__pyx_k_ipp), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_n_samples, __pyx_k_n_samples, sizeof(__pyx_k_n_samples), 0, 0, 1, 1},
  {&__pyx_kp_u_n_samples_must_be_at_least, __pyx_k_n_samples_must_be_at_least, sizeof(__pyx_k_n_samples_must_be_at_least), 0, 1, 0, 0},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_ni, __pyx_k_ni, sizeof(__pyx_k_ni), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_noise_buffer, __pyx_k_noise_buffer, sizeof(__pyx_k_noise_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_omega, __pyx_k_omega, sizeof(__pyx_k_omega), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_p_arp_mod, __pyx_k_p_arp_mod, sizeof(__pyx_k_p_arp_mod), 0, 0, 1, 1},
  {&__pyx_n_s_p_arp_speed, __pyx_k_p_arp_speed, sizeof(__pyx_k_p_arp_speed), 0, 0, 1, 1},
  {&__pyx_n_s_p_base_freq, __pyx_k_p_base_freq, sizeof(__pyx_k_p_base_freq), 0, 0, 1, 1},
  {&__pyx_n_s_p_duty, __pyx_k_p_duty, sizeof(__pyx_k_p_duty), 0, 0, 1, 1},
  {&__pyx_n_s_p_duty_ramp, __pyx_k_p_duty_ramp, sizeof(__pyx_k_p_duty_ramp), 0, 0, 1, 1},
  {&__pyx_n_s_p_env_attack, __pyx_k_p_env_attack, sizeof(__pyx_k_p_env_attack), 0, 0, 1, 1},
  {&__pyx_n_s_p_env_decay, __pyx_k_p_env_decay, sizeof(__pyx_k_p_env_decay), 0, 0, 1, 1},
  {&__pyx_n_s_p_env_punch, __pyx_k_p_env_punch, sizeof(__pyx_k_p_env_punch), 0, 0, 1, 1},
  {&__pyx_n_s_p_env_sustain, __pyx_k_p_env_sustain, sizeof(__pyx_k_p_env_sustain), 0, 0, 1, 1},
  {&__pyx_n_s_p_freq_dramp, __pyx_k_p_freq_dramp, sizeof(__pyx_k_p_freq_dramp), 0, 0, 1, 1},
  {&__pyx_n_s_p_freq_limit, __pyx_k_p_freq_limit, sizeof(__pyx_k_p_freq_limit), 0, 0, 1, 1},
  {&__pyx_n_s_p_freq_ramp, __pyx_k_p_freq_ramp, sizeof(__pyx_k_p_freq_ramp), 0, 0, 1, 1},
  {&__pyx_n_s_p_hpf_freq, __pyx_k_p_hpf_freq, sizeof(__pyx_k_p_hpf_freq), 0, 0, 1, 1},
  {&__pyx_n_s_p_hpf_ramp, __pyx_k_p_hpf_ramp, sizeof(__pyx_k_p_hpf_ramp), 0, 0, 1, 1},
  {&__pyx_n_s_p_lpf_freq, __pyx_k_p_lpf_freq, sizeof(__pyx_k_p_lpf_freq), 0, 0, 1, 1},
  {&__pyx_n_s_p_lpf_ramp, __pyx_k_p_lpf_ramp, sizeof(__pyx_k_p_lpf_ramp), 0, 0, 1, 1},
  {&__pyx_n_s_p_lpf_resonance, __pyx_k_p_lpf_resonance, sizeof(__pyx_k_p_lpf_resonance), 0, 0, 1, 1},
  {&__pyx_n_s_p_pha_offset, __pyx_k_p_pha_offset, sizeof(__pyx_k_p_pha_offset), 0, 0, 1, 1},
  {&__pyx_n_s_p_pha_ramp, __pyx_k_p_pha_ramp, sizeof(__pyx_k_p_pha_ramp), 0, 0, 1, 1},
  {&__pyx_n_s_p_repeat_speed, __pyx_k_p_repeat_speed, sizeof(__pyx_k_p_repeat_speed), 0, 0, 1, 1},
  {&__pyx_n_s_p_vib_delay, __pyx_k_p_vib_delay, sizeof(__pyx_k_p_vib_delay), 0, 0, 1, 1},
  {&__pyx_n_s_p_vib_speed, __pyx_k_p_vib_speed, sizeof(__pyx_k_p_vib_speed), 0, 0, 1, 1},
  {&__pyx_n_s_p_vib_strength, __pyx_k_p_vib_strength, sizeof(__pyx_k_p_vib_strength), 0, 0, 1, 1},
  {&__pyx_n_s_period, __pyx_k_period, sizeof(__pyx_k_period), 0, 0, 1, 1},
  {&__pyx_n_s_phase, __pyx_k_phase, sizeof(__pyx_k_phase), 0, 0, 1, 1},
  {&__pyx_n_s_phaser_buffer, __pyx_k_phaser_buffer, sizeof(__pyx_k_phaser_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pitch, __pyx_k_pitch, sizeof(__pyx_k_pitch), 0, 0, 1, 1},
  {&__pyx_n_s_pluck, __pyx_k_pluck, sizeof(__pyx_k_pluck), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_s_pp, __pyx_k_pp, sizeof(__pyx_k_pp), 0, 0, 1, 1},
  {&__pyx_n_s_prev, __pyx_k_prev, sizeof(__pyx_k_prev), 0, 0, 1, 1},
  {&__pyx_n_s_pyfxr, __pyx_k_pyfxr, sizeof(__pyx_k_pyfxr), 0, 0, 1, 1},
  {&__pyx_kp_s_pyfxr_pyx, __pyx_k_pyfxr_pyx, sizeof(__pyx_k_pyfxr_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_pyglet_media_codecs, __pyx_k_pyglet_media_codecs, sizeof(__pyx_k_pyglet_media_codecs), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_PygletSource, __pyx_k_pyx_unpickle_PygletSource, sizeof(__pyx_k_pyx_unpickle_PygletSource), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Wavetable, __pyx_k_pyx_unpickle_Wavetable, sizeof(__pyx_k_pyx_unpickle_Wavetable), 0, 0, 1, 1},
  {&__pyx_n_s_randval, __pyx_k_randval, sizeof(__pyx_k_randval), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_release, __pyx_k_release, sizeof(__pyx_k_release), 0, 0, 1, 1},
  {&__pyx_n_s_release_samples, __pyx_k_release_samples, sizeof(__pyx_k_release_samples), 0, 0, 1, 1},
  {&__pyx_n_s_rep_limit, __pyx_k_rep_limit, sizeof(__pyx_k_rep_limit), 0, 0, 1, 1},
  {&__pyx_n_s_rep_time, __pyx_k_rep_time, sizeof(__pyx_k_rep_time), 0, 0, 1, 1},
  {&__pyx_n_s_rfperiod, __pyx_k_rfperiod, sizeof(__pyx_k_rfperiod), 0, 0, 1, 1},
  {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_sample, __pyx_k_sample, sizeof(__pyx_k_sample), 0, 0, 1, 1},
  {&__pyx_n_s_sample_rate, __pyx_k_sample_rate, sizeof(__pyx_k_sample_rate), 0, 0, 1, 1},
  {&__pyx_n_s_sample_size, __pyx_k_sample_size, sizeof(__pyx_k_sample_size), 0, 0, 1, 1},
  {&__pyx_n_s_samples, __pyx_k_samples, sizeof(__pyx_k_samples), 0, 0, 1, 1},
  {&__pyx_n_s_saw, __pyx_k_saw, sizeof(__pyx_k_saw), 0, 0, 1, 1},
  {&__pyx_n_s_setframerate, __pyx_k_setframerate, sizeof(__pyx_k_setframerate), 0, 0, 1, 1},
  {&__pyx_n_s_setnchannels, __pyx_k_setnchannels, sizeof(__pyx_k_setnchannels), 0, 0, 1, 1},
  {&__pyx_n_s_setnframes, __pyx_k_setnframes, sizeof(__pyx_k_setnframes), 0, 0, 1, 1},
  {&__pyx_n_s_setsampwidth, __pyx_k_setsampwidth, sizeof(__pyx_k_setsampwidth), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_sfx, __pyx_k_sfx, sizeof(__pyx_k_sfx), 0, 0, 1, 1},
  {&__pyx_n_s_si, __pyx_k_si, sizeof(__pyx_k_si), 0, 0, 1, 1},
  {&__pyx_n_s_sine, __pyx_k_sine, sizeof(__pyx_k_sine), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_square, __pyx_k_square, sizeof(__pyx_k_square), 0, 0, 1, 1},
  {&__pyx_n_s_square_duty, __pyx_k_square_duty, sizeof(__pyx_k_square_duty), 0, 0, 1, 1},
  {&__pyx_n_s_square_slide, __pyx_k_square_slide, sizeof(__pyx_k_square_slide), 0, 0, 1, 1},
  {&__pyx_n_s_ssample, __pyx_k_ssample, sizeof(__pyx_k_ssample), 0, 0, 1, 1},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_sustain, __pyx_k_sustain, sizeof(__pyx_k_sustain), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_tone, __pyx_k_tone, sizeof(__pyx_k_tone), 0, 0, 1, 1},
  {&__pyx_n_s_triangle, __pyx_k_triangle, sizeof(__pyx_k_triangle), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_vib_amp, __pyx_k_vib_amp, sizeof(__pyx_k_vib_amp), 0, 0, 1, 1},
  {&__pyx_n_s_vib_phase, __pyx_k_vib_phase, sizeof(__pyx_k_vib_phase), 0, 0, 1, 1},
  {&__pyx_n_s_vib_speed, __pyx_k_vib_speed, sizeof(__pyx_k_vib_speed), 0, 0, 1, 1},
  {&__pyx_n_s_video_format, __pyx_k_video_format, sizeof(__pyx_k_video_format), 0, 0, 1, 1},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {&__pyx_n_s_wave, __pyx_k_wave, sizeof(__pyx_k_wave), 0, 0, 1, 1},
  {&__pyx_n_s_wave_type, __pyx_k_wave_type, sizeof(__pyx_k_wave_type), 0, 0, 1, 1},
  {&__pyx_n_s_wavetable, __pyx_k_wavetable, sizeof(__pyx_k_wavetable), 0, 0, 1, 1},
  {&__pyx_n_u_wb, __pyx_k_wb, sizeof(__pyx_k_wb), 0, 1, 0, 1},
  {&__pyx_n_s_writeframesraw, __pyx_k_writeframesraw, sizeof(__pyx_k_writeframesraw), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 26, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 32, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_builtin_round = __Pyx_GetBuiltinName(__pyx_n_s_round); if (!__pyx_builtin_round) __PYX_ERR(0, 111, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 162, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_OverflowError = __Pyx_GetBuiltinName(__pyx_n_s_OverflowError); if (!__pyx_builtin_OverflowError) __PYX_ERR(1, 81, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "_pyfxr.pyx":32
 * 
 *         if i != 1023:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 "Wavetable generator generated too few values."
 *             )
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_Wavetable_generator_generated_to); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "_pyfxr.pyx":109
 * 
 *         if not (0.0 < duty_cycle < 1.0):
 *             raise ValueError("duty_cycle must be between 0 and 1")             # <<<<<<<<<<<<<<
 * 
 *         split = round(duty_cycle * 1024)
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_duty_cycle_must_be_between_0_and); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "_pyfxr.pyx":162
 *         if i >= 0:
 *             if i >= self.n_samples:
 *                 raise IndexError("index out of range")             # <<<<<<<<<<<<<<
 *         else:
 *             i = self.n_samples + i
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_u_index_out_of_range); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "_pyfxr.pyx":178
 *         import wave
 * 
 *         with wave.open(filename, 'wb') as wav:             # <<<<<<<<<<<<<<
 *             wav.setframerate(SAMPLE_RATE)
 *             wav.setnchannels(1)
 */
  __pyx_tuple__4 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "_pyfxr.pyx":37
 * 
 *     @staticmethod
 *     def from_function(f):             # <<<<<<<<<<<<<<
 *         """Generate a wavetable by calling a function f.
 * 
 */
  __pyx_tuple__9 = PyTuple_Pack(4, __pyx_n_s_f, __pyx_n_s_w, __pyx_n_s_t, __pyx_n_s_i); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_codeobj__10 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__9, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyfxr_pyx, __pyx_n_s_from_function, 37, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__10)) __PYX_ERR(0, 37, __pyx_L1_error)

  /* "_pyfxr.pyx":53
 * 
 *     @staticmethod
 *     def sine():             # <<<<<<<<<<<<<<
 *         """Construct a sine waveform."""
 *         cdef Wavetable w
 */
  __pyx_tuple__11 = PyTuple_Pack(2, __pyx_n_s_w, __pyx_n_s_i); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__11, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyfxr_pyx, __pyx_n_s_sine, 53, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(0, 53, __pyx_L1_error)

  /* "_pyfxr.pyx":66
 * 
 *     @staticmethod
 *     def triangle():             # <<<<<<<<<<<<<<
 *         """Construct a triangle waveform."""
 *         cdef Wavetable w
 */
  __pyx_tuple__13 = PyTuple_Pack(3, __pyx_n_s_w, __pyx_n_s_i, __pyx_n_s_v); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__13, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyfxr_pyx, __pyx_n_s_triangle, 66, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(0, 66, __pyx_L1_error)

  /* "_pyfxr.pyx":84
 * 
 *     @staticmethod
 *     def saw():             # <<<<<<<<<<<<<<
 *         """Construct a saw waveform."""
 *         cdef Wavetable w
 */
  __pyx_tuple__15 = PyTuple_Pack(3, __pyx_n_s_w, __pyx_n_s_i, __pyx_n_s_v); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyfxr_pyx, __pyx_n_s_saw, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "_pyfxr.pyx":96
 * 
 *     @staticmethod
 *     def square(float duty_cycle=0.5):             # <<<<<<<<<<<<<<
 *         """Generate a square-wave waveform.
 * 
 */
  __pyx_tuple__17 = PyTuple_Pack(5, __pyx_n_s_duty_cycle, __pyx_n_s_w, __pyx_n_s_i, __pyx_n_s_split, __pyx_n_s_v); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyfxr_pyx, __pyx_n_s_square, 96, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 96, __pyx_L1_error)

  /* "_pyfxr.pyx":277
 * 
 * 
 * def tone(             # <<<<<<<<<<<<<<
 *     Wavetable wavetable,
 *     double pitch=440.0,  # Hz, default = A
 */
  __pyx_tuple__19 = PyTuple_Pack(14, __pyx_n_s_wavetable, __pyx_n_s_pitch, __pyx_n_s_attack, __pyx_n_s_decay, __pyx_n_s_sustain, __pyx_n_s_release, __pyx_n_s_time, __pyx_n_s_n_samples, __pyx_n_s_i, __pyx_n_s_t, __pyx_n_s_samples, __pyx_n_s_omega, __pyx_n_s_v, __pyx_n_s_amplitude); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(6, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyfxr_pyx, __pyx_n_s_tone, 277, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 277, __pyx_L1_error)

  /* "_pyfxr.pyx":391
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * def sfx(             # <<<<<<<<<<<<<<
 *     int wave_type=0,
 *     float p_base_freq=0.3,
 */
  __pyx_tuple__21 = PyTuple_Pack(68, __pyx_n_s_wave_type, __pyx_n_s_p_base_freq, __pyx_n_s_p_freq_limit, __pyx_n_s_p_freq_ramp, __pyx_n_s_p_freq_dramp, __pyx_n_s_p_duty, __pyx_n_s_p_duty_ramp, __pyx_n_s_p_vib_strength, __pyx_n_s_p_vib_speed, __pyx_n_s_p_vib_delay, __pyx_n_s_p_env_attack, __pyx_n_s_p_env_sustain, __pyx_n_s_p_env_decay, __pyx_n_s_p_env_punch, __pyx_n_s_p_lpf_resonance, __pyx_n_s_p_lpf_freq, __pyx_n_s_p_lpf_ramp, __pyx_n_s_p_hpf_freq, __pyx_n_s_p_hpf_ramp, __pyx_n_s_p_pha_offset, __pyx_n_s_p_pha_ramp, __pyx_n_s_p_repeat_speed, __pyx_n_s_p_arp_speed, __pyx_n_s_p_arp_mod, __pyx_n_s_phase, __pyx_n_s_period, __pyx_n_s_env_stage, __pyx_n_s_env_time, __pyx_n_s_iphase, __pyx_n_s_ipp, __pyx_n_s_fperiod, __pyx_n_s_fmaxperiod, __pyx_n_s_fslide, __pyx_n_s_fdslide, __pyx_n_s_arp_mod, __pyx_n_s_square_duty, __pyx_n_s_square_slide, __pyx_n_s_env_vol, __pyx_n_s_fphase, __pyx_n_s_fdphase, __pyx_n_s_env_length, __pyx_n_s_phaser_buffer, __pyx_n_s_noise_buffer, __pyx_n_s_fltp, __pyx_n_s_fltdp, __pyx_n_s_fltphp, __pyx_n_s_fltw, __pyx_n_s_fltw_d, __pyx_n_s_fltdmp, __pyx_n_s_flthp, __pyx_n_s_flthp_d, __pyx_n_s_vib_phase, __pyx_n_s_vib_speed, __pyx_n_s_vib_amp, __pyx_n_s_rep_time, __pyx_n_s_rep_limit, __pyx_n_s_arp_time, __pyx_n_s_arp_limit, __pyx_n_s_rfperiod, __pyx_n_s_ssample, __pyx_n_s_sample, __pyx_n_s_fp, __pyx_n_s_pp, __pyx_n_s_si, __pyx_n_s_ni, __pyx_n_s_n_samples, __pyx_n_s_s, __pyx_n_s_i); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(24, 0, 68, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyfxr_pyx, __pyx_n_s_sfx, 391, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 391, __pyx_L1_error)

  /* "_pyfxr.pyx":630
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * def pluck(float duration, float pitch, float release=0.1):             # <<<<<<<<<<<<<<
 *     """Generate a pluck sound using the Karplus-Strong algorithm."""
 *     cdef size_t delay, i, n_samples, release_samples, pos
 */
  __pyx_tuple__23 = PyTuple_Pack(12, __pyx_n_s_duration, __pyx_n_s_pitch, __pyx_n_s_release, __pyx_n_s_delay, __pyx_n_s_i, __pyx_n_s_n_samples, __pyx_n_s_release_samples, __pyx_n_s_pos, __pyx_n_s_randval, __pyx_n_s_prev, __pyx_n_s_fsample, __pyx_n_s_s); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(3, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyfxr_pyx, __pyx_n_s_pluck, 630, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 630, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_Wavetable(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__25 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Wavetable, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__27 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PygletSource, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_512 = PyInt_FromLong(512); if (unlikely(!__pyx_int_512)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_93127629 = PyInt_FromLong(93127629L); if (unlikely(!__pyx_int_93127629)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_116404587 = PyInt_FromLong(116404587L); if (unlikely(!__pyx_int_116404587)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967296 = PyInt_FromString((char *)"4294967296", 0, 0); if (unlikely(!__pyx_int_4294967296)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6_pyfxr_Wavetable) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6_pyfxr_Wavetable.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6_pyfxr_Wavetable.tp_dictoffset && __pyx_type_6_pyfxr_Wavetable.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6_pyfxr_Wavetable.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Wavetable, (PyObject *)&__pyx_type_6_pyfxr_Wavetable) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6_pyfxr_Wavetable) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __pyx_ptype_6_pyfxr_Wavetable = &__pyx_type_6_pyfxr_Wavetable;
  if (PyType_Ready(&__pyx_type_6_pyfxr_SoundBuffer) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6_pyfxr_SoundBuffer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6_pyfxr_SoundBuffer.tp_dictoffset && __pyx_type_6_pyfxr_SoundBuffer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6_pyfxr_SoundBuffer.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SoundBuffer, (PyObject *)&__pyx_type_6_pyfxr_SoundBuffer) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6_pyfxr_SoundBuffer) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __pyx_ptype_6_pyfxr_SoundBuffer = &__pyx_type_6_pyfxr_SoundBuffer;
  if (PyType_Ready(&__pyx_type_6_pyfxr_CachedSound) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6_pyfxr_CachedSound.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6_pyfxr_CachedSound.tp_dictoffset && __pyx_type_6_pyfxr_CachedSound.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6_pyfxr_CachedSound.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CachedSound, (PyObject *)&__pyx_type_6_pyfxr_CachedSound) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6_pyfxr_CachedSound) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  __pyx_ptype_6_pyfxr_CachedSound = &__pyx_type_6_pyfxr_CachedSound;
  if (PyType_Ready(&__pyx_type_6_pyfxr_PygletSource) < 0) __PYX_ERR(0, 243, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6_pyfxr_PygletSource.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6_pyfxr_PygletSource.tp_dictoffset && __pyx_type_6_pyfxr_PygletSource.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6_pyfxr_PygletSource.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PygletSource, (PyObject *)&__pyx_type_6_pyfxr_PygletSource) < 0) __PYX_ERR(0, 243, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6_pyfxr_PygletSource) < 0) __PYX_ERR(0, 243, __pyx_L1_error)
  __pyx_ptype_6_pyfxr_PygletSource = &__pyx_type_6_pyfxr_PygletSource;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_pyfxr(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_pyfxr(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__pyfxr(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__pyfxr(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__pyfxr(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_pyfxr' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__pyfxr(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_pyfxr", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main__pyfxr) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "_pyfxr")) {
      if (unlikely(PyDict_SetItemString(modules, "_pyfxr", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "_pyfxr.pyx":12
 * 
 * 
 * cdef float AMPLITUDE = (1 << 15) - 1             # <<<<<<<<<<<<<<
 * cdef uint32_t SAMPLE_RATE = 44100
 * 
 */
  __pyx_v_6_pyfxr_AMPLITUDE = 32767.0;

  /* "_pyfxr.pyx":13
 * 
 * cdef float AMPLITUDE = (1 << 15) - 1
 * cdef uint32_t SAMPLE_RATE = 44100             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_6_pyfxr_SAMPLE_RATE = 0xAC44;

  /* "_pyfxr.pyx":37
 * 
 *     @staticmethod
 *     def from_function(f):             # <<<<<<<<<<<<<<
 *         """Generate a wavetable by calling a function f.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_9Wavetable_3from_function, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_from_function, __pyx_t_1) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":36
 *             )
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_function(f):
 *         """Generate a wavetable by calling a function f.
 */
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_6_pyfxr_Wavetable, __pyx_n_s_from_function); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_from_function, __pyx_t_2) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":53
 * 
 *     @staticmethod
 *     def sine():             # <<<<<<<<<<<<<<
 *         """Construct a sine waveform."""
 *         cdef Wavetable w
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_9Wavetable_5sine, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_sine, __pyx_t_2) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":52
 *         return w
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def sine():
 *         """Construct a sine waveform."""
 */
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_6_pyfxr_Wavetable, __pyx_n_s_sine); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_sine, __pyx_t_1) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":66
 * 
 *     @staticmethod
 *     def triangle():             # <<<<<<<<<<<<<<
 *         """Construct a triangle waveform."""
 *         cdef Wavetable w
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_9Wavetable_7triangle, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_triangle, __pyx_t_1) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":65
 *         return w
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def triangle():
 *         """Construct a triangle waveform."""
 */
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_6_pyfxr_Wavetable, __pyx_n_s_triangle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_triangle, __pyx_t_2) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":84
 * 
 *     @staticmethod
 *     def saw():             # <<<<<<<<<<<<<<
 *         """Construct a saw waveform."""
 *         cdef Wavetable w
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_9Wavetable_9saw, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_saw, __pyx_t_2) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":83
 *         return w
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def saw():
 *         """Construct a saw waveform."""
 */
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_6_pyfxr_Wavetable, __pyx_n_s_saw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_saw, __pyx_t_1) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":96
 * 
 *     @staticmethod
 *     def square(float duty_cycle=0.5):             # <<<<<<<<<<<<<<
 *         """Generate a square-wave waveform.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_9Wavetable_11square, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_square, __pyx_t_1) < 0) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":95
 *         return w
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def square(float duty_cycle=0.5):
 *         """Generate a square-wave waveform.
 */
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_6_pyfxr_Wavetable, __pyx_n_s_square); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_Wavetable->tp_dict, __pyx_n_s_square, __pyx_t_2) < 0) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_Wavetable);

  /* "_pyfxr.pyx":142
 *     cdef int16_t *samples
 * 
 *     sample_rate: int = SAMPLE_RATE             # <<<<<<<<<<<<<<
 *     channels: int = 1
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_6_pyfxr_SAMPLE_RATE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_SoundBuffer->tp_dict, __pyx_n_s_sample_rate, __pyx_t_2) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6_pyfxr_SoundBuffer);

  /* "_pyfxr.pyx":143
 * 
 *     sample_rate: int = SAMPLE_RATE
 *     channels: int = 1             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, size_t n_samples):
 */
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_SoundBuffer->tp_dict, __pyx_n_s_channels, __pyx_int_1) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6_pyfxr_SoundBuffer);

  /* "_pyfxr.pyx":251
 *         self.pos = 0
 * 
 *     video_format = None             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6_pyfxr_PygletSource->tp_dict, __pyx_n_s_video_format, Py_None) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6_pyfxr_PygletSource);

  /* "_pyfxr.pyx":277
 * 
 * 
 * def tone(             # <<<<<<<<<<<<<<
 *     Wavetable wavetable,
 *     double pitch=440.0,  # Hz, default = A
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_1tone, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tone, __pyx_t_2) < 0) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pyfxr.pyx":391
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * def sfx(             # <<<<<<<<<<<<<<
 *     int wave_type=0,
 *     float p_base_freq=0.3,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_3sfx, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sfx, __pyx_t_2) < 0) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pyfxr.pyx":630
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * def pluck(float duration, float pitch, float release=0.1):             # <<<<<<<<<<<<<<
 *     """Generate a pluck sound using the Karplus-Strong algorithm."""
 *     cdef size_t delay, i, n_samples, release_samples, pos
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_5pluck, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pluck, __pyx_t_2) < 0) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Wavetable(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_7__pyx_unpickle_Wavetable, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Wavetable, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Wavetable__set_state(<Wavetable> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Wavetable__set_state(Wavetable __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.wavetable = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6_pyfxr_9__pyx_unpickle_PygletSource, NULL, __pyx_n_s_pyfxr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PygletSource, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pyfxr.pyx":1
 * #cython: language_level=3             # <<<<<<<<<<<<<<
 * 
 * from libc.stdint cimport int16_t, uint32_t, uint64_t
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "carray.from_py":77
 * 
 * @cname("__Pyx_carray_from_py_int16_t")
 * cdef int __Pyx_carray_from_py_int16_t(object o, base_type *v, Py_ssize_t length) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i = length
 *     try:
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init _pyfxr", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init _pyfxr");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(b)
            if (8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53))) {
                return PyFloat_FromDouble((double)a / (double)b);
            }
            return PyInt_Type.tp_as_number->nb_true_divide(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            }
        }
                __Pyx_PyInt_TrueDivideObjC_ZeroDivisionError(b)
                if ((8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53)))
                        || __Pyx_sst_abs(size) <= 52 / PyLong_SHIFT) {
                    return PyFloat_FromDouble((double)a / (double)b);
                }
                return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            return PyLong_FromLong(x);
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            if (unlikely(zerodivision_check && b == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero");
                return NULL;
            }
            PyFPE_START_PROTECT("divide", return NULL)
            result = ((double)a) / (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned short    uint16_t;
        #else
           typedef unsigned __int16  uint16_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_size_t(size_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(size_t)*3+2];
    char *dpos, *end = digits + sizeof(size_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    size_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (size_t) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (size_t) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (size_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* GetNameInClass */
static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int16_t(int16_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int16_t neg_one = (int16_t) -1, const_zero = (int16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int16_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int16_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int16_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int16_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int16_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int16_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int16_t __Pyx_PyInt_As_int16_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int16_t neg_one = (int16_t) -1, const_zero = (int16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int16_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int16_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int16_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int16_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int16_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) >= 2 * PyLong_SHIFT) {
                            return (int16_t) (((((int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) >= 3 * PyLong_SHIFT) {
                            return (int16_t) (((((((int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) >= 4 * PyLong_SHIFT) {
                            return (int16_t) (((((((((int16_t)digits[3]) << PyLong_SHIFT) | (int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int16_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int16_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int16_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int16_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int16_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int16_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int16_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int16_t) (((int16_t)-1)*(((((int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int16_t) ((((((int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int16_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int16_t) (((int16_t)-1)*(((((((int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int16_t) ((((((((int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int16_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int16_t) (((int16_t)-1)*(((((((((int16_t)digits[3]) << PyLong_SHIFT) | (int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int16_t) ((((((((((int16_t)digits[3]) << PyLong_SHIFT) | (int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int16_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int16_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int16_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int16_t) -1;
        }
    } else {
        int16_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int16_t) -1;
        val = __Pyx_PyInt_As_int16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int16_t");
    return (int16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int16_t");
    return (int16_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint32_t) -1;
        }
    } else {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint64_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint64_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint64_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint64_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint64_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint64_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint64_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint64_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint64_t) -1;
        }
    } else {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyInt_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
